/*
 * Module:   AODV Routing Protocol for Castalia Simulator
 * Author:   Xu Zongque 
 * Version:  0.1
 * Contact:  Paolo Roberto Grassi <grassi@elet.polimi.it>
 * Project:  Multi Sensor and Actuator Networks
 * Owner:    Politecnico di Milano (MI) - ITALY
 *
 *                _/      _/              _/_/_/    _/_/    
 *               _/_/  _/_/  _/    _/  _/        _/    _/   
 *              _/  _/  _/  _/    _/    _/_/    _/_/_/_/    
 *             _/      _/  _/    _/        _/  _/    _/     
 *            _/      _/    _/_/_/  _/_/_/    _/    _/      
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "AodvRouting.h"
#include <csimulation.h>

Define_Module(AodvRouting);

void AodvRouting::startup()
{
	//parametres
	activeRouteTimeout = (double)par("activeRouteTimeout") / 1000.0;
	allowedHelloLoss = par("allowedHelloLoss");
	helloInterval = (double)par("helloInterval") / 1000.0;
	localAddTTL = par("localAddTTL");
	netDiameter = par("netDiameter");
	nodeTraversalTime = (double)par("nodeTraversalTime") / 1000.0;
	rerrRatelimit = par("rerrRatelimit");
	rreqRetries = par("rreqRetries");
	rreqRatelimit = par("rreqRatelimit");
	timeoutBuffer = par("timeoutBuffer");
	ttlStart = par("ttlStart");
	ttlIncrement = par("ttlIncrement");
	ttlThreshould = par("ttlThreshould");

	currSeqNum = 1; //start from 1, 0 represents invalid value
	currRreqID = 0; //start from 1, initialized by 0 because it will be incremented before the first rreq was sent

	rt = new RoutingTable();

}

void AodvRouting::finish(){
	VirtualRouting::finish();
	AodvRREQPacket* rreqpkt;
	AodvRERRPacket* rerrpkt;
	// clear the buffer
	while (!rreqBuffer.empty()) {
		rreqpkt = rreqBuffer.front();
		rreqBuffer.pop();
		cancelAndDelete(rreqpkt);
	}
	while (!rerrBuffer.empty()) {
		rerrpkt = rerrBuffer.front();
		rerrBuffer.pop();
		cancelAndDelete(rerrpkt);
	}
}

void AodvRouting::setRreqBoardcastDropTimer(const char* dest, const char* source, int id){
	if(checkExpireEntry(&rreqBroadcast, source, id)){
		return;
	}
	RreqExpireTimerSet rets;
	rets.canceled = false;
	rets.dest = string(dest);
	rets.lifetime = simTime().dbl() + 2 * nodeTraversalTime * 1000; //PATH_DISCOVERY_TIME = 2 * NODE_TRAVERSAL_TIME
	rets.originator = string(source);
	rets.rreqID = id;
	//if the first to expire changes
	if(!rreqBroadcast.empty() && rets.lifetime < rreqBroadcast.top().lifetime){
		cancelTimer(AODV_RREQ_BOARDCAST_DROP_TIMER);
	}
	rreqBroadcast.push(rets);
	if(getTimer(AODV_RREQ_BOARDCAST_DROP_TIMER).dbl()<=0){
		setTimer(AODV_RREQ_BOARDCAST_DROP_TIMER, rreqBroadcast.top().lifetime - simTime().dbl());
	}

}

bool AodvRouting::isInRreqBlacklist(const char* dest){
    for(list<string>::const_iterator i=rreqBlacklist.begin();i!=rreqBlacklist.end();++i){
        string s = *i;
        if(s.compare(dest)==0){
            return true;
        }
    }
    return false;
}

bool AodvRouting::checkExpireEntry(const priority_queue <RreqExpireTimerSet, vector<RreqExpireTimerSet>, RreqExpireLifetimeCompare>* exq,
								const char* dest){
	const RreqExpireTimerSet* q0;
	if(!(exq->empty())){
		q0 = &(exq->top());
		for(unsigned int i = 0; i < exq->size(); i++){
			if(q0[i].dest.compare(dest)==0){
				return true;
			}
		}
	}
	return false;
}

void AodvRouting::cancelExpireEntry(priority_queue <RreqExpireTimerSet, vector<RreqExpireTimerSet>, RreqExpireLifetimeCompare>* exq,
								const char* dest){
	RreqExpireTimerSet* q0;
		if(!(exq->empty())){
			q0 = (RreqExpireTimerSet*)&(exq->top());
			for(unsigned int i = 0; i < exq->size(); i++){
				if(q0[i].dest.compare(dest)==0){
					q0[i].canceled = true;
					return;
				}
			}
		}
		return;
}

void AodvRouting::setRrepAckTimer(const char* neib){
	//check if is already waiting for the neighbor to ack
	const RoutingTableEntryExpireTimerSet* q0;
	if(!(rrepAckExpire.empty())){
		q0 = &(rrepAckExpire.top());
		for(unsigned int i = 0; i < rrepAckExpire.size(); i++){
			if(q0[i].destination.compare(neib)==0){
				return;
			}
		}
	}
	rrepAcked[string(neib)]=false;
	RoutingTableEntryExpireTimerSet ratts;
	ratts.destination = string(neib);
	ratts.lifetime = simTime().dbl() + 2 * nodeTraversalTime;
	rrepAckExpire.push(ratts);
	if(getTimer(AODV_RREP_ACK_WAIT_TIMER).dbl()<=0){
		setTimer(AODV_RREP_ACK_WAIT_TIMER, rrepAckExpire.top().lifetime - simTime().dbl());
	}
}

void AodvRouting::setRreqBlacklistTimer(const char* neib){
	//check if is already in the blacklist
	const RoutingTableEntryExpireTimerSet* q0;
	if(!(rreqBlacklistTimeout.empty())){
		q0 = &(rreqBlacklistTimeout.top());
		for(unsigned int i = 0; i < rreqBlacklistTimeout.size(); i++){
			if(q0[i].destination.compare(neib)==0){
				return;
			}
		}
	}
	RoutingTableEntryExpireTimerSet rbetts;
	rbetts.destination = string(neib);
	//BLACKLIST_TIMEOUT = RREQ_RETRIES * NET_TRAVERSAL_TIME
	//NET_TRAVERSAL_TIME = 2 * NODE_TREVERSAL_TIME * NET_DIAMETER
	rbetts.lifetime = simTime().dbl() + 2 * rreqRetries * nodeTraversalTime * netDiameter;
	rreqBlacklistTimeout.push(rbetts);
	if(getTimer(AODV_RREQ_BLACKLIST_REMOVE_TIMER).dbl()<=0){
		setTimer(AODV_RREQ_BLACKLIST_REMOVE_TIMER, rreqBlacklistTimeout.top().lifetime - simTime().dbl());
	}
}

void AodvRouting::setRreqExpirationTimer(const char* dest){
	if(checkExpireEntry(&rreqRequest, dest)){
		return;
	}
	//wait time should be (2^RETRY_COUNT)*NET_TRAVERSAL_TIME
	//NET_TRAVERSAL_TIME = 2*NODE_TRAVERSAL_TIME*NET_DIAMETER
	double expTime = 2 * netDiameter * nodeTraversalTime;
	for (int i=1;i<rreqRetryCount[string(dest)];i++){
		expTime *= 2;
	}
	RreqExpireTimerSet rets;
	rets.canceled = false;
	rets.dest = string(dest);
	rets.lifetime = simTime().dbl() + expTime; //PATH_DISCOVERY_TIME = 2 * NODE_TRAVERSAL_TIME
	trace() << "exptime is: " << simTime().dbl() + expTime;
	rets.originator = string(SELF_NETWORK_ADDRESS);
	rets.rreqID = currRreqID;
	//if the first to expire changes
	if(!rreqRequest.empty() && rets.lifetime < rreqRequest.top().lifetime){
		cancelTimer(AODV_RREQ_EXPIRATION_TIMER);
	}
	rreqRequest.push(rets);
	if(getTimer(AODV_RREQ_EXPIRATION_TIMER).dbl()<=0){
		setTimer(AODV_RREQ_EXPIRATION_TIMER, rreqRequest.top().lifetime - simTime().dbl());
		trace() << "10 - timer set for " << rreqRequest.top().lifetime - simTime().dbl() << " seconds later";
	}

}

bool AodvRouting::isRreqRequstActive(const char* dest){
	const RreqExpireTimerSet* q0;
	AodvRREQPacket** q1;
	//on going
	if(!(rreqRequest.empty())){
		q0 = &(rreqRequest.top());
		for(unsigned int i = 0; i < rreqRequest.size(); i++){
			if(q0[i].dest.compare(dest)==0){
				return true;
			}
		}
	}
	//buffered
	if(!rreqBuffer.empty()){
		q1 = &(rreqBuffer.front());
		for(unsigned int i = 0; i < rreqBuffer.size(); i++){
			if(string(q1[i]->getDestinationAodv()).compare(dest)==0){
				return true;
			}
		}
	}
	return false;
}

bool AodvRouting::checkExpireEntry(const priority_queue <RreqExpireTimerSet, vector<RreqExpireTimerSet>, RreqExpireLifetimeCompare>* exq,
		const char* orig, int idx){
	const RreqExpireTimerSet* q0;
	if(!(exq->empty())){
		q0 = &(exq->top());
		for(unsigned int i = 0; i < exq->size(); i++){
			if(q0[i].rreqID==idx && q0[i].originator.compare(orig)==0){
				return true;
			}
		}
	}
	return false;
}

double AodvRouting::getRemainingLifetimeTime(const char * dest){
	const RoutingTableEntryExpireTimerSet* q0;
	double x = 0;
	if(!rtExpireSeq.empty()){
		q0 = &rtExpireSeq.top();
		for (int i = rtExpireSeq.size() - 1 ; i >=0; i--){
			if (!q0[i].canceled && q0[i].destination.compare(dest)==0){
				if (q0[i].lifetime > x){
					x = q0[i].lifetime;
				}
			}
		}
	}
	return x;
}

void AodvRouting::updateLifetimeEntry(const char * dest, double time){
	if(rt->getState(dest)!=VALID){
		return;
	}
	RoutingTableEntryExpireTimerSet rteet;
	rteluc[string(dest)]++;
	rteet.destination = string(dest);
	rteet.lifetime = simTime().dbl() + time;
	rteet.canceled = false;
	//trace() << "rtlf04 - lifetime for " << string(dest) << " updated. count " << rteluc[string(dest)] << " value " << rteet.lifetime;
	rtExpireSeq.push(rteet);
}

void AodvRouting::resetLifetimeEntry(const char * dest){
	RoutingTableEntryExpireTimerSet* q0;
	if(!(rtExpireSeq.empty())){
		q0 = (RoutingTableEntryExpireTimerSet*)&(rtExpireSeq.top());
		for(unsigned int i = 0; i < rtExpireSeq.size(); i++){
			if(q0[i].destination.compare(dest)==0){
				q0[i].canceled = true;
				return;
			}
		}
	}
	rteluc[string(dest)]=0;
	return;
}

void AodvRouting::updateRTEntry(const char* destinationAddr,unsigned long destinationSeqNum,
					bool isDestinationValid,RoutingFlag state,int hopCount,
					const char* nextHopAddr,std::list<std::string>* precursor){
	//refer to RFC3561 chapter 6.2
	RoutingTableEntryExpireTimerSet rteet;
	double oldLifetime = getRemainingLifetimeTime(destinationAddr);
	rteluc[string(destinationAddr)]++;
	rteet.destination = string(destinationAddr);
	rteet.lifetime = oldLifetime<=0 ? simTime().dbl() + activeRouteTimeout : oldLifetime + activeRouteTimeout;
	rteet.canceled = false;
	//trace() << "rtlf02 - lifetime for " << string(destinationAddr) << " updated. count " << rteluc[string(destinationAddr)] << " value " << rteet.lifetime;
	//if the first to expire changes
	if(!rtExpireSeq.empty() && rteet.lifetime < rtExpireSeq.top().lifetime){
		cancelTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER);
	}
	rtExpireSeq.push(rteet);
	if(getTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER).dbl()<=0){
		setTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER, rtExpireSeq.top().lifetime - simTime().dbl());
	}
	rt->updateEntry(destinationAddr,  destinationSeqNum, isDestinationValid, state, hopCount, nextHopAddr, precursor);
}

void AodvRouting::updateRTEntry(const char* destinationAddr,unsigned long destinationSeqNum,
					bool isDestinationValid,RoutingFlag state,int hopCount,
					const char* nextHopAddr,std::list<std::string>* precursor,
					double lifetime){
	RoutingTableEntryExpireTimerSet rteet;
	double oldLifetime = getRemainingLifetimeTime(destinationAddr);
	rteluc[string(destinationAddr)]++;
	rteet.destination = string(destinationAddr);
	rteet.lifetime = oldLifetime<=0 ? simTime().dbl() + lifetime : oldLifetime + lifetime;
	rteet.canceled = false;
	//trace() << "rtlf01 - lifetime for " << string(destinationAddr) << " updated. count " << rteluc[string(destinationAddr)] << " value " << rteet.lifetime;
	//if the first to expire changes
	if(!rtExpireSeq.empty() && rteet.lifetime < rtExpireSeq.top().lifetime){
		cancelTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER);
	}
	rtExpireSeq.push(rteet);
	if(getTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER).dbl()<=0){
		setTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER, rtExpireSeq.top().lifetime - simTime().dbl());

	}
	rt->updateEntry(destinationAddr,  destinationSeqNum, isDestinationValid, state, hopCount, nextHopAddr, precursor);
}

void AodvRouting::sendRreqRequest(int hopCount, int rreqID, const char* sourceAodv, const char* destAodv,
								unsigned long sSeq, unsigned long dSeq){
	AodvRREQPacket *rreqPacket =
	    new AodvRREQPacket("AODV routing RREQ packet", NETWORK_LAYER_PACKET);
	rreqPacket->setFlagD(false);
	rreqPacket->setFlagG(false);
	rreqPacket->setFlagJ(false);
	rreqPacket->setFlagR(false);
	rreqPacket->setFlagU(false);
	rreqPacket->setHopCount(hopCount);
	rreqPacket->setRreqID(rreqID);
	rreqPacket->setDestinationAodv(destAodv);
	rreqPacket->setDestinationSeqNum(dSeq);
	rreqPacket->setSourceAodv(sourceAodv);
	rreqPacket->setSourceSeqNum(sSeq);
	rreqPacket->setSource(SELF_NETWORK_ADDRESS);
	rreqPacket->setDestination(destAodv);
	if (getTimer(AODV_RREQ_RATE_LIMIT_TIMER).dbl() <= 0){
		++rreqRetryCount[string(destAodv)];
		setRreqBoardcastDropTimer(destAodv, sourceAodv, rreqID);
		if(string(sourceAodv).compare(SELF_NETWORK_ADDRESS)==0){
			trace() << "08A - set the rreq expire timer for dest: " << string(destAodv) << " id: " << rreqID;
			setRreqExpirationTimer(destAodv);
		}
		setTimer(AODV_RREQ_RATE_LIMIT_TIMER, (double)1/rreqRatelimit);
		trace() << "01 - broadcast rreq request for destination: " << string(destAodv) << " sourceAddr: " << string(sourceAodv) << " id: " << rreqID;
		toMacLayer(rreqPacket, BROADCAST_MAC_ADDRESS);
	} else {
		rreqBuffer.push(rreqPacket);
	}

}


void AodvRouting::fromApplicationLayer(cPacket * pkt, const char *destination)
{
	//the application module can only send data packets
	DataPacket *dataPacket =
	    new DataPacket("AODV routing data packet", NETWORK_LAYER_PACKET);
	dataPacket->setSource(SELF_NETWORK_ADDRESS);
	dataPacket->setDestinationAodv(destination);
	dataPacket->setDestination(destination);
	encapsulatePacket(dataPacket, pkt);
	if(string(destination).compare(BROADCAST_NETWORK_ADDRESS)==0){
		toMacLayer(dataPacket, BROADCAST_MAC_ADDRESS);
		return;
	}
	
	//refer to RFC3561 chapter 6.3
	//if in the routing table there is already an valid entry for the destination, send the package
	if(rt->isRouteValid(destination)){
		updateLifetimeEntry(destination, activeRouteTimeout);
		trace() << "06 - send data packet. next hop: " << string(rt->getNextHop(destination))
				<< " destination: " << string(destination);
		dataPacket->setDestination(rt->getNextHop(destination));
		toMacLayer(dataPacket, resolveNetworkAddress(rt->getNextHop(destination)));
		return;
	} else if(!isRreqRequstActive(destination)){
		//if in current time there are no rreq requests on going for the destination
		//send the RREQ to find a route
		currSeqNum++;
		currRreqID++;
		sendRreqRequest(0, currRreqID,SELF_NETWORK_ADDRESS, destination,
						currSeqNum, rt->getDestSeqNum(destination));
	}
	//buffer the packet
	bufferPacket(dataPacket);
}

void AodvRouting::fromMacLayer(cPacket * pkt, int srcMacAddress, double rssi, double lqi)
{

	PacketTypes packetType;
	//get the packet type
	AodvRREPPacket *rrepPacket = dynamic_cast <AodvRREPPacket*>(pkt);
	AodvRREQPacket *rreqPacket = dynamic_cast <AodvRREQPacket*>(pkt);
	AodvRERRPacket *rerrPacket = dynamic_cast <AodvRERRPacket*>(pkt);
	AodvHelloMessage* helloMsg = dynamic_cast <AodvHelloMessage*>(pkt);
	AodvRREPAckPacket* rrepaPacket = dynamic_cast <AodvRREPAckPacket*>(pkt);
	DataPacket *dataPacket = dynamic_cast <DataPacket*>(pkt);
	if (dataPacket){
		packetType = AODV_DATA_PACKET;
	} else if (helloMsg){
		packetType = AODV_HELLO_MESSAGE_PACKET;
	} else if(rreqPacket){
		packetType = AODV_RREQ_PACKET;
	} else if(rrepPacket){
		packetType = AODV_RREP_PACKET;
	} else if(rrepaPacket){
		packetType = AODV_RREP_ACK_PACKET;
	} else if(rerrPacket){
		packetType = AODV_RERR_PACKET;
	} else{
		packetType = AODV_UNKNOWN_PACKET_TYPE;
	}

	switch(packetType){
		case AODV_DATA_PACKET:
			if(string(dataPacket->getDestinationAodv()).compare(BROADCAST_NETWORK_ADDRESS)==0){
				toApplicationLayer(dataPacket->decapsulate());
				return;
			}
			if(string(dataPacket->getDestination()).compare(SELF_NETWORK_ADDRESS)!=0){
				return;
			}
			trace() << "07 - received data packet. source: " << string(dataPacket->getSource())
					<< " dest hop: " << string(dataPacket->getDestination())
					<< " final destination: " << string(dataPacket->getDestinationAodv());
			processDataPackage(dataPacket);
			break;
		case AODV_RREQ_PACKET:
			//trace() << "02 - received rreq packet. source: " << string(rreqPacket->getSourceAodv())
			//		<< " id: " << rreqPacket->getRreqID()
			//		<< " destination: " << string(rreqPacket->getDestinationAodv())
			//		<< " packet from: " << srcMacAddress << " / " << string(rreqPacket->getSource());
			processAodvRreqPackage(rreqPacket, srcMacAddress, rssi, lqi);
			break;
		case AODV_RREP_PACKET:
			if(string(rrepPacket->getDestination()).compare(SELF_NETWORK_ADDRESS)!=0){
				//trace() << "03B - received WRONG rrep packet. " << (void*)rrepPacket << " originator: " << string(rrepPacket->getOriginator())
				//					<< " destination(for aodv information): " << string(rrepPacket->getDestinationAodv())
				//					<< " last hop: " << srcMacAddress << " / " << rrepPacket->getSource()
				//					<< " destMac: " << string(rrepPacket->getDestination())
				//					<< " hopCount: " << rrepPacket->getHopCount();
				return;
			}
			trace() << "03A - received rrep packet. " << (void*)rrepPacket << " originator: " << string(rrepPacket->getOriginator())
					<< " destination(for aodv information): " << string(rrepPacket->getDestinationAodv())
					<< " last hop: " << srcMacAddress << " / " << rrepPacket->getSource()
					<< " destMac: " << string(rrepPacket->getDestination())
					<< " hopCount: " << rrepPacket->getHopCount();
			processAodvRrepPackage(rrepPacket, srcMacAddress, rssi, lqi);
			break;
		case AODV_RREP_ACK_PACKET:
			rrepAcked[string(rrepaPacket->getSource())] = true;
			break;
		case AODV_HELLO_MESSAGE_PACKET:
			processHelloMsg(helloMsg);
			break;
		case AODV_RERR_PACKET:
			processAodvRerrPackage(rerrPacket, srcMacAddress, rssi, lqi);
			break;
		default:
			return;
	}
	return;

}

void AodvRouting::processDataPackage(DataPacket* pkt){
	DataPacket *ndPacket = pkt->dup();
	if(string(pkt->getDestinationAodv()).compare(SELF_NETWORK_ADDRESS)==0){
		toApplicationLayer(pkt->decapsulate());
		trace() << "12 - receiving data packet, give packet to app layer. source: " << string(ndPacket->getSource());
		return;
	} else if(rt->isRouteValid(pkt->getDestinationAodv())&&rt->getNextHop(pkt->getDestinationAodv())){
		updateLifetimeEntry(pkt->getDestinationAodv(), activeRouteTimeout);
		ndPacket->setDestination(rt->getNextHop(pkt->getDestinationAodv()));
		trace() << "11 - forwarding data packet. source: " << string(ndPacket->getSource())
							<< " next hop: " << string(rt->getNextHop(pkt->getDestinationAodv()))
							<< " final destination: " << string(ndPacket->getDestinationAodv());
		toMacLayer(ndPacket, resolveNetworkAddress(rt->getNextHop(pkt->getDestinationAodv())));
		return;
	} else if(!isRreqRequstActive(pkt->getDestinationAodv())){
		//if in current time there are no rreq requests on going for the destination
		//send the RREQ to find a route
		currSeqNum++;
		currRreqID++;
		sendRreqRequest(0, currRreqID,SELF_NETWORK_ADDRESS, pkt->getDestinationAodv(),
								currSeqNum, rt->getDestSeqNum(pkt->getDestinationAodv()));
		//buffer the packet
		bufferPacket(ndPacket);
	}
}

void AodvRouting::processAodvRreqPackage(AodvRREQPacket * rreqPackage,int srcMacAddress, double rssi, double lqi){
	if(isInRreqBlacklist(rreqPackage->getSource())){
		return;
	}
	//updates a route to the previous hop without a valid seq number
	updateRTEntry(rreqPackage->getSource(), 0, false, VALID, 1, rreqPackage->getSource(),NULL);

	//check if this node has already processed for this rreq, if yes drop packet
	if(checkExpireEntry(&rreqBroadcast,rreqPackage->getSourceAodv(), rreqPackage->getRreqID())){
		return;
	}
	//update route for the originator
	updateRTEntry(rreqPackage->getSourceAodv(), rreqPackage->getSourceSeqNum(), true, VALID, rreqPackage->getHopCount() + 1, rreqPackage->getSource(),NULL);

	if(!((string(rreqPackage->getDestinationAodv()).compare(SELF_NETWORK_ADDRESS)==0))
			/*&&(!rt->getNextHop(rreqPackage->getDestinationAodv()))*/){ // PRG: this line was commented but without commenting it's faster
		//forwarding the rreq
		//refer to RFC3561 chapter 6.5
		sendRreqRequest(rreqPackage->getHopCount() + 1,
						rreqPackage->getRreqID(),
						rreqPackage->getSourceAodv(),
						rreqPackage->getDestinationAodv(),
						rreqPackage->getSourceSeqNum(),
						rt->getDestSeqNum(rreqPackage->getDestinationAodv()) > rreqPackage->getDestinationSeqNum() ?
							rt->getDestSeqNum(rreqPackage->getDestinationAodv()) : rreqPackage->getDestinationSeqNum());
	}

	else if(rt->getNextHop(rreqPackage->getDestinationAodv())){
		//current node has an active route to the destination
		//refer to RFC3561 chapter 6.6.2
		trace() << "current node has an active route to the destination";
		setRreqBoardcastDropTimer(rreqPackage->getDestinationAodv(), rreqPackage->getSourceAodv(), rreqPackage->getRreqID());
		sendRrepPacket(rt->getHopCount(rreqPackage->getDestinationAodv()),
						rreqPackage->getSourceAodv(),
						rreqPackage->getDestinationAodv(),
						rt->getDestSeqNum(rreqPackage->getDestinationAodv()),
						getRemainingLifetimeTime(rreqPackage->getDestinationAodv())-simTime().dbl(),
						false);
	} else {
		//current node is the destination
		//refer to RFC3561 chapter 6.6.1
		if(rreqPackage->getDestinationSeqNum()==currSeqNum){
			currSeqNum++;
		}
		setRreqBoardcastDropTimer(rreqPackage->getDestinationAodv(), rreqPackage->getSourceAodv(), rreqPackage->getRreqID());
		//MY_ROUTE_TIMOUT = 2 * ACTIVE_ROUTE_TIMEOUT
		sendRrepPacket(0, rreqPackage->getSourceAodv(), SELF_NETWORK_ADDRESS,
						currSeqNum, 2 * activeRouteTimeout,  false);
	}
}

void AodvRouting::processAodvRrepPackage(AodvRREPPacket * rrepPackage,int srcMacAddress, double rssi, double lqi){
	sendRrepAckPacket(rrepPackage->getSource());
	/* TODO: temporary disabled for debugging*/
	if(getTimer(AODV_HELLO_MESSAGE_REFRESH_TIMER).dbl()<=0){
		sendHelloMessage();
	}


	//updates a route to the previous hop without a valid seq number
	updateRTEntry(rrepPackage->getSource(), 0, false, VALID, 1, rrepPackage->getSource(),NULL);
	//update route for the destination
	updateRTEntry(rrepPackage->getDestinationAodv(), rrepPackage->getDestinationSeqNum(),
				true, VALID, rrepPackage->getHopCount() + 1, rrepPackage->getSource(),
				NULL,rrepPackage->getLifetime());
	if (!(string(rrepPackage->getOriginator()).compare(SELF_NETWORK_ADDRESS)==0)){
		//forward the rrep
		sendRrepPacket(rrepPackage->getHopCount() + 1,
						rrepPackage->getOriginator(),
						rrepPackage->getDestinationAodv(),
						rrepPackage->getDestinationSeqNum(),
						rrepPackage->getLifetime(),
						true);
	} else {
		//routing table already updated for destination
		//send buffered packages for destination
		trace() << "route to " << string(rrepPackage->getDestinationAodv()) << " found.";
		rreqRetryCount[string(rrepPackage->getDestinationAodv())] = 0;
		processPackagesInBuffer(rrepPackage->getDestinationAodv(), false);
		cancelExpireEntry(&rreqRequest,rrepPackage->getDestinationAodv());
	}
}

void AodvRouting::processPackagesInBuffer(const char * dest, bool drop){
	queue< cPacket* > bufferTemp;
	DataPacket* currPkt;
	while (!TXBuffer.empty()) {
		currPkt = dynamic_cast <DataPacket*>(TXBuffer.front());
		if(currPkt && string(dest).compare(currPkt->getDestinationAodv())==0){
			if(!drop && rt->getNextHop(dest)){
				currPkt->setDestination(rt->getNextHop(dest));
				trace() << "send data packet in buffer to next hop: " << string(rt->getNextHop(dest))
						<< " final destination: " << string(dest);
				toMacLayer(currPkt, resolveNetworkAddress(rt->getNextHop(dest)));
			}
		}
		else{
			bufferTemp.push(TXBuffer.front());
		}
		TXBuffer.pop();
		updateLifetimeEntry(dest, activeRouteTimeout);
	}
	while (!bufferTemp.empty()) {
		TXBuffer.push(bufferTemp.front());
		bufferTemp.pop();
	}

}

void AodvRouting::timerFiredCallback(int index){
	switch(index){
	case AODV_RREQ_BOARDCAST_DROP_TIMER:
		rreqBroadcast.pop();
		if(!rreqBroadcast.empty()){
			setTimer(AODV_RREQ_BOARDCAST_DROP_TIMER, rreqBroadcast.top().lifetime - simTime().dbl());
		}
		break;

	case AODV_RREQ_EXPIRATION_TIMER:
	{
		if (rreqRequest.top().canceled){
			trace() << "09A - rreq broadcase timer (cancelled) expired for dest: " << rreqRequest.top().dest << " id: " << rreqRequest.top().rreqID;
			rreqRequest.pop();
			if(!rreqRequest.empty()){
				setTimer(AODV_RREQ_EXPIRATION_TIMER, rreqRequest.top().lifetime - simTime().dbl());
				trace() << "10 - timer set for " << rreqRequest.top().lifetime - simTime().dbl();
			}
			break;
		}
		trace() << "09B - rreq broadcase timer expired for dest: " << rreqRequest.top().dest << " id: " << rreqRequest.top().rreqID;
		string dest1 = string(rreqRequest.top().dest);
		rreqRequest.pop();
		if(!rreqRequest.empty()){
			setTimer(AODV_RREQ_EXPIRATION_TIMER, rreqRequest.top().lifetime - simTime().dbl());
		}
		if(rreqRetryCount[dest1] < rreqRetries){
			//send another rreq
			currRreqID++;
			sendRreqRequest(0, currRreqID, SELF_NETWORK_ADDRESS, dest1.c_str(),
							currSeqNum, rt->getDestSeqNum(dest1.c_str()));
		} else {
			rreqRetryCount[dest1] = 0;
			processPackagesInBuffer(dest1.c_str(),true);
			//TODO: notify the application layer that the link is not available
		}
	}
		break;

	case AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER:
	{
		if(!rtExpireSeq.top().canceled){
			string dest2 = string(rtExpireSeq.top().destination);
			rteluc[dest2]--;
			//trace() << "rtlf03 - lifetime for " << dest2
			//		<< " timed out. count " << rteluc[dest2];
			if (rteluc[dest2]==0){
				if(rt->getState(dest2.c_str())==VALID){
				//	trace() << "rtlf03B - lifetime for " << dest2 << "expired." ;
					rt->updateState(dest2, INVALID);
				}
			}
		}
		rtExpireSeq.pop();
		//it is possible that multiple entry expires at the same time
		while(!rtExpireSeq.empty() && rtExpireSeq.top().lifetime - simTime().dbl()<=0){
			if(!rtExpireSeq.top().canceled){
				string dest3 = string(rtExpireSeq.top().destination);
				rteluc[dest3]--;
				//trace() << "rtlf03 - lifetime for " << dest3
				//		<< " timed out. count " << rteluc[dest3];
				if (rteluc[dest3]==0){
					if(rt->getState(dest3.c_str())==VALID){
						//trace() << "rtlf03B - lifetime for " << dest3 << "expired." ;
						rt->updateState(dest3, INVALID);
					}
				}
			}
			rtExpireSeq.pop();
		}
		if(!rtExpireSeq.empty()){
			setTimer(AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER, rtExpireSeq.top().lifetime - simTime().dbl());
		}
	}
		break;

	case AODV_RREQ_RATE_LIMIT_TIMER:
		if(!rreqBuffer.empty()){
			processrreqBuffer();
		}
		break;

	case AODV_HELLO_MESSAGE_REFRESH_TIMER:
		sendHelloMessage();
		break;

	case AODV_HELLO_MESSAGE_EXPIRE_TIMER:
	{
		string neib = hmExpireSeq.top().destination;
		hmeluc[neib]--;
		if (hmeluc[neib]==0){
			list<string>* affectedDest = new list<string>();
			list<string>* affectedPrecursor = new list<string>();
			rt->reportLinkBroken(neib.c_str(), affectedDest, affectedPrecursor);
			resetLifetimeEntry(neib.c_str());
			sendRerrMessage(affectedDest, affectedPrecursor);
		}
		hmExpireSeq.pop();
		while(!hmExpireSeq.empty() && hmExpireSeq.top().lifetime - simTime().dbl()<=0){
			string dest3 = hmExpireSeq.top().destination;
			hmeluc[dest3]--;
			if (hmeluc[dest3]==0){
				list<string>* affectedDest = new list<string>();
				list<string>* affectedPrecursor = new list<string>();
				rt->reportLinkBroken(dest3.c_str(), affectedDest, affectedPrecursor);
				resetLifetimeEntry(dest3.c_str());
				sendRerrMessage(affectedDest, affectedPrecursor);
			}
			hmExpireSeq.pop();
		}
		if(!hmExpireSeq.empty()){
			setTimer(AODV_HELLO_MESSAGE_EXPIRE_TIMER, hmExpireSeq.top().lifetime - simTime().dbl());
		}
	}
		break;

	case AODV_RERR_RATE_LIMIT_TIMER:
		if(!rerrBuffer.empty()){
			toMacLayer(rerrBuffer.front(), resolveNetworkAddress(rerrBuffer.front()->getDestination()));
			rerrBuffer.pop();
			setTimer(AODV_RREQ_RATE_LIMIT_TIMER, (double)1/rerrRatelimit);
		}
		break;

	case AODV_RREP_ACK_WAIT_TIMER:
		if(!rrepAcked[string(rrepAckExpire.top().destination)]){
			rreqBlacklist.push_front(string(rrepAckExpire.top().destination));
			setRreqBlacklistTimer(rrepAckExpire.top().destination.c_str());
		}
		rrepAckExpire.pop();
		if(!rrepAckExpire.empty()){
			setTimer(AODV_RREP_ACK_WAIT_TIMER, rrepAckExpire.top().lifetime - simTime().dbl());
		}
		break;

	case AODV_RREQ_BLACKLIST_REMOVE_TIMER:
		{
			string node = string(rreqBlacklistTimeout.top().destination);
			rreqBlacklist.remove(node);
			rreqBlacklistTimeout.pop();
			if(!rreqBlacklistTimeout.empty()){
				setTimer(AODV_RREQ_BLACKLIST_REMOVE_TIMER, rreqBlacklistTimeout.top().lifetime - simTime().dbl());
			}
		}
		break;

	default:
		return;

	}
	return;
}

void AodvRouting::sendHelloMessage(){
	AodvHelloMessage* helloMsg = new AodvHelloMessage("AODV hello message packet", NETWORK_LAYER_PACKET);
	helloMsg->setSource(SELF_NETWORK_ADDRESS);
	helloMsg->setDestination(BROADCAST_NETWORK_ADDRESS);
	toMacLayer(helloMsg, BROADCAST_MAC_ADDRESS);
	setTimer(AODV_HELLO_MESSAGE_REFRESH_TIMER, helloInterval);
}

void AodvRouting::processrreqBuffer(){
	string dest=string(rreqBuffer.front()->getDestinationAodv());
	++rreqRetryCount[dest];
	setRreqBoardcastDropTimer(rreqBuffer.front()->getDestinationAodv(),
								rreqBuffer.front()->getSourceAodv(),
								rreqBuffer.front()->getRreqID());
	if(string(rreqBuffer.front()->getSourceAodv()).compare(SELF_NETWORK_ADDRESS)==0){
		trace() << "08B - set the rreq expire timer for dest: " << dest << " id: " << rreqBuffer.front()->getRreqID();
		setRreqExpirationTimer(dest.c_str());
	}
	trace() << "04 - broadcast rreq request for destination: " << string(rreqBuffer.front()->getDestinationAodv())
			<< " sourceAddr: " << string(rreqBuffer.front()->getSourceAodv())
			<< " id: " << rreqBuffer.front()->getRreqID();
	toMacLayer(rreqBuffer.front(), BROADCAST_MAC_ADDRESS);
	rreqBuffer.pop();
	setTimer(AODV_RREQ_RATE_LIMIT_TIMER, (double)1/rreqRatelimit);
}

void AodvRouting::sendRrepPacket(int hopCount, const char* orig, const char* destAodv,
								unsigned long dSeq, double lifetime,  bool forwarding){
	if(!rt->getNextHop(orig)) return;
	AodvRREPPacket *rrepPackage =
					new AodvRREPPacket("AODV routing RREP packet", NETWORK_LAYER_PACKET);
	rrepPackage->setFlagA(true);
	rrepPackage->setFlagR(false);
	rrepPackage->setPrefixSz(0);
	rrepPackage->setHopCount(hopCount);
	rrepPackage->setOriginator(orig);
	rrepPackage->setDestinationAodv(destAodv);
	rrepPackage->setDestinationSeqNum(dSeq);
	rrepPackage->setSource(SELF_NETWORK_ADDRESS);
	rrepPackage->setDestination(rt->getNextHop(orig));
	rrepPackage->setLifetime(lifetime);
	if(string(destAodv).compare(SELF_NETWORK_ADDRESS)!=0){
		rt->updatePrecursor(destAodv, rt->getNextHop(orig));
	}
	if(forwarding){
		updateLifetimeEntry(orig, activeRouteTimeout);
	}
	trace() << "05 - send rrep package " << (void*)rrepPackage << " to (next hop) " << resolveNetworkAddress(rt->getNextHop(orig)) << " / " << string(rt->getNextHop(orig))
			<< " originator: " << string(orig) << " / " << string(rrepPackage->getOriginator())
			<< " destination " << string(destAodv) << " / " << string(rrepPackage->getDestinationAodv())
			<< " source: " << string(SELF_NETWORK_ADDRESS);
	toMacLayer(rrepPackage, resolveNetworkAddress(rt->getNextHop(orig)));
	setRrepAckTimer(rt->getNextHop(orig));
}

void AodvRouting::processHelloMsg(AodvHelloMessage* msg){
	trace() << "14 - received hello msg from " << string(msg->getSource());
	RoutingTableEntryExpireTimerSet hmeet;
	hmeluc[string(msg->getSource())]++;
	hmeet.destination = string(msg->getSource());
	hmeet.lifetime = allowedHelloLoss * helloInterval + simTime().dbl();
	hmExpireSeq.push(hmeet);
	if(getTimer(AODV_HELLO_MESSAGE_EXPIRE_TIMER).dbl()<=0){
		setTimer(AODV_HELLO_MESSAGE_EXPIRE_TIMER, hmExpireSeq.top().lifetime - simTime().dbl());
	}
}

void AodvRouting::sendRerrMessage(list<string>* affDest, list<string>* affPrecur){
	if(affDest->empty()||affPrecur->empty()){
		return;
	}
	for (list<string>::const_iterator i=affPrecur->begin();i!=affPrecur->end();++i){
		sendSingleRerrMessage(affDest, affPrecur, (*i).c_str());
	}
}

void AodvRouting::sendSingleRerrMessage(list<string>* affDest, list<string>* affPrecur, const char* dest){
	int destCnt = affDest->size();
	AodvRERRPacket *rerrPackage =
						new AodvRERRPacket("AODV routing RERR packet", NETWORK_LAYER_PACKET);
	rerrPackage->setSource(SELF_NETWORK_ADDRESS);
	rerrPackage->setFlagN(true);
	rerrPackage->setDestCount(destCnt);
	rerrPackage->setUnreachableDestAddrArraySize(destCnt);
	rerrPackage->setUnreachableDestSeqNumArraySize(destCnt);
	rerrPackage->setDestination(dest);
	for(int i=0;i<destCnt;i++){
		rerrPackage->setUnreachableDestAddr(i,affDest->front().c_str());
		rerrPackage->setUnreachableDestSeqNum(i,rt->getDestSeqNum(affDest->front().c_str()));
		affDest->pop_front();
	}
	if (getTimer(AODV_RERR_RATE_LIMIT_TIMER).dbl() <= 0){
		trace() << "13 - send rerr packet to " << string(dest) << " notifying link broken to node: ";
		for(int i=0;i<destCnt;i++){
				trace() << "addr: " << rerrPackage->getUnreachableDestAddr(i) << " SeqNum: " << rerrPackage->getUnreachableDestSeqNum(i);
		}
		toMacLayer(rerrPackage, resolveNetworkAddress(dest));
		setTimer(AODV_RREQ_RATE_LIMIT_TIMER, (double)1/rerrRatelimit);
	} else {
		rerrBuffer.push(rerrPackage);
	}
}

void AodvRouting::processAodvRerrPackage(AodvRERRPacket* rerrPacket,int srcMacAddress, double rssi, double lqi){
//	if(string(rerrPacket->getDestination()).compare("99")==0 && string("99").compare(SELF_NETWORK_ADDRESS)==0){
//		trace() << "spec03: received rerr packet from " << string(rerrPacket->getSource())
//				<< " destCnt: " << rerrPacket->getDestCount();
//		for(int i = 0; i < (int)rerrPacket->getUnreachableDestAddrArraySize() ; i++){
//			trace() << "reported link broken for node " << rerrPacket->getUnreachableDestAddr(i);
//		}
//	}
	if(string(rerrPacket->getDestination()).compare(SELF_NETWORK_ADDRESS)!=0){
		return;
	}
	unsigned int affDestCount = (unsigned int)rerrPacket->getDestCount();
	if (affDestCount!=rerrPacket->getUnreachableDestAddrArraySize() ||
		affDestCount!=rerrPacket->getUnreachableDestSeqNumArraySize()){
		//broken rerr packet
		return;
	}
	list<string>* affDest = new list<string>();
	list<string>* affPre = new list<string>();
	for (unsigned int i = 0; i < affDestCount; i++){
		affDest->push_front(string(rerrPacket->getUnreachableDestAddr(i)));
		rt->updateSeqNum(rerrPacket->getUnreachableDestAddr(i),rerrPacket->getUnreachableDestSeqNum(i));
	}
	rt->forwardLinkBroken(rerrPacket->getSource(),affDest,affPre);
	sendRerrMessage(affDest,affPre);
}

void AodvRouting::sendRrepAckPacket(const char* neib){
	AodvRREPAckPacket *rrepaPacket =
						new AodvRREPAckPacket("AODV routing RREP ACK packet", NETWORK_LAYER_PACKET);
	rrepaPacket->setSource(SELF_NETWORK_ADDRESS);
	rrepaPacket->setDestination(neib);
	toMacLayer(rrepaPacket, resolveNetworkAddress(neib));
}
