/*
 * Module:   AODV Routing Protocol for Castalia Simulator
 * Author:   Xu Zongque 
 * Version:  0.1
 * Contact:  Paolo Roberto Grassi <grassi@elet.polimi.it>
 * Project:  Multi Sensor and Actuator Networks
 * Owner:    Politecnico di Milano (MI) - ITALY
 *
 *                _/      _/              _/_/_/    _/_/    
 *               _/_/  _/_/  _/    _/  _/        _/    _/   
 *              _/  _/  _/  _/    _/    _/_/    _/_/_/_/    
 *             _/      _/  _/    _/        _/  _/    _/     
 *            _/      _/    _/_/_/  _/_/_/    _/    _/      
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef _AODVROUTING_H_
#define _AODVROUTING_H_

#include <list>
#include <map>
#include <queue>
#include "VirtualRouting.h"
#include "AodvRoutingPacket_m.h"
#include "AodvRoutingRrepPacket_m.h"
#include "AodvRoutingDataPacket_m.h"
#include "RoutingTable_rt.h"

using namespace std;

enum TimerIndex{
	AODV_RREQ_BOARDCAST_DROP_TIMER = 0,
	AODV_RREQ_EXPIRATION_TIMER = 1,
	AODV_ROUTING_TABLE_ENTRY_EXPIRATION_TIMER = 2,
	AODV_RREQ_RATE_LIMIT_TIMER = 3,
	AODV_HELLO_MESSAGE_REFRESH_TIMER = 4,
	AODV_HELLO_MESSAGE_EXPIRE_TIMER = 5,
	AODV_RERR_RATE_LIMIT_TIMER = 6,
	AODV_RREP_ACK_WAIT_TIMER = 7,
	AODV_RREQ_BLACKLIST_REMOVE_TIMER = 8,
};

enum PacketTypes {
	AODV_UNKNOWN_PACKET_TYPE = 0,
	AODV_RREQ_PACKET = 1,
	AODV_RREP_PACKET = 2,
	AODV_RERR_PACKET = 3,
	AODV_RREP_ACK_PACKET = 4,
	AODV_HELLO_MESSAGE_PACKET = 5,
	AODV_DATA_PACKET = 6,

};

struct RoutingTableEntryExpireTimerSet{
	double lifetime;
	string destination;
	bool canceled;
};

struct RoutingTableEntryLifetimeCompare{
	bool operator()(const RoutingTableEntryExpireTimerSet &a,
					const RoutingTableEntryExpireTimerSet &b){
		return a.lifetime > b.lifetime ;
	}
};

struct RreqExpireTimerSet{
	bool canceled;
	int rreqID;
	string originator;
	string dest;
	double lifetime;
};

struct RreqExpireLifetimeCompare{
	bool operator()(const RreqExpireTimerSet &a,
					const RreqExpireTimerSet &b){
		return a.lifetime > b.lifetime ;
	}
};

class AodvRouting: public VirtualRouting {

 private:
	unsigned long currSeqNum; //sequence number of current node
	RoutingTable* rt; //The routing table;
	int currRreqID; //next RreqID to be used in the RREQ packet

	map<string, int> rreqRetryCount;

	priority_queue <RoutingTableEntryExpireTimerSet,
					vector<RoutingTableEntryExpireTimerSet>,
					RoutingTableEntryLifetimeCompare > rtExpireSeq;
	//routing table entry lifetime update count
	map<string, int> rteluc;

	priority_queue <RreqExpireTimerSet,
					vector<RreqExpireTimerSet>,
					RreqExpireLifetimeCompare> rreqBroadcast;
	priority_queue <RreqExpireTimerSet,
					vector<RreqExpireTimerSet>,
					RreqExpireLifetimeCompare> rreqRequest;
	queue <AodvRREQPacket* > rreqBuffer;
	queue <AodvRERRPacket* > rerrBuffer;

	//for hello message
	priority_queue <RoutingTableEntryExpireTimerSet,
					vector<RoutingTableEntryExpireTimerSet>,
					RoutingTableEntryLifetimeCompare > hmExpireSeq;
	map<string, int> hmeluc;

	//for unidirectional route discovery
	map<string, bool> rrepAcked;
	list<string> rreqBlacklist;
	priority_queue <RoutingTableEntryExpireTimerSet,
					vector<RoutingTableEntryExpireTimerSet>,
					RoutingTableEntryLifetimeCompare > rrepAckExpire;
	priority_queue <RoutingTableEntryExpireTimerSet,
					vector<RoutingTableEntryExpireTimerSet>,
					RoutingTableEntryLifetimeCompare > rreqBlacklistTimeout;

	
	//parametres
	double activeRouteTimeout; //in s
	int allowedHelloLoss;  
	double helloInterval; //in s
	int localAddTTL;
	int netDiameter;
	double nodeTraversalTime; //in s
	int rerrRatelimit;
	int rreqRetries;
	int rreqRatelimit;
	int timeoutBuffer;
	int ttlStart;
	int ttlIncrement;
	int ttlThreshould;

	void setRreqBoardcastDropTimer(const char* dest, const char* source, int id);
	void setRreqExpirationTimer(const char* dest);
	void setRrepAckTimer(const char* neib);
	void setRreqBlacklistTimer(const char* neib);

	//returns true if there is an on going rreq request for the dest
	bool isRreqRequstActive(const char* dest);

	bool isInRreqBlacklist(const char* dest);

	bool checkExpireEntry(const priority_queue <RreqExpireTimerSet, vector<RreqExpireTimerSet>, RreqExpireLifetimeCompare>* ,
							const char* orig, int idx);
	bool checkExpireEntry(const priority_queue <RreqExpireTimerSet, vector<RreqExpireTimerSet>, RreqExpireLifetimeCompare>* ,
							const char* dest);
	void cancelExpireEntry(priority_queue <RreqExpireTimerSet, vector<RreqExpireTimerSet>, RreqExpireLifetimeCompare>* ,
							const char* dest);

	//update the routing table and set the timer
	void updateRTEntry(const char* destinationAddr,unsigned long destinationSeqNum,
						bool isDestinationValid,RoutingFlag state,int hopCount,
						const char* nextHopAddr,std::list<std::string>* precursor);
	void updateRTEntry(const char* destinationAddr,unsigned long destinationSeqNum,
						bool isDestinationValid,RoutingFlag state,int hopCount,
						const char* nextHopAddr,std::list<std::string>* precursor,
						double lifetime);
	double getRemainingLifetimeTime(const char * dest);
	//update a lifetime of a VALID routing table entry (to no less then simTime() + time)
	void updateLifetimeEntry(const char * dest, double time);
	void resetLifetimeEntry(const char * dest);
	

 protected:
	void startup();
	void finish();
	void fromApplicationLayer(cPacket *, const char *);
	void fromMacLayer(cPacket *, int, double, double);
	void processAodvRreqPackage(AodvRREQPacket *,int srcMacAddress, double rssi, double lqi);
	void processAodvRrepPackage(AodvRREPPacket *,int srcMacAddress, double rssi, double lqi);
	void processAodvRerrPackage(AodvRERRPacket *,int srcMacAddress, double rssi, double lqi);
	void processDataPackage(DataPacket*);
	void processHelloMsg(AodvHelloMessage*);
	void processPackagesInBuffer(const char *, bool drop);
	void processrreqBuffer();
	void timerFiredCallback(int index);
	void sendRreqRequest(int hopCount, int rreqID, const char* sourceAodv, const char* destAodv,
						unsigned long sSeq, unsigned long dSeq);
	void sendRerrMessage(list<string>* affDest, list<string>* affPrecur);
	void sendSingleRerrMessage(list<string>* affDest, list<string>* affPrecur, const char* dest);
	void sendHelloMessage();
	void sendRrepPacket(int hopCount, const char* orig, const char* destAodv, unsigned long dSeq, double lifetime, bool forwarding);
	void sendRrepAckPacket(const char* neib);
};

#endif
