/*
 * Module:   AODV Routing Protocol for Castalia Simulator
 * Author:   Xu Zongque 
 * Version:  0.1
 * Contact:  Paolo Roberto Grassi <grassi@elet.polimi.it>
 * Project:  Multi Sensor and Actuator Networks
 * Owner:    Politecnico di Milano (MI) - ITALY
 *
 *                _/      _/              _/_/_/    _/_/    
 *               _/_/  _/_/  _/    _/  _/        _/    _/   
 *              _/  _/  _/  _/    _/    _/_/    _/_/_/_/    
 *             _/      _/  _/    _/        _/  _/    _/     
 *            _/      _/    _/_/_/  _/_/_/    _/    _/      
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ROUTINGTABLE_RT_H
#define ROUTINGTABLE_RT_H

#endif // ROUTINGTABLE_H
#include <string>
#include <list>
#include "CastaliaModule.h"

enum RoutingFlag {
    NOT_DEFINED = 0,
    VALID = 1,
    INVALID = 2,
    REPAIRABLE = 3,
    BEING_REPAIRED = 4,
};

//expiration of enties managed by timers
typedef struct _ROW {
        std::string destinationAddr;
        unsigned long destinationSeqNum;
        bool isDestinationValid;
        RoutingFlag state;
        int hopCount;
        std::string nextHopAddr;
        std::list<std::string>* precursor;

}Entry;

class RoutingTable{
private:
    std::list<Entry>* table;
    Entry* searchByDest(const char* destination);
    Entry* searchByDest(std::string destination);

public:
    RoutingTable();
    ~RoutingTable();
    /* the function will first search an entry row by destination address in the routing
     * table. If it finds one, the function updates the entry with other parameters,
     * otherwise it adds a new entry to the table.
     */
    void updateEntry(Entry& e);
    void updateEntry(const std::string& destinationAddr,unsigned long destinationSeqNum,
                     bool isDestinationValid,RoutingFlag state,int hopCount,
                     const std::string& nextHopAddr,std::list<std::string>* precursor);
    void updateEntry(const char* destinationAddr,unsigned long destinationSeqNum,
                     bool isDestinationValid,RoutingFlag state,int hopCount,
                     const char* nextHopAddr,std::list<std::string>* precursor);

    /* the function will first search an entry row by destination address in the routing
     * table. If it finds one, the function updates the state field in the entry,
     * otherwise it does nothing.
     */
    void updateState(const std::string& destAddr, RoutingFlag st);
    void updateState(const char* destAddr, RoutingFlag st);

    /* the function will first search an entry row by destination address in the routing
     * table. If it finds one, the function return the nexthop address recorded in the entry,
     * otherwise it returns NULL.
     */
    const char* getNextHop(const char* destination);

    /* the function will first search an entry row by destination address in the routing
     * table. If it finds one, the function return the destination sequence number recorded
     * in the entry, otherwise it returns 0.
     */
    unsigned long getDestSeqNum(const char* destination);

    bool isDestSeqNumValid(const char* destination);
    RoutingFlag getState(const char* destination);
    int getHopCount(const char* destination);
    const std::list<std::string>& getPrecursors(const char* destination);
    void updatePrecursor(const char* destination, const char* precursor);
    void deletePrecursor(const char* destination, const char* precursor);
    void updateSeqNum(const char* destination, unsigned long seq);

    bool isRouteValid(const char* destination);

    void reportLinkBroken(const char* neib, std::list<std::string>* affectedDest, std::list<std::string>* affectedPrecursor);
    void forwardLinkBroken(const char* src, const std::list<std::string>* affectedDest, std::list<std::string>* affectedPrecursor);
};


