// C. Pham, Congduc.Pham@univ-pau.fr, www.univ-pau.fr/~cpham 

#include <math.h>
#include <omnetpp.h>
#include "coordNode.h"
#include "videoSensorNode.h"

// The module class needs to be registered with OMNeT++
Define_Module(coordNode);

#ifdef DISPLAY_SENSOR
SDL_Surface *screen = NULL;  
SDL_Event event;
Uint32 currentSDLColor;
#endif

Point POSITION_INTERNAL_OFFSET;

void draw_point (int x, int y, int radius)
{
// just to avoid segmentation fault if graphical display is not used
#ifdef DISPLAY_SENSOR
    Uint32 *pix;  
    pix=(Uint32*)screen->pixels+(x+y*GRAPHICS_WIDTH);
	*pix=currentSDLColor;
#endif
}  

//Bresenham line from (x1,y1) to (x2,y2) with rgb color
bool draw_line(int x1, int y1, int x2, int y2)
{
#ifdef DISPLAY_SENSOR
  int w = GRAPHICS_WIDTH;
  int h = GRAPHICS_HEIGHT;

  if(x1 < 0 || x1 > w - 1 || x2 < 0 || x2 > w - 1 || y1 < 0 || y1 > h - 1 || y2 < 0 || y2 > h - 1) return 0;
  
  int deltax = abs(x2 - x1); //The difference between the x's
  int deltay = abs(y2 - y1); //The difference between the y's
  int x = x1; //Start x off at the first pixel
  int y = y1; //Start y off at the first pixel
  int xinc1, xinc2, yinc1, yinc2, den, num, numadd, numpixels, curpixel;

  if(x2 >= x1) //The x-values are increasing
  {
    xinc1 = 1;
    xinc2 = 1;
  }
  else //The x-values are decreasing
  {
    xinc1 = -1;
    xinc2 = -1;
  }
  if(y2 >= y1) //The y-values are increasing
  {
    yinc1 = 1;
    yinc2 = 1;
  }
  else //The y-values are decreasing
  {
    yinc1 = -1;
    yinc2 = -1;
  }
  if (deltax >= deltay) //There is at least one x-value for every y-value
  {
    xinc1 = 0; //Don't change the x when numerator >= denominator
    yinc2 = 0; //Don't change the y for every iteration
    den = deltax;
    num = deltax / 2;
    numadd = deltay;
    numpixels = deltax; //There are more x-values than y-values
  }
  else //There is at least one y-value for every x-value
  {
    xinc2 = 0; //Don't change the x for every iteration
    yinc1 = 0; //Don't change the y when numerator >= denominator
    den = deltay;
    num = deltay / 2;
    numadd = deltax;
    numpixels = deltay; //There are more y-values than x-values
  }
  for (curpixel = 0; curpixel <= numpixels; curpixel++)
  {
	draw_point(x % w, y % h, 0);  //Draw the current pixel
    num += numadd;  //Increase the numerator by the top of the fraction
    if (num >= den) //Check if numerator >= denominator
    {
      num -= den; //Calculate the new numerator value
      x += xinc1; //Change the x as appropriate
      y += yinc1; //Change the y as appropriate
    }
    x += xinc2; //Change the x as appropriate
    y += yinc2; //Change the y as appropriate
  }
 #endif
  return 1;
}

void obstacleManager::init(int n, int maxX, int maxY)
{
	Point left, right;
	Segment s_a;

	this->nbObstacle=n;

	for (int i=0; i<n; i++) {
		// get starting point
		left.x=uniform(0, maxX-MAX_OBSTACLE_SIZE);
		left.y=uniform(MAX_OBSTACLE_SIZE, maxY-MAX_OBSTACLE_SIZE);

		left = left + POSITION_INTERNAL_OFFSET;

		double theta=uniform(0,1)*PI/2;

		// give a down orientation in 50% of cases
		if (uniform(0,1)>0.5)
			theta = theta + PI;

		// find the ending point
		double l=uniform(MIN_OBSTACLE_SIZE, MAX_OBSTACLE_SIZE);
		right.x=left.x+l*cos(theta);
		right.y=left.y+l*sin(theta);

		// store the new segment obstacle
		obstacles.push_back(Segment(left, right));

#ifdef DISPLAY_SENSOR
		left = left*DISPLAY_SCALING_FACTOR;
		right = right*DISPLAY_SCALING_FACTOR;

		s_a=Segment(left, right);
		// store the new segment obstacle
		obstacles_display.push_back(s_a);
		// draw the obstacle
		if (screen!=NULL)
			currentSDLColor=SDL_MapRGB(screen->format, 255, 0, 0); // in red
		if( SDL_MUSTLOCK( screen ) )
			SDL_LockSurface( screen );
		s_a.draw();
		if( SDL_MUSTLOCK( screen ) )
			SDL_UnlockSurface( screen );
		SDL_Flip(screen);
		if (screen!=NULL)
			currentSDLColor=SDL_MapRGB(screen->format, 0, 0, 0); // in black
#endif
	}
#ifdef DISPLAY_SENSOR
	cout << "Press key to resume simulation with sensor node's FoV" << endl ;
	char c=getchar();
#endif
}

bool obstacleManager::intersects(Segment s)
{
#ifdef WITH_OBSTACLES
	std::list<Segment>::iterator k;

	for (k = obstacles.begin(); k != obstacles.end(); k++) {
		if (k->intersects(s)) return true;
	}
	return false;
#else
	return false;
#endif
}

void obstacleManager::draw()
{
#ifdef DISPLAY_SENSOR
	std::list<Segment>::iterator k;
	if (screen!=NULL) {
		currentSDLColor=SDL_MapRGB(screen->format, 255, 0, 0); // in red
		if( SDL_MUSTLOCK( screen ) )
			SDL_LockSurface( screen );
		for (k = obstacles_display.begin(); k != obstacles_display.end(); k++)
			k->draw();
		if( SDL_MUSTLOCK( screen ) )
			SDL_UnlockSurface( screen );
		SDL_Flip(screen);
		currentSDLColor=SDL_MapRGB(screen->format, 0, 0, 0); // in black
	}
#endif
}

void coordNode::initialize()
{
	// get the number of sensor node in the network, defined in the .ned file
	nbrNode = par("nodesNb");
	fieldMaxX = par("fieldMaxX"); // the read value already has the scaling factor
	fieldMaxY = par("fieldMaxY"); // the read value already has the scaling factor
	allNodes = new noeud[nbrNode];
	nbRegisteredSensorNodeCounter = 0;

	// just in case because normally all nodes will be registered
	for (int i=0; i<nbrNode; i++)
			this->allNodes[i].id=-1;

	POSITION_INTERNAL_OFFSET=Point(MAX_DOV_INTERNAL,MAX_DOV_INTERNAL);

	// create a message for self-notification of simulation increment
        this->displaystatsEv = new cMessage("displaystats");
#ifdef CASTALIA
	// need more time before displaying stat and counting the number the active nodes.
        scheduleAt(simTime()+DISPLAY_STATS_INTERVAL*3, this->displaystatsEv);
#else
        scheduleAt(simTime()+DISPLAY_STATS_INTERVAL, this->displaystatsEv);
#endif
	// the random generator seed is initialized in the omnetpp.ini file
        // get the value from the omnetpp.ini file
	double mySeed=strtod(ev.getConfig()->getConfigValue("seed-0-mt"), NULL);
	// put in the the .sca file to differentiate between runs
	recordScalar("info.sim.initial_seed", mySeed);
	recordScalar("info.topology.nbNodes", nbrNode);

	percentageCoverageStatVector.setName("percentageCoverage");
	percentageActiveNodeStatVector.setName("percentageActiveNode");
	stddevEnergyLevelStatVector.setName("meanEnergyLevel");

	meanPercentageCoversetCoverage.setName("meanPercentageCoversetCoverage");
	meanPercentageCoversetCoverage.setRangeAutoUpper(0, 10, 1.5);
	meanNumberCoverset.setName("meanNumberCoverset");
	meanNumberCoverset.setRangeAutoUpper(0, 10, 1.5);
	meanCaptureRate.setName("meanCaptureRate");
	meanCaptureRate.setRangeAutoUpper(0, 10, 1.5);
	meanEnergyLevel.setName("meanEnergyLevel");
	meanEnergyLevel.setRangeAutoUpper(0, 10, 1.5);

#ifdef DISPLAY_SENSOR
	if (SDL_Init(SDL_INIT_VIDEO)==0) {
   		screen = SDL_SetVideoMode(GRAPHICS_WIDTH, GRAPHICS_HEIGHT, BITDEPTH, SDL_SWSURFACE);
	
		if (screen != NULL) {
			currentSDLColor=SDL_MapRGB(screen->format, 0, 0, 0); // black
			SDL_FillRect(screen, NULL, 0xFFFFFFFF);
			Point P1;
			Polygon theField;

			// draw the field with a rectangle
			P1 = Point(MAX_DOV*METER_TO_DISPLAY_FACTOR,MAX_DOV*METER_TO_DISPLAY_FACTOR);
			theField.push_back(P1);
			P1 = Point((MAX_DOV+MAXFIELDMETER)*METER_TO_DISPLAY_FACTOR,MAX_DOV*METER_TO_DISPLAY_FACTOR);
			theField.push_back(P1);
			P1 = Point((MAX_DOV+MAXFIELDMETER)*METER_TO_DISPLAY_FACTOR,(MAX_DOV+MAXFIELDMETER)*METER_TO_DISPLAY_FACTOR);
			theField.push_back(P1);
			P1 = Point(MAX_DOV*METER_TO_DISPLAY_FACTOR,(MAX_DOV+MAXFIELDMETER)*METER_TO_DISPLAY_FACTOR);
			theField.push_back(P1);
			if( SDL_MUSTLOCK( screen ) ) 
				SDL_LockSurface( screen ); 
			theField.draw();
			if( SDL_MUSTLOCK( screen ) ) 
				SDL_UnlockSurface( screen ); 
			SDL_Flip(screen);
		}
	}
#endif

	obstacleList.nbObstacle=0;
#ifdef WITH_OBSTACLES
	// create random obstacle
	obstacleList.init(25, fieldMaxX, fieldMaxY);
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
	// TODO try avoid the statically defined number of nodes
	for (int i=0; i<200; i++) {
		alertNumber[i]=0;
		
		for (int j=0; j<700; j++)
			this->alertResponseArray[i][j]=0;
    }
#endif
}

void coordNode::handleMessage(cMessage *msg)
{
	//
	// periodically display statistics
	//
    if (strcmp("displaystats", msg->getName())== 0) {
		
		int nbActiveNode=0;
		int nbAliveNode=0;
		double percentageActiveNode;

		// meanEnergyLevel.clearResult();

                for(int i=0; i<this->nbrNode; i++) {
                        if (this->allNodes[i].isActive)
                            nbActiveNode++;
                        if (this->allNodes[i].isAlive)
                            nbAliveNode++;
#ifdef CASTALIA
                        meanEnergyLevel.collect(this->allNodes[i].theNode->getCastaliaEnergyLevel());
#else
                        meanEnergyLevel.collect(this->allNodes[i].theNode->getEnergyLevel());
#endif
                }
	       
		percentageActiveNode = (double)nbActiveNode * 100.0 / this->nbrNode;
		this->roundCoverage = this->percentageCoverageRound(this->allNodes, this->nbrNode);

		recordScalar("**display_stats_time", simTime());
		recordScalar("**display_stats_percentage_coverage", this->roundCoverage);
		recordScalar("**display_stats_number_active_nodes", nbActiveNode);
		recordScalar("**display_stats_percentage_active_nodes", percentageActiveNode);

		percentageCoverageStatVector.record(this->roundCoverage);
		percentageActiveNodeStatVector.record(percentageActiveNode);
		stddevEnergyLevelStatVector.record(meanEnergyLevel.getMean());
		//stddevEnergyLevelStatVector.record(meanEnergyLevel.getStddev());

#ifdef DEBUG_OUTPUT_LEVEL0
                ev << "**display-stats** " << simTime() << "\t pert.coverage " <<
        	this->roundCoverage << "% \t nb active nodes " << percentageActiveNode << "%" << endl;
#endif  

		if (nbAliveNode <= 4) {
			this->nodeFinish(nbAliveNode);
			delete msg;
			msg = NULL;
                }
                else
                        // schedules next display
			scheduleAt(simTime()+DISPLAY_STATS_INTERVAL, this->displaystatsEv);	
    }
}
 
float coordNode::distance(Point M, Point N)
{
  return sqrt((M.x - N.x)*(M.x - N.x) + (M.y - N.y)*(M.y - N.y));
}

// the percentage of coverage when considering all the nodes
double coordNode::percentageCoverage(noeud* N, int nb_nodes)
{
  	Enter_Method("percentageCoverage()");
  	int nbCouvertInitial = 0;
	Point P;
  	double percentage;
  	bool found;

        srand (time(NULL));

#ifdef DISPLAY_SENSOR
#ifdef WITH_OBSTACLES
	// re-draw the obstacles that might have been cleared by the node's FoV drawings
	cout << "Re-drawing obstacles" << endl ;
	obstacleList.draw();
#endif
	cout << "Press key to start initial coverage computation" << endl ;
	char c=getchar();
//	if (screen!=NULL) {
//		if( SDL_MUSTLOCK( screen ) )
//			SDL_LockSurface( screen );
//	}
#endif

        // take random points in the field
	for (int i=0; i<=100000; i++) {
            found=false;
#ifdef DISPLAY_SENSOR
            currentSDLColor=SDL_MapRGB(screen->format, 0, 255, 0); // in green
#endif
            // Define the position of the sensor node
            P.x = uniform(0, this->fieldMaxX);
            P.y = uniform(0, this->fieldMaxY);

            P = P + POSITION_INTERNAL_OFFSET;

            // if at least one sensor node, with multiple camera support, covers the point then increases the number of random point covered
            for (int j=0; j<nb_nodes; j++) {

                    for (int k=0; k<N[j].nbCamera; k++) {
                            if (P.is_inside(N[j].fov[k])) {
#ifdef WITH_OBSTACLES
                                    if (!obstacleList.intersects(Segment(N[j].p,P))) {
#endif
                                            nbCouvertInitial=nbCouvertInitial+1;
                                            found=true;
#ifdef DISPLAY_SENSOR
                                            if (k==0)
                                                    currentSDLColor=SDL_MapRGB(screen->format, 255, 0, 0); // in red
                                            if (k==1)
                                                    currentSDLColor=SDL_MapRGB(screen->format, 0, 0, 255); // in blue
                                            if (k==3)
                                                    currentSDLColor=SDL_MapRGB(screen->format, 0, 255, 255); // in yellow
#endif
                                            break;
#ifdef WITH_OBSTACLES
                                    }
#endif
                            }
                    }

#ifdef DISPLAY_SENSOR
                    if (screen!=NULL && found) {
                            P=P*DISPLAY_SCALING_FACTOR;
                            P.draw();
                    }
#endif

                    if (found)
                            break;
            }
        }

#ifdef DISPLAY_SENSOR
//	if( SDL_MUSTLOCK( screen ) )
//		SDL_UnlockSurface( screen );

	if (screen!=NULL) {
		SDL_Flip(screen);
	}
#endif

	percentage = (double)nbCouvertInitial * 100.0 /100000.0;

	cout << "Percentage coverage " << percentage << endl ;

#ifdef DISPLAY_SENSOR	
//	if( SDL_MUSTLOCK( screen ) )
//		SDL_LockSurface( screen );

	cout << "Re-drawing sensor's FoV" << endl ;

	if (screen!=NULL) {
		currentSDLColor=SDL_MapRGB(screen->format, 0, 0, 0); // in black
		for (int j=0; j<nb_nodes; j++) {
			for (int k=0; k<N[j].nbCamera; k++) {
				N[j].fov_display[k].draw();
			}
		}
	}
	cout << "Re-drawing obstacles" << endl ;
	obstacleList.draw();

//	if( SDL_MUSTLOCK( screen ) )
//		SDL_UnlockSurface( screen );

	SDL_Flip(screen);
/*
        bool nokeypress=true;
        SDL_Event event;

        cout << "Press key to resume simulation" << endl;

        while (nokeypress) {
                SDL_WaitEvent(&event);
                switch(event.type)
                {
                        case SDL_QUIT:
                                nokeypress = false;
                                break;
                        case SDL_KEYDOWN:
                                nokeypress= false;
                                break;
                }
        }
*/
        c=getchar();
        //c=getchar();
        //SDL_Quit();
#endif
    return percentage;
}

void coordNode::fillNodes(Point p, Point v, Point b, Point c, int id, int nbCamera, int cameraIndex, VideoSensorNode* theNode)
{
	this->allNodes[id].id = id;
	this->allNodes[id].theNode = theNode;
	this->allNodes[id].p = p;
        this->allNodes[id].v = v;
	this->allNodes[id].b = b;
	this->allNodes[id].c = c;
	this->allNodes[id].nbCamera = nbCamera;
	this->allNodes[id].fov[cameraIndex]=Triangle(p,b,c);
#ifdef DISPLAY_SENSOR
	this->allNodes[id].fov_display[cameraIndex]=Triangle(p*DISPLAY_SCALING_FACTOR, b*DISPLAY_SCALING_FACTOR, c*DISPLAY_SCALING_FACTOR);
#endif
	this->allNodes[id].isAlive=true;
	this->allNodes[id].isActive = true;
#ifdef SCHEDULE_PROB
	this->allNodes[id].nbWakeup=0;
#endif

        // one more node registered
	// only if we have registered the last camera of each node
	if (cameraIndex==0)
		this->nbRegisteredSensorNodeCounter++;
    
        // have all node registered? if yes, display initial coverage (should be 100% as we considered that
        // initial coverage is always 100%, coverage computer at each round is then relative to the initial coverage
        // the goal here is not the get full coverage of the field
        if (this->nbRegisteredSensorNodeCounter == this->nbrNode) {
            this->initialCoverage = this->percentageCoverage(this->allNodes, this->nbrNode);
        
#ifdef DEBUG_OUTPUT_LEVEL0
            ev << "**Initial_coverage** t=" << simTime() <<  "\t pert.coverage " << this->initialCoverage << endl;
#endif
            recordScalar("stat.coverset.initial_coverage", this->initialCoverage);
#ifdef DISPLAY_SENSOR
            //SDL_Quit();
#endif
        }
}

void coordNode::updateNodesStatus(int id, bool isActive, bool isAlive)
{
	this->allNodes[id].isActive = isActive;
	this->allNodes[id].isAlive=isAlive;
}

noeud* coordNode::getNodeInfo(int theNodeId) {

	if (this->allNodes[theNodeId].id == -1)
		return NULL;
	else
		return (noeud*)&this->allNodes[theNodeId];
}

#ifdef CRITICALITY_ADAPTATIVE_MAC

void coordNode::countAlertNumber(int id) {
	// count the number of alert sent by node id
	this->alertNumber[id]++;
}

void coordNode::countAlertResponse(int id, int alertNb) {
	// count the number of response for that alert, store by node id
	this->alertResponseArray[id][alertNb]++;
}

#endif

#ifdef SCHEDULE_PROB

bool coordNode::probTestNode(unsigned int id, int senderId)
{
#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "coordNode " << "<-- prob test node " << id << " by node " << senderId << endl;
#endif
	if (this->allNodes[id].isActive)
		return false;
	if (this->allNodes[id].isAlive==false)
		return false;
	return true;
}

bool coordNode::probWakeNode(unsigned int id, int senderId)
{
#ifdef DEBUG_OUTPUT_LEVEL1
	ev << "coordNode " << "<-- prob wakes node " << id << " by node " << senderId << endl;
#endif
	// node is already active
	if (this->allNodes[id].isActive) {
		this->allNodes[id].nbWakeup++;
#ifdef DEBUG_OUTPUT_LEVEL1
		ev << "coordNode " << "<-- node " << id << " has already " << this->allNodes[id].theNode->prob_nbActivityRound++ <<
			" pending rounds" << endl;
#endif
#ifdef SCHEDULE_PROB_CHAOS
		// increases the number of iterations the node has to do before going to sleep
		this->allNodes[id].theNode->chaos_nbIteration++;
#endif
		// increases the number of rounds it has to be awake
		this->allNodes[id].theNode->prob_nbActivityRound++;

		// indicates that the sender should wake up another node also to maintain a constant proportion of
		// active nodes. Maybe hard to do in real world as it needs more signaling.
		return false;
	}

	if (this->allNodes[id].isAlive==false) {
#ifdef DEBUG_OUTPUT_LEVEL1
		ev << "coordNode " << "<-- node " << id << " is dead." << endl;
#endif
		// if we return true, then a dead node is selected, which will decrease the number of active nodes
		// however, the side effect is that more nodes will be in sleep mode, without being waken up
		return false;
		// if we return false, then the sender node keeps trying to find a next node to wake up.
		// many iterations may be needed when the number of dead nodes is high.
	}

	// call the node's function to wake up
	this->allNodes[id].theNode->probWakeNode();
	this->allNodes[id].nbWakeup++;
	return true;
}
#endif

// the percentage of coverage when considering only active nodes
double coordNode::percentageCoverageRound(noeud* N, int nb_nodes){
    Enter_Method("percentageCoverageRound()");
    double nbCouvertInitial = 0;
    Point P;
    double percentage;
    bool found;

    srand (time(NULL));
    
    for (int i=0; i<=50000; i++){
        found=false;
		P.x=uniform(0, this->fieldMaxX);
		P.y=uniform(0, this->fieldMaxY);

		P=P+POSITION_INTERNAL_OFFSET;

    	// if at least one sensor node covers the point then increases the number of random point covered
    	for (int j=0; j<nb_nodes && !found; j++) {
			for (int k=0; k<N[j].nbCamera && !found; k++) {
				if (N[j].isActive && P.is_inside(N[j].fov[k])) {
#ifdef WITH_OBSTACLES
					if (!obstacleList.intersects(Segment(N[j].p,P))) {
#endif
						nbCouvertInitial=nbCouvertInitial+1;
						found=true;
#ifdef WITH_OBSTACLES
					}
#endif
				}
			}
		}
    }

    percentage = (double)nbCouvertInitial * 100 /50000;
    // compared to the initial coverage when all nodes are considered
    percentage = percentage * 100 / this->initialCoverage;
    
    return percentage;
}


void coordNode::nodeFinish(int nbAliveNode)
{
#ifdef DEBUG_OUTPUT_LEVEL0
	ev << "**coord node is terminating** " << simTime() << "\t nb alive nodes " << nbAliveNode << endl;
	ev << "meanPercentageCoversetCoverage " << meanPercentageCoversetCoverage.getMean() << endl;

#endif
	// write the statistics in the .sca file
	recordScalar("info.sim.end_time", simTime());
	meanPercentageCoversetCoverage.recordAs("meanPercentageCoversetCoverage");
	meanNumberCoverset.recordAs("meanNumberCoverset");
	meanCaptureRate.recordAs("meanCaptureRate");
#ifdef SCHEDULE_PROB
	char s[20];
	for(int i=0; i<this->nbrNode; i++){
		sprintf(s, "info.schedule.chaos.node_%d_nb_wake_up", i);
		recordScalar(s, this->allNodes[i].nbWakeup);
	}
#endif

	endSimulation();
}

void coordNode::finish()
{
	cancelAndDelete(displaystatsEv);	
	delete[] allNodes;
	
#ifdef CRITICALITY_ADAPTATIVE_MAC

        for (int i=0; i<this->nbrNode; i++) {

                int responseDistribution[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
                int maxNumberOfResponse=0;

                char s[120];

                recordScalar("********************************************************* Node", i);

                sprintf(s, "Node %d --------------- nbAlert", i);

                recordScalar((const char*)s, this->alertNumber[i]);


                for (int j=1; j<=this->alertNumber[i]; j++) {

                                    sprintf(s, "Node %d - Alert %d ----- alertResponse", i, j);

                                    recordScalar((const char*)s, this->alertResponseArray[i][j]);

                                    if (this->alertResponseArray[i][j]>maxNumberOfResponse)
                                            maxNumberOfResponse = this->alertResponseArray[i][j];

                                    responseDistribution[this->alertResponseArray[i][j]]++;
                }

                for (int k=0; k<=maxNumberOfResponse; k++) {
                        sprintf(s,"Node %d ---------alertResponseCount_%d", i, k);
                        recordScalar((const char*)s, responseDistribution[k]);
                }



                for (int k=0; k<=maxNumberOfResponse; k++) {
                        sprintf(s,"Node %d --alertResponseDistribution_%d ", i, k);
                        recordScalar((const char*)s, (double)responseDistribution[k]*100.0/(double)this->alertNumber[i]);
                }
        }
#endif	
}


