// C. Pham, Congduc.Pham@univ-pau.fr, www.univ-pau.fr/~cpham

/* SOME REMARKS ON THE DEBUGGING OUTPUT OF THE SIMULATION MODEL

The code uses 4 levels of output for debugging purposes. These levels are defined by #ifdef statements which may not be the best solution for readability but is quite efficient. There may be more elegant way to do it but I haven't had time to redesign the entire process. These levels are: DEBUG_OUTPUT_LEVEL0, DEBUG_OUTPUT_LEVEL1, DEBUG_OUTPUT_LEVEL2 and DEBUG_OUTPUT_LEVEL3. I usually use level 0 for most common output,typically those that are needed quite often to know what the simulation is doing. Then as the level increases, the output are more and more optional and will mostly be triggered in case of problems.

These outputs write to the OMNET++'s standard ev stream (which is the text part of the OMNET++'s tkenv graphical interface). When the cmd env is used there will be no output.

Castalia provides the trace() fonctionnality that writes in a trace file. In this case, we use a tee class that allows simultaneous outputs in both the ev stream and the trace file. In order to have this feature, outputs should use the ev_v stream. The following table summarizes how outputs are currently managed:

OMNET++: ev_v is simply the regular ev stream
Castalia: ev_v is both ev and trace file

I usually use ev_v for level 0 of debug output so that under Castalia I have both the standard output, when available, and the trace file. The other debug output level still use the regular ev stream in order to not make the trace file too big difficult to read. In intense debugging phase, you could turn some ev into ev_v.
*/

#include <algorithm>
#include <cassert>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <cstdlib>
#include <limits>
#include <stdlib.h>
#include <time.h>
#include <vector>
#include <set>
#include <list>

#include <omnetpp.h>

#include "coordNode.h"
#include "videoSensorNode.h"
#include "activity_m.h"
#include "position_m.h"
#include "dead_m.h"

#ifdef CASTALIA
#include "idreport_m.h"
#include "image_m.h"
#ifdef SEND_REAL_IMAGE
#include "handleImage.c"
#endif
#endif

#ifdef WITH_INTRUSION
#include "intrusion.h"
#endif

#ifdef FUNCTION_TRACE_DEBUG
#include "mydebug.h"
#endif

// The module class needs to be registered with OMNeT++
Define_Module(VideoSensorNode);

#ifdef DISPLAY_SENSOR
extern SDL_Surface *screen;  
extern SDL_Event event;
#endif

extern Point POSITION_INTERNAL_OFFSET;

// half the angle of the field of view, about 36° in total
#ifdef COVERAGE_MIXED_ANGLEVIEW
static double defaultAlpha = PI/10;
#else
#ifdef SENSOR_XS_ANGLEVIEW
// PI/18 gives 2.alpha=20°
static double defaultAlpha = PI/18;
#else
// PI/10 gives 2.alpha=36° per default
static double defaultAlpha = PI/10;
#endif
#endif

// For set and list manipulation

void printSet(const SetInt& theSet) {

	SetInt::iterator i;
	ev << "SetSize = " << theSet.size() << "\t(";
	for (i=theSet.begin(); i!=theSet.end(); ) { 
		ev << *i ;
		i++;
		if (i!=theSet.end()) ev << ", ";
	}	
	ev << ")" << endl;
}

void printSetSet(const SetSetInt& theSetSet) {

	SetSetInt::iterator i;
	ev << "SetSetSize = " << theSetSet.size() << "\n[[ \n";
	for (i=theSetSet.begin(); i!=theSetSet.end(); i++) {
        printSet (*i);
    }
	ev << "]]" << endl;
}

void printList(ListInt& theList) {

	ListInt::iterator i;
	ev << "ListSize = " << theList.size() << "\t(";
	for (i=theList.begin(); i!=theList.end(); ) { 
		ev << *i ;
		i++;
		if (i!=theList.end()) ev << ", ";
	}	
	ev << ")" << endl;
}

void printListList(ListListInt& theListList) {

	ListListInt::iterator i;
	ev << "ListListSize = " << theListList.size() << "\n[[ \n";
	for (i=theListList.begin(); i!=theListList.end(); i++) {
        printList (*i);
    }
	ev << "]]" << endl;
}

bool isIncluded(std::list<int> small,std::list<int> great) {
    std::list<int>::iterator i;
    std::list<int>::iterator j;
    j=great.begin();
    i=small.begin();

    while(i!=small.end()){
        j=std::find(great.begin(),great.end(),(*i));
        if(j==great.end())
        return false;
        i++;
    }
    return true;
}

// For managing chaotic scheduling
#ifdef SCHEDULE_PROB_CHAOS
unsigned int prng_xor(unsigned int xq) {

	xq ^= (xq << 13);
	xq ^= (xq >> 17);
	xq ^= (xq << 5);
	return xq;
}

double CC(int n, int r) {
	double  Anr = 1;
	double  Arr = 1;

	for(; r > 0; Anr *= n--, Arr *= r--);

	return Anr / Arr;
}

struct chaos_x ci_ver2(struct chaos_x in, unsigned int NQ) {
	double a,cq[NQ+2];
	int i1,i,j,n,w,MQ;

	a=4294967296.0/(pow(2,NQ));
	cq[0]=0;

	for (i=0;i<NQ+1;i++)  {
		cq[i+1]=cq[i]+CC(NQ,i)*a;
	}

	//MQ=32/NQ;
	in.out=0;

	in.xq=prng_xor(in.xq);
	//printf("xq=%u,",in.xq);
	for (i1=0;i1<NQ+1;i1++) {
		if (in.xq>=cq[i1]&&in.xq<cq[i1+1]) break;
	}

	n=i1;
	//printf("%u",i1);

	int aq[32]={1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};

	for(j=0;j<n;j++) {
		in.yq=prng_xor(in.yq);
		//printf("yq=%u\n",in.yq);
		w=in.yq%(NQ);
		if (aq[w]==1) {
			in.y=(in.y^(1<<w));
			aq[w]=0;
		}
		else n++;
	}

	//printf("%u, ",in.out);
	in.out=in.y;

	return in;
}
#endif


// For managing criticality level
double getCaptureRate(double r0, int maxDefinedCoverSetNumber, double minCaptureRate, double maxCaptureRate, int myCoversetSize) {

	double captureRate=minCaptureRate;
	int X;

	X=myCoversetSize+1;

	// TODO artificially increase the size of a mobile (camera rotation) node?	

	if (X >= maxDefinedCoverSetNumber) X=maxDefinedCoverSetNumber;

#ifdef CRITICALITY_WITH_BEZIER
	// Bezier behavior, see WCNC'10 paper
	if (myCoversetSize>=0) {
		double hx = (double)maxDefinedCoverSetNumber;
		double hy = maxCaptureRate;

		if (r0 < 0) r0=0;

		double bx = -hx*r0+hx;
		double by = hy*r0;

		double aX = (-bx+sqrt(bx*bx-2*bx*X+hx*X))/(hx-2*bx);

		if ( (hx-2*bx)==0 ) {
			captureRate=((hy-2*by)/(4*bx*bx))*X*X+(by/bx)*X;
		}
		else {
			captureRate=(hy-2*by)*aX*aX+2*by*aX;
		}

		if (captureRate<minCaptureRate) captureRate=minCaptureRate;
	}

	return captureRate;
#else
	// dumb behavior
	return minCaptureRate;

	if (r0 <=0) {
		return minCaptureRate;
	}
	else {
		if (X==0)
			return minCaptureRate;

		if (X==maxDefinedCoverSetNumber)
			return maxCaptureRate;
		
		if (r0 == 0.5)
			return (double)X;
	}
	return minCaptureRate;
#endif
}

#ifdef CRITICALITY_ADAPTATIVE_MAC

int VideoSensorNode::masterSelection() {

	int x = minCaptureRate;
	int master = this->neighbors[0].id;

	for (int i=0; i<this->neighbors.size(); i++) {
		if (this->neighbors[i].capturingRate > x) {
			//trace() << " node has sent an activity msg :\n";
			x = this->neighbors[i].capturingRate;
			master = this->neighbors[i].id;
		}
	}

	return master;
}
#endif

// ----------------------------
// IMPORTANT: gate numbering rules under OMNET++
// Assume each sensor has N(4) neighbors. It will have N out gates numbered from 0..N-1(3).

void VideoSensorNode::initialize() {

#ifdef CASTALIA
	// set the teestream output to both ev (OMNET++) and trace() (Castalia)
	videoOutputStream.set_streambuf(ev.getOStream(), (hasPar("collectTraceInfo") &&
														par("collectTraceInfo"))?DebugInfoWriter::getStream():(std::ostream&)empty);

	// get the coordinator node which is just a logical node for the simulation
	// In Castalia, the sensor application module is a son of Node (Castalia defined)
	// which is contained in the SensorNetwork module 
	cModule *parent = getParentModule();
	cModule *coordNodeModule = parent->getParentModule()->getSubmodule("coordinator");
    
	//get a valid reference to the object of the Resources Manager module so that we can make direct calls to its public methods
	//instead of using extra messages & message types for tighlty couplped operations
	if (parent->findSubmodule("ResourceManager") != -1) {
		resMgrModule = check_and_cast <ResourceManager*>(parent->getSubmodule("ResourceManager"));
	} else {
		opp_error("\n[Application]:\n Error in geting a valid reference to ResourceManager for direct method calls.");
	}

	if (parent->findSubmodule("MobilityManager") != -1) {
		mobilityModule = check_and_cast <VirtualMobilityManager*>(parent->getSubmodule("MobilityManager"));
	} else {
		opp_error("\n[Application]:\n Error in geting a valid reference to MobilityManager for direct method calls.");
	}

	// we make no checks here
	radioModule = check_and_cast <Radio*>(parent->getSubmodule("Communication")->getSubmodule("Radio"));

	self = parent->getIndex();
	// set this->id to the self value used by Castalia
	this->id = self;

	cpuClockDrift = resMgrModule->getCPUClockDrift();

	// get the initial energy of batteries
	double maxEnergyLevel = resMgrModule->par("initialEnergy");

	// the node is disabled at simulation starting
	disabled = 1;

	stringstream out;
	out << self;
	selfAddress = out.str();

	applicationID = par("applicationID").stringValue();
	priority = par("priority");
	packetHeaderOverhead = par("packetHeaderOverhead");
	constantDataPayload = par("constantDataPayload");
	isSink = hasPar("isSink") ? par("isSink") : false;
#else
	// node's id is its number
    this->id = std::atoi(this->getName() + 4);
    
	// Without Castalia, both node modules and coordinator module are contained in the SensorNetwork module 
    cModule *coordNodeModule = getParentModule()->getSubmodule("coordinator");
#endif    
    CoordNode = check_and_cast<coordNode *>(coordNodeModule);
	// starts with no cover-sets	
	this->nodeCoverset.clear();
	// read parameters from .ini file
#ifdef CASTALIA
	this->alwaysActive = par("alwaysActive");
#endif
	this->minCaptureRate = par("minCaptureRate");
	this->maxCaptureRate = par("maxCaptureRate");
	this->staticCaptureRate = par("staticCaptureRate");
	this->nbCamera = par("nbCamera");
	this->aov[0] = par("aov");
	this->aov[1] = par("aov1");
	this->aov[2] = par("aov2");
	this->aov[3] = par("aov3");
	this->dov[0] = par("dov");
	this->dov[1] = par("dov1");
	this->dov[2] = par("dov2");
	this->dov[3] = par("dov3");

	this->dov[0]=this->dov[0]*(MAXFIELDUNIT/MAXFIELDMETER);
	this->dov[1]=this->dov[1]*(MAXFIELDUNIT/MAXFIELDMETER);
	this->dov[2]=this->dov[2]*(MAXFIELDUNIT/MAXFIELDMETER);
	this->dov[3]=this->dov[3]*(MAXFIELDUNIT/MAXFIELDMETER);
	// we start with the first camera
	this->nextActiveCamera = 0;
	this->cameraCycling = par("cameraCycling");
	this->cameraSimultaneous = par("cameraSimultaneous");

	this->maxBatteryLevel = par("maxBatteryLevel");
	this->energyPerCapture = par("energyPerCapture");
	this->criticalityLevel = par("criticalityLevel");
	this->minCriticalityLevel = par("minCriticalityLevel");
	this->maxCriticalityLevelPeriod = par("maxCriticalityLevelPeriod");
	this->maxDefinedCoverSetNumber = par("maxDefinedCoverSetNumber");
	this->forceNumberOfCoversets = par("forceNumberOfCoversets");
	this->isMobile = par("isMobile");
	this->isMobileProb = par("isMobileProb");
	this->isCamRotatable = par("isCamRotatable");
	this->isCamRotatableProb = par("isCamRotatableProb");
	this->maxMobilitySpeed = par("maxMobilitySpeed");
	this->maxCamRotationTime = par("maxCamRotationTime");
	this->maxCamRotationCount = par("maxCamRotationCount");
	this->load = par("load");

#ifdef CASTALIA

	this->selectiveFOV=par("selectiveFOV");
	this->virtualFOV=par("virtualFOV");
	this->onlyAltBC=par("onlyAltBC");

	this->idReportToSink = par("idReportToSink");
	this->recipientAddress = par("nextRecipient").stringValue();

	// for sending images
	this->sendImageOnIntrusion = par("sendImageOnIntrusion");
	this->forceSendImage = par("forceSendImage");
	this->forceSendImageCounter = 0;
	this->forceSendImageInterval = par("forceSendImageInterval");
	this->imageByteSize = par("imageByteSize");
	this->imageChunkSize = par("imageChunkSize");
	this->imageFilename = par("imageFilename").stringValue();

	this->cranEncoding = par("cranEncoding");
	// use to get the original information such as image size (in pixels),...
	this->imageBMPOriginalFilename = par("imageBMPOriginalFilename").stringValue(); // in .ini file

	this->displayReceivedImage = par("displayReceivedImage");
	this->waitForKeyPressWhenDisplayImage = par("waitForKeyPressWhenDisplayImage");
	this->displayReceivedImageTimer = par("displayReceivedImageTimer");
	this->keepImageFile = par("keepImageFile");
	this->imagePacketLossProb = par("imagePacketLossProb");
	this->imageSuccPacketLossCount=par("imageSuccPacketLossCount");
	this->imageByteCorruptedProb = par("imageByteCorruptedProb");

        this->imageNodeBaselineConsumption=((double)(resMgrModule->par("baselineNodePower"))) / 1000.0;

	// for realistic energy management
        this->measuredEnergyPerImageCapture = ((double)par("measuredEnergyPerImageCapture"))/1000.0;
        this->measuredEnergyPerImageEncoding = ((double)par("measuredEnergyPerImageEncoding"))/1000.0;
        this->measuredEnergyPerImageProcessing = ((double)par("measuredEnergyPerImageProcessing"))/1000.0;
	this->appConsumedEnergy = 0.0;

        this->timeForImageCapture = ((double)par("timeForImageCapture"))/1000.0;
        this->timeForImageEncoding = ((double)par("timeForImageEncoding"))/1000.0;
        this->timeForImageProcessing = ((double)par("timeForImageProcessing"))/1000.0;

	// for realistic sending overhead
	this->a_time_send =  par("aTimeSend");
	this->b_time_send =  par("bTimeSend");
        this->globalSendTime = ((double)par("globalSendTime"))/1000.0;
        this->globalRelayTime = ((double)par("globalRelayTime"))/1000.0;

	this->nextSendTime = 0.0;

#endif

	this->imageCaptureCount=0;

	// define whether the node is mobile
	if (this->isMobile)
		this->isMobile=(uniform(0,1)<this->isMobileProb)? true:false;
	// define whether the camera is rotatable
	if (this->isCamRotatable)
		this->isCamRotatable=(uniform(0,1)<this->isCamRotatableProb)? true:false;			
	// variable for the video node mobility model
	this->isMoving=false;
	this->startMobilityTime=0.0;
	this->mobilityDuration=0.0;
	// for camera rotation capability
	this->isCamRotating=false;
	this->startCamRotationTime=-1.0;
	this->camRotationCount=this->maxCamRotationCount;
	this->camRotationTime=this->maxCamRotationTime*this->camRotationCount;

	// the energy level is expressed in units	
    this->energyLevel = this->maxBatteryLevel;
	this->nbActivityMsgSent = 0;
    this->nbPositionPacketReceived = 0;
    this->nbStatusPacketReceived = 0;
	this->maxCriticalityLevel=this->criticalityLevel;
	
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	// enforce that the node start with the minimum criticality level
	this->criticalityLevel=this->minCriticalityLevel;
	this->hasAlertedNeighbors=false;
	this->isAlerted=false;
	this->nbAlertMsgRec=0;
	this->nbAlertMsgSent=0;
	this->nbAlertMsgProc=0;
	this->nbHasBeenInAlertedMode=0;
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
	this->canReinforce=false;
	this->nbTotalReinforcementMsgRec=0;
	this->nbReinforcementMsgRec=0;
	this->nbReinforcementMsgSent=0;
#endif	
#else
	// this->criticalityLevel is already set
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
	this->coverAndCapturePacketReceived = 0;
	this->alertResponseCount = 0;
#endif

#ifdef CRITICALITY_WITH_BEZIER	
	// this->minCaptureRate is already set
#else
	this->minCaptureRate=this->staticCaptureRate;
#endif
	// set to minCaptureRate at the beginning since no cover sets are found yet
	setCaptureRate(this->minCaptureRate);
    // includes the sensor itself
    this->minimumCoverSetNumber = 1; 
#ifdef WITH_INTRUSION
	// local counter for he sensor node to count to number of detected intrusions
  	this->intrusionViewCount=0;
	this->intrusionOccludedCount=0;
	this->forcedIntrusionProb = par("forcedIntrusionProb");
#ifdef CASTALIA
	this->activateCoversetOnIntrusion = par("activateCoversetOnIntrusion");
	this->propagateCoversetActivation = par("propagateCoversetActivation");
	this->receivedCoversetActivation = false;
	this->forcedCoversetActivationAt = par("forcedCoversetActivationAt");
	this->imageCountOnIntrusion = par("imageCountOnIntrusion");
	this->imageCountOnIntrusionCounter = 0;
	this->imageCountOnCoversetActivation = par("imageCountOnCoversetActivation");
	this->imageCountOnCoversetActivationCounter = 0;
#endif
#endif

#ifdef CASTALIA
	// keep the sequence number for images that will be sent by the video sensor node
	this->imageSN=0;

	// get position from the mobility manager
	NodeLocation_type nodeLocation;

	nodeLocation=mobilityModule->getLocation();
	this->p.x = nodeLocation.x;
	this->p.y = nodeLocation.y;

	// with a scaling factor for the omnet++ display window and for internal coverage computation
	this->p = this->p * MAXFIELDUNIT/MAXFIELDMETER;

	// then need to set the display string so that the nodes are correctly positioned
	cDisplayString &nodes = getParentModule()->getDisplayString();
	char s[10];
	sprintf(s, "%d", (int)this->p.x);
	nodes.setTagArg("p",0,s);
	sprintf(s, "%d", (int)this->p.y);
	nodes.setTagArg("p",1,s);
#else
    // Get position of the node, from the .ned file
    // Without Castalia, the coordinates in the .ned file are already scaled by a factor of MAXFIELDUNIT/MAXFIELDMETER=10
    cDisplayString &nodes = getDisplayString();
	this->p.x = std::atoi (nodes.getTagArg("p",0));
	this->p.y = std::atoi (nodes.getTagArg("p",1));
#endif

	// used for realizing the translation of all coordinates in meters into internal representation
	POSITION_INTERNAL_OFFSET=Point(MAX_DOV_INTERNAL, MAX_DOV_INTERNAL);

	// avoid negative coordinate
	// in this way, point v, b and c will not be outside the positive space coordinate
	// these coordinates are the real coordinate of the sensor nodes for FoV computation
	this->p = this->p + POSITION_INTERNAL_OFFSET;

#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << this->getName() << "\tid " << this->id <<"\tpx " << this->p.x << "\tpy " << this->p.y << endl;
#endif

#ifdef CASTALIA
	// initialize the connectivityMap object for distance and neighborhood discovery
	connectivityTool.init(this->id,
		getParentModule()->getParentModule()->par("numNodes"), this->p.x, this->p.y, this->radioModule);
#ifdef DEBUG_OUTPUT_LEVEL0
	ev << this->getName() << "\tTX output power=" << connectivityTool.initialTxOutputPower << "dBm." << endl;
#ifdef CASTALIA_MODIFIED_RADIO
	ev << "has Castalia modified radio" << endl;
#else
	ev << "has NO Castalia modified radio" << endl;
#endif

#ifdef CASTALIA_CHANGE_POWER_LEVELS
	ev << "will try to change TX power level" << endl;
#else
	ev << "will NOT try to change TX power level" << endl;
#endif
	ev << this->getName() << "\tTX nb output levels=" << connectivityTool.nbTxOutputPowerLevels << endl;
#endif
#endif

	// will depend on the value of aov
	for (int i=0; i<this->nbCamera; i++)
		this->alpha[i] = ((this->aov[i]/2)*2*PI)/360;

	// this->proba will be also used for other purposes
	this->proba = uniform(0,1);
#ifdef COVERAGE_MIXED_ANGLEVIEW
	// some Node(80%) have a smaller angle of view of PI/18 = 20°
	if (this->proba > 0.2)
		this->alpha[0] = PI/18;
#endif

	// find coordinate of point v of vector pv: set a direction for the camera
	// node that can rotate their camera will update their lineOfSight during the simulation
	if (hasPar("lineOfSight"))
		this->lineOfSight = par("lineOfSight");
	else
		this->lineOfSight = -1.0;

	// random line of sight
	if (this->lineOfSight < 0)
		this->lineOfSight = uniform(0,1) * 2 * PI;
	else
		this->lineOfSight = (this->lineOfSight * 2 * PI) / 360;

	// in decreasing order so that at the end v, b and c correspond to aov[0]
        // we have to register all cameras
	for (int i=this->nbCamera; i>=0 ; i--) {
		// LOS of multiple camera systems are uniformly distributed in the 2*PI range
		this->v.x=this->dov[i]*cos(this->lineOfSight+i*2*PI/this->nbCamera)+this->p.x;
		this->v.y=this->dov[i]*sin(this->lineOfSight+i*2*PI/this->nbCamera)+this->p.y;

		// find point b, c for each sensor node
		this->findTriangleCoord(this->p, this->v, this->b, this->c, this->dov[i], this->alpha[i]);

		// fill in the Triangle structure for geometric manipulation
		myFoV[i]=Triangle(this->p,this->b,this->c);

		// registers this node in the coordinator's node table
		CoordNode->fillNodes(this->p, this->v, this->b, this->c, this->id, this->nbCamera, i, this);
	}

	// for the moment, g does not take into account multiple aov from multiple cameras
	// so cover-set computation is only limited to only one camera
	// TODO: extends to multiple cameras
	// point g
	this->g.x = (this->p.x + this->b.x + this->c.x)/3;
	this->g.y = (this->p.y + this->b.y + this->c.y)/3;

	// alternate point gp between g and p
	this->gp.x = (this->p.x + this->g.x)/2;
	this->gp.y = (this->p.y + this->g.y)/2;

	// alternate point gv between g and v
	this->gv.x = (this->g.x + this->v.x)/2;
	this->gv.y = (this->g.y + this->v.y)/2;

	// alternate point gc between g and c
	this->gc.x = (this->c.x + this->g.x)/2;
	this->gc.y = (this->c.y + this->g.y)/2;

	// alternate point gb between g and b
	this->gb.x = (this->g.x + this->b.x)/2;
	this->gb.y = (this->g.y + this->b.y)/2;


#ifdef DISPLAY_SENSOR

	Point ap, av, ab, ac;

	ap = this->p * DISPLAY_SCALING_FACTOR;

	for (int i=0; i<this->nbCamera; i++) {

		av.x=this->dov[i]*DISPLAY_SCALING_FACTOR*cos(this->lineOfSight+i*2*PI/this->nbCamera)+ap.x;
		av.y=this->dov[i]*DISPLAY_SCALING_FACTOR*sin(this->lineOfSight+i*2*PI/this->nbCamera)+ap.y;

		// find point b, c for each sensor node
		double ah = this->dov[i]*DISPLAY_SCALING_FACTOR*sin(this->alpha[i]);
		double ad = this->dov[i]*DISPLAY_SCALING_FACTOR*cos(this->alpha[i]);

		ab.x = av.x + ah*(av.y - ap.y)/ad;
		ab.y = av.y - ah*(av.x - ap.x)/ad;

		ac.x = av.x - ah*(av.y - ap.y)/ad;
		ac.y = av.y + ah*(av.x - ap.x)/ad;

		Triangle displayFOV(ap,ab,ac);

		if (screen!=NULL) {
			if( SDL_MUSTLOCK( screen ) )
				SDL_LockSurface( screen );
			// draw the triangle FoV
			displayFOV.draw();
			SDL_Flip(screen);
			if( SDL_MUSTLOCK( screen ) )
				SDL_UnlockSurface( screen );
		}
	}
#endif

#ifdef CASTALIA
	// initialize the neighbor table
	this->neighbors.clear();
	// some nodes may not have neighbors, so need to set it to 0 first
	this->nbNeighbors = 0;
#else
    // get neighborhood, based on communication range, that should have been defined by the .ned file
    this->nbNeighbors = this->gateSize("in");
    assert (this->gateSize("in") == this->gateSize("out"));
#endif

#ifdef SCHEDULE_PROB
#ifdef SCHEDULE_PROB_CHAOS
	// initialize the chaotic probabilities
	this->chaos_in.xq=rand()%1000+1;
	this->chaos_in.yq=rand()%1000+1;
	this->chaos_in.y=this->id;
	// for 128 nodes -> 7 bits
	this->chaos_NQ=7;
	this->chaos_nbIteration=CHAOS_NB_INITIAL_ITERATION;
	this->chaos_totalIteration=0;

	this->nbChaosIterationStatVector.setName("nbChaosIteration");
#endif
	this->prob_totalWakeup=0;
	this->prob_nbActivityRound=1;
	this->prob_nbRetry=0;
	this->prob_nbSelfSchedule=0;
	// stop prob-based scheduling activity
	this->probStopActivityEv = new cMessage("probStopActivity", TIMER_SERVICE);

	// start chaotic-based scheduling activity, needed to avoid infinite sleeping
	this->probStartActivityEv = new cMessage("probStartActivity", TIMER_SERVICE);

#endif

#ifndef CASTALIA_CONNECTIVITYMAP
	// for broadcasting position information
	// with CASTALIA && CASTALIA_CONNECTIVITYMAP we use the HELLO mechanism instead
	this->diffusionEv = new cMessage("positionBroadcast", TIMER_SERVICE);
#endif

#ifdef CASTALIA
	// creates a message for self-notification to start compute cover set
	this->startComputeCoverEv = new cMessage("computeCover", TIMER_SERVICE);
	// creates a message for self-notification of forced sending of images
	this->forceSendImageEv = new cMessage("forceSendImage", TIMER_SERVICE);
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
	// creates a message for self-notification to trigger 2_hop data diffusion
	this->start2HopDataDiffusion = new cMessage("2hopdatadiffusion", TIMER_SERVICE);
	this->startCoverMemberDiffusion = new cMessage("covermemberdiffusion", TIMER_SERVICE);
	this->startComputeScore = new cMessage("computeScore", TIMER_SERVICE);
#endif
#endif

	// creates a message for self-notification of simulation increment
	this->imageCaptureEv = new cMessage("capture", TIMER_SERVICE);

	// creates a message for self-notification that status can change
	this->allowChangeStatusEv= new cMessage("allowChangeStatus", TIMER_SERVICE);

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	// creates a message for self-notification to go back to normal criticality level
	this->toNormalCriticalityLevelEv= new cMessage("toNormalCriticalityLevel", TIMER_SERVICE);

#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
	// creates a message for self-notification to enable refinforcement	
	this->canReinforceEv= new cMessage("canReinforce", TIMER_SERVICE);
#endif	
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
	this->startMasterSelectionEv = new cMessage("masterSelection", TIMER_SERVICE);
#endif

	// creates a message for self-notification that camera stops rotating
	this->stopCamRotationEv = new cMessage("stopCamRotation", TIMER_SERVICE);

	this->isDead = false;
#ifdef RETRY_CHANGE_STATUS
	this->retryChangeStatus=false;
#endif
	this->canChangeStatus=true;
#if defined(CASTALIA) && defined(CASTALIA_CONNECTIVITYMAP)
	// with Castalia when neighbors discovery is performed, we need to indicate that the node is not active yet 
	// in order to delay the node's position broadcast
	this->isActive = false;	
#else	
	// then turn the node to active at the beginning	
	this->isActive = true;
#endif	
	if (ev.isGUI()) {
		if (this->isMobile)
			setIcon("robot",0);
		else
			if (this->isCamRotatable) 
				setIcon("sensor-video-rot",0);
			else
				setIcon("sensor-video",0);
	}
	CoordNode->updateNodesStatus(this->id, this->isActive); 
	
#ifdef CASTALIA
	if (ev.isGUI())
			if (this->isSink)
				setIcon("block/filter_s",0);

	/* Send the STARTUP message to 1)Sensor_Manager, 2)Commmunication module, and
	 * 3)APP (self message) so that the node starts to operate.
	 */

	send(new cMessage("Sensor Dev Mgr [STARTUP]", NODE_STARTUP), "toSensorDeviceManager");
	send(new cMessage("Communication [STARTUP]", NODE_STARTUP), "toCommunicationModule");

	/*
	 * We are using a random delay because we don't want the nodes to be synchronized
	 */

	double random_startup_delay = parent->par("startupOffset");
	random_startup_delay += genk_dblrand(0) * (double)parent->par("startupRandomization");

	sendDirect(new cMessage("Resource Mgr [STARTUP]", NODE_STARTUP), random_startup_delay, 0, resMgrModule, "powerConsumption");

	scheduleAt(simTime() + random_startup_delay,
		   new cMessage("App [STARTUP]", NODE_STARTUP));

	// node is now enabled so that it can accept incoming packets
	disabled = 0;

	declareOutput("AppPacket sent");
	declareOutput("AppPacket received");
	declareOutput("Image received");
	declareOutput("Image sent");
	declareOutput("Image displayed");
	declareOutput("Image discarded past");
	declareOutput("Image discarded future");
	declareOutput("Consumed AppEnergy");
	declareOutput("Intrusion breakdown");
	declareOutput("Coverset stats");
#else
	// without Castalia, we schedule the diffusion msg and the next capture msg in the initialize() method
	// otherwise, it will be done on reception of the NODE_STARTUP msg

	// avoids synchronizing all nodes, similar to the Castalia approach
	double random_startup_delay = genk_dblrand(0) * STARTUP_DELAY;
	// schedule initial messages
	//
	scheduleAt(simTime()+random_startup_delay, this->diffusionEv);

	// if Bezier model, then need to find the cover sets first before scheduling capture events
#ifndef CRITICALITY_WITH_BEZIER	
	if (this->isActive)
		scheduleAt(simTime()+random_startup_delay+1.0/this->myCaptureRate, this->imageCaptureEv);
#endif
#endif
	
#ifdef DEBUG_OUTPUT_LEVEL0
    ev << "node " << this->id << ": initialized with criticality level "<< this->criticalityLevel <<
		" MAX Coverset " << this->maxDefinedCoverSetNumber << " capture at " << this->myCaptureRate << " fps" <<
		 endl;
	if (this->isMobile)
		ev << "--> node " << this->id << ": has mobility feature" << endl;
	if (this->isCamRotatable)
		ev << "--> node " << this->id << ": has rotatable camera" << endl;
#ifdef CASTALIA
	if (this->cranEncoding)
		ev << "--> node " << this->id << ": will use CRAN encoding for real images" << endl;
	if (this->isSink)
		ev << "--> node " << this->id << ": is the SINK" << endl;
#endif
	// check whether we should emulate more coversets
	if (this->forceNumberOfCoversets >= 0)
		ev << "--> node " << this->id << ": is forced with " <<  this->forceNumberOfCoversets << " coversets" << endl;
#endif

#ifdef DEBUG_OUTPUT_LEVEL1
    float dist = distance(this->p,this->v);

	ev << this->getName() << "\tid: " << this->id << "\tpx: " << this->p.x <<"\tpy: " << this->p.y << "\talpha: " << this->alpha[0]*180/PI
			<< "\tvx: " << this->v.x << "\tvy: " << this->v.y << "\tLoS: " << this->lineOfSight*360/(2*PI) << endl;
	ev <<"\tbx: " << this->b.x << "\tby: " << this->b.y << "\tcx: " << this->c.x <<"\tcy: " << this->c.y << endl;
	ev << "\tDpv: " << dist << "\tDpb: "<< distance(this->p,this->b) << "\tDpc: "<< distance(this->p,this->c) << endl;

#endif
}

void VideoSensorNode::handleMessage(cMessage *msg)
{
	// update the text above the icon that indicates the capture rate, the nb of possible coverset
	// and the energy level
	if (ev.isGUI()) updateIconText();

	int msgKind = msg->getKind();

#ifdef CASTALIA
	if ((disabled) && (msgKind != NODE_STARTUP) && (msgKind != TIMER_SERVICE))
	{
		cout << "Node " << this->id << ": event " << msg->getName() << " rejected, node is disabled!" << endl;
		delete msg; msg = NULL;	return;
	}
#endif
 
	switch (msgKind) {

#ifdef CASTALIA
	/*--------------------------------------------------------------------------------------------------------------
	 * Sent by ourself in order to start running the Application.
	 *--------------------------------------------------------------------------------------------------------------*/
	case NODE_STARTUP: {
		handleNodeStartup();
		delete msg; msg=NULL; break;
	}

	case OUT_OF_ENERGY: {
		disabled = 1;
		// get the dead time
		collectOutput("Consumed AppEnergy", "0-DeadTime", SIMTIME_DBL(simTime()));

#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id <<": DEAD due to REAL energy shortage" << endl;
#endif
		setActiveStatus(false,true);
		break;
	}

	/*--------------------------------------------------------------------------------------------------------------
	 * Message received by the ResourceManager module. It commands the module to stop its operation.
	 *--------------------------------------------------------------------------------------------------------------*/
	case DESTROY_NODE: {
		disabled = 1;
		break;
	}
#endif

	//
	// all our internal message are of type TIMER_SERVICE for compatibility with the Castalia framework when used
	//
	case TIMER_SERVICE: {
	//////////////////////////////////////////////////////////////////////////////////////////////
    // SELF MESSAGES
    //////////////////////////////////////////////////////////////////////////////////////////////
    
		ASSERT(msg->isSelfMessage());

		////////////////////////////////////////////
		// diffusion msg
		//
		if (strcmp("positionBroadcas", msg->getName()) == 0) {
			handlePositionBroadcast();
			// this msg serves only once
			delete msg; msg=NULL; break;
		}

#ifdef CASTALIA
#ifdef CASTALIA_CONNECTIVITYMAP
		////////////////////////////////////////////
		// hello initiate msg, self message that will trigger a hello packet if it is the node's turn to do so.
		//
		if (strcmp("hello initiate", msg->getName()) == 0) {
			handleHelloInitiate();
			// this msg serves only once
			delete msg; msg=NULL; break;
		}
#endif
		////////////////////////////////////////////
		// idreport msg
		//
		if (strcmp("idreport", msg->getName()) == 0) {
			handleIdReport();
			// this msg serves only once
			delete msg; msg=NULL; break;
		}

		////////////////////////////////////////////
		// periodic forced sending of image msg
		//
		if (strcmp("forceSendImage", msg->getName()) == 0) {
			collectOutput("Image sent", "forced");
			this->forceSendImageCounter = 1;
			imageCapture();
			if (this->isActive)
				scheduleAt(simTime()+this->forceSendImageInterval, this->forceSendImageEv);
		}

		////////////////////////////////////////////
		// computeCover msg  
		//
		if (strcmp("computeCover", msg->getName()) == 0) {
			handleComputeCover();
			// this msg serves only once
			delete msg; msg=NULL; break;
		}

#ifdef BEST_COVERSET_SELECTION_STRATEGIES
		////////////////////////////////////////////
		// 2-hop data diffusion msg
		//
		if (strcmp("2hopdatadiffusion", msg->getName()) == 0) {
			handle2hopDataDiffusion();
			delete msg; msg=NULL; break;
		}

		////////////////////////////////////////////
		// Cover member diffusion msg
		//
		if (strcmp("covermemberdiffusion", msg->getName()) == 0) {
			handleCoverMemberDiffusionInitiate();
			delete msg; msg=NULL; break;
		}

		////////////////////////////////////////////
		// computeScore msg
		//
		if (strcmp("computeScore", msg->getName()) == 0) {
			handleComptuteScore();
			delete msg; msg=NULL; break;
		}
#endif

#ifdef SEND_REAL_IMAGE
		////////////////////////////////////////////
		// display image received
		//
		if (strcmp("displayImage", msg->getName()) == 0) {
			handleDisplayImage(msg->par("nodeID").longValue(), msg->par("imageSN").longValue());
			// we need to delete the message because each new image will create a new message
			delete msg; msg=NULL; break;
		}
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
		////////////////////////////////////////////
		///////Starting selection of Master/////////
		////////////////////////////////////////////
		if (strcmp("masterSelection", msg->getName()) == 0) {
			handleMasterSelection();
		}
#endif

#endif // CASTALIA

		////////////////////////////////////////////
		// periodic image capture msg
		//
		if (strcmp("capture", msg->getName()) == 0) {
			handleImageCapture();
		}

		////////////////////////////////////////////
		// allowChangeStatus msg
		//
		if (strcmp("allowChangeStatus", msg->getName()) == 0 && !this->isDead) {
			handleAllowChangeStatus();
		}

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
		////////////////////////////////////////////
		// toNormalCriticalityLevel msg
		//
		if (strcmp("toNormalCriticalityLevel", msg->getName()) == 0) {
			handleToNormalCriticalityLevel();
		}

#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER	
		////////////////////////////////////////////
		// canReinforce msg
		//	
		if (strcmp("canReinforce", msg->getName()) == 0) {	
			// can now reinforce	
			this->canReinforce=true;
		}		
#endif				
#endif
		////////////////////////////////////////////
		// camStopRotation msg
		//	
		if (strcmp("camStopRotation", msg->getName()) == 0) {	
			// stops the camera. As the camera rotation is virtual, the original camera line of sight is unchanged
			this->isCamRotating=false;
			this->startCamRotationTime=-1.0;
		}		

#ifdef SCHEDULE_PROB
		////////////////////////////////////////////
		// probStartActivity msg
		//
		if (strcmp("probStartActivity", msg->getName()) == 0) {
			this->probWakeNode();
		}

		////////////////////////////////////////////
		// probStopActivity msg
		//
		if (strcmp("probStopActivity", msg->getName()) == 0) {
			handleProbStopActivity();
		}
#endif

#if defined(CASTALIA) && defined(CASTALIA_CONNECTIVITYMAP)
		////////////////////////////////////////////
		// changeTxPower msg
		//
		if (strcmp("changeTxPower", msg->getName()) == 0) {
			handleChangeTxPower();
			// this msg serves only once
			delete msg; msg=NULL; break;
		}
#endif

		break;
	} // end of case TIMER_SERVICE

	//////////////////////////////////////////////////////////////////////////////////////////////
    // REMOTE MESSAGES
    //////////////////////////////////////////////////////////////////////////////////////////////

	//
	// all our remote message are of type APPLICATION_PACKET for compatibility with the Castalia framework when used
	//

	case APPLICATION_PACKET: {

#ifdef CASTALIA

		ApplicationPacket *rcvPacket = check_and_cast <ApplicationPacket*>(msg);
		AppNetInfoExchange_type info = rcvPacket->getAppNetInfoExchange();

#ifdef CASTALIA_CONNECTIVITYMAP	

		/////////////////////////////////////////////
		// it is a HELLO packet from other sensors
		//
		if (strcmp("hello", msg->getName()) == 0) {
			handleHelloPacket(msg);
		}
#endif		
#endif
		/////////////////////////////////////////////
		// it is an alert packet from other sensors
		//
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
		if (strcmp("alert", msg->getName()) == 0) {
			handleAlertPacket(msg);
		}
#endif    

#ifdef BEST_COVERSET_SELECTION_STRATEGIES
		/////////////////////////////////////////////
		// it is a 2-hop data diffusion packet from other sensors
		//
		if (strcmp("2-hop data diffusion", msg->getName()) == 0) {
			handle2hopDataDiffusionPacket(msg);
		}

		/////////////////////////////////////////////
		// it is a start cover member diffusion packet from other sensors
		//
		if (strcmp("start cover member diffusion", msg->getName()) == 0) {
			handleStartCoverMemberDiffusionPacket(msg);
		}

		/////////////////////////////////////////////
		// it is a cover member diffusion packet from other sensors
		//
		if (strcmp("cover member diffusion", msg->getName()) == 0) {
			handleCoverMemberDiffusionPacket(msg);
		}
#endif

		/////////////////////////////////////////////
		// it is a position packet from other sensors
		//
		if (strcmp("position", msg->getName()) == 0) {
			handlePositionPacket(msg);
		}

		/////////////////////////////////////////////
		// it is an update packet from other sensors
		// (a node starts moving or its camera starts rotating)
		if (strcmp("update_pos", msg->getName()) == 0) {
			handleUpdatePosPacket(msg);
		}

		/////////////////////////////////////////////
		// An activity/dead status packet
		// As we need to do the same thing if is a dead message, we put all this together.
		//
		if ( (strcmp("activity",msg->getName())== 0 || strcmp("dead",msg->getName())== 0) && !this->isDead) {
			handleActivityPacket(msg);
		}

#ifdef CASTALIA

#ifdef CRITICALITY_ADAPTATIVE_MAC
		//////////////////////////////////////////////////////////////////////
		// msg to send number of coversets and capture rate to other nodes
		//
		if (strcmp("coversets and capturerate",msg->getName())== 0) {
			handleCoverSetsAndCaptureRatePacket(msg);
		}
#endif

		/////////////////////////////////////////////
		// A coverset activation packet
		//
		if (strcmp("coverset activation",msg->getName()) == 0) {
			handleCoversetActivationPacket(msg);
		}

		/////////////////////////////////////////////
		// it is an IDREPORT packet from other sensors
		// should only be received by a sink node when the purpose is to test the routing layer

		if (strcmp("idreport", msg->getName()) == 0) {
			handleIdReportPacket(msg);
		}

		/////////////////////////////////////////////
		// it is an IMAGE packet from other sensors
		// should be received by a sink node, but can be received by any other node if specified in sendImage()

		if (strcmp("image", msg->getName()) == 0) {
			handleImagePacket(msg);
		}
#endif
		delete msg; msg=NULL; break;
	} // end of case APPLICATION_PACKET

#ifdef CASTALIA

	case RADIO_CONTROL_MESSAGE: {

		ev_v << "Node " << this->id << " <-- RADIO_CONTROL_MESSAGE " << msg->getName() << endl;

		if (strcmp("Radio buffer full", msg->getName()) == 0) {
#ifdef DEBUG_OUTPUT_LEVEL0
				ev_v << "Node " << this->id << ": RADIO BUFFERS FULL" << endl;
#endif
		}

		delete msg; msg=NULL; break;
	}

	case MAC_CONTROL_MESSAGE: {

		ev_v << "Node " << this->id << " <-- MAC_CONTROL_MESSAGE " << msg->getName() << endl;

		if (strcmp("MAC buffer full", msg->getName()) == 0) {
#ifdef DEBUG_OUTPUT_LEVEL0
				ev_v << "Node " << this->id << ": MAC BUFFERS FULL" << endl;
#endif
		}

		delete msg; msg=NULL; break;
	}

	case NETWORK_CONTROL_MESSAGE: {

		ev_v << "Node " << this->id << " <-- NETWORK_CONTROL_MESSAGE " << msg->getName() << endl;
		delete msg; msg=NULL; break;
	}

	default:
	{
		ev_v << "Node " << this->id << ": WARNING received packet of unknown type = " << msgKind << endl;
		delete msg; msg=NULL; break;
	}
#endif

	} // end of switch
}

void VideoSensorNode::handlePositionBroadcast() {
	sendPosition(false);
#ifdef CASTALIA
	// schedule a computeCover msg as the node may not have neighbors in which case it will never receive position packets
	// ONLY when CASTALIA_CONNECTIVITYMAP is not defined otherwise a real neighbor discovery is performed
	scheduleAt(simTime()+NEIGHBORS_DISCOVERY_DELAY, this->startComputeCoverEv);
#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << "Node " << this->id << ": will start building cover-sets in " << NEIGHBORS_DISCOVERY_DELAY << "s from now=" << simTime() << endl;
#endif
#endif
}

#ifdef CASTALIA

void VideoSensorNode::handleNodeStartup() {

	if (this->isActive) {

		// schedule the sending of an image if the node is forced to do it periodically
		if (this->forceSendImage)
			scheduleAt(simTime()+this->forceSendImageInterval, this->forceSendImageEv);
#ifdef CASTALIA_CONNECTIVITYMAP
		// start the cover-set computation, no position diffusion as this step has been done during the connectivityMap process
		scheduleAt(simTime(), this->startComputeCoverEv);
#else
		// schedules a broadcast of the sensor's position
		scheduleAt(simTime(), this->diffusionEv);
#endif
#ifndef CRITICALITY_WITH_BEZIER
		// if Bezier model, then need to find the cover sets first before scheduling capture events
		if (this->imageCaptureEv->isScheduled())
			cancelEvent(this->imageCaptureEv);

		scheduleAt(simTime()+1.0/this->myCaptureRate, this->imageCaptureEv);
#endif
	}
	else {
		// only at the first STARTUP
#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << ": received a STARTUP MSG" << endl;
#endif

#ifdef GPSR_ROUTING
		// set the node's position for the routing layer in case we have a geographic routing layer
		GpsrRoutingControlCommand *cmd = new GpsrRoutingControlCommand("GPSR set node pos", NETWORK_CONTROL_COMMAND);
		cmd->setGpsrRoutingCommandKind(SET_GPSR_NODE_POS);
		cmd->setDouble1(this->p.x);
		cmd->setDouble2(this->p.y);
#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << ": sends command to set node's position in routing layer" << endl;
#endif
		toNetworkLayer(cmd);
#endif

#ifdef CASTALIA_CONNECTIVITYMAP
		// need to schedule the various timer for neighbors discovery
#if defined(CASTALIA_CHANGE_POWER_LEVELS) && defined(CASTALIA_MODIFIED_RADIO)
		// start with the highest output power
		connectivityTool.currentTxOutputPowerIndex=0;
		setRadioTxOutputPower(radioModule->getTxOutputPowerByIndex(connectivityTool.currentTxOutputPowerIndex));
#endif

		// Start the timer(s) for the connectivity tool
		scheduleAt(simTime()+ INTER_PACKET_SPACING, new cMessage("hello initiate", TIMER_SERVICE));
		// will display the results and change the transmission power if needed
		// add a security time of 2s so that all messages have been received and processed,
		// otherwise some neighbors will be missing
		// this event is scheduled at the end of the neighbors discovery phase of all nodes
		scheduleAt(simTime()+ connectivityTool.txInterval_perOutputPowerLevel+2,
				   new cMessage("changeTxPower", TIMER_SERVICE));
#endif
	}
}

#ifdef CASTALIA_CONNECTIVITYMAP
void VideoSensorNode::handleHelloInitiate() {
	double now = SIMTIME_DBL(simTime());

	int currentNodeTx = ((int)floor(now / connectivityTool.txInterval_perNode))%connectivityTool.totalSNnodes;

	// all nodes schedule a "hello initiate" but they will only be activated in sequence: 0 then 1 then 2,...
	if ( (connectivityTool.self == currentNodeTx) && (connectivityTool.packetsSent < SENT_PACKETS_PER_NODE_PER_LEVEL) ) {

		PositionPacket *pkt = new PositionPacket("hello", APPLICATION_PACKET);

		pkt->setId(this->id); // BYTE_T*2
		pkt->setPx(this->p.x); // FLOAT_T
		pkt->setPy(this->p.y); // FLOAT_T
		pkt->setVx(this->v.x); // FLOAT_T
		pkt->setVy(this->v.y); // FLOAT_T
		// for the moment, the cover-set computation does not take into account multiple camera systems
		pkt->setAlpha(this->alpha[0]); // FLOAT_T
		pkt->setActive(false); // BYTE_BIT_FIELD_T
		pkt->setSN(connectivityTool.serialNumber); // BYTE_T

		int payloadSize=BYTE_T*2+FLOAT_T*5+BYTE_BIT_FIELD_T+BYTE_T;

		pkt->setByteLength(payloadSize);

		// now send it to the network layer
		toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);

		collectOutput("AppPacket sent", "HELLO msg");

#ifdef DEBUG_OUTPUT_LEVEL2
		if (connectivityTool.serialNumber==0)
			ev << "Node " << this->id << ": initiates connectivityMap. simTime=" << now << endl;
			ev << "Node " << this->id << ": will use " << connectivityTool.nbTxOutputPowerLevels << " output power levels" << endl;
#endif

#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "Node " << this->id << " --> HELLO(" << connectivityTool.packetsSent << ") to its neighbors. simTime=" << now
				<< " output power=" << connectivityTool.currentTxOutputPower << "dBm" << endl;
#endif
		connectivityTool.serialNumber++;
		connectivityTool.packetsSent++;
	}

	// schedule a new event
	if (connectivityTool.packetsSent < SENT_PACKETS_PER_NODE_PER_LEVEL) {
		scheduleAt(now + INTER_PACKET_SPACING, new cMessage("hello initiate", TIMER_SERVICE));
	}
}
#endif

void VideoSensorNode::handleIdReport() {
	IdReportData tmpData;
	tmpData.nodeID = this->id;
	tmpData.locX = p.x;
	tmpData.locY = p.y;

	// get either NULL in which case it is the default sink string which is "SINK" for pro-active routing layers.
	string theSink=(getSinkAddressString())?this->recipientAddress:SINK_NETWORK_ADDRESS;

	for (int i=0; i<NB_IDREPORT_PACKET; i++) {

		IdReportDataPacket *packet2Net =
			new IdReportDataPacket("idreport", APPLICATION_PACKET);
		packet2Net->setExtraData(tmpData); // BYTE_T*2+FLOAT_T*2
		packet2Net->setSequenceNumber(i); // BYTE_T

#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << " --> IDREPORT to node " << theSink << " seq=" << i << endl;
#endif
		int payloadSize=BYTE_T*2+FLOAT_T*2+BYTE_T;

		packet2Net->setByteLength(payloadSize);

		toNetworkLayer(packet2Net, theSink.c_str());

		collectOutput("AppPacket sent", "ID report msg");
	}
}

void VideoSensorNode::handleComputeCover() {
#ifndef CASTALIA_CONNECTIVITYMAP
			// assume that after a NEIGHBORS_DISCOVERY_DELAY period, all possible neighbors have sent their position
			// from now on, this->nbNeighbors will keep the number of neighbors
			// Note that nbPositionPacketReceived may be 0
			this->nbNeighbors=this->nbPositionPacketReceived;
#ifdef DEBUG_OUTPUT_LEVEL2
			ev << "Node " << this->id << ": assumes " << this->nbPositionPacketReceived << " neighbors" << endl;
#endif
#endif
			computeCoverset();

			// after construction of coversets, if the node has to send a test packet, then schedule it 1 sec. later
			if (this->idReportToSink)
				scheduleAt(simTime()+1.0, new cMessage("idreport", TIMER_SERVICE));

#ifdef SCHEDULE_PROB
			// node will be active depending on a given probability
			// for the moment, half of them are active
			if (this->proba > 0.5) {
#ifdef DEBUG_OUTPUT_LEVEL1
				ev << "Node " << this->id << ": is INITIALLY ACTIVE FOR prob scheduling" << endl;
#endif
				this->probWakeNode();
			}
			else {
				this->canChangeStatus=true;
				setActiveStatus(false);
			}
#endif
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
			// initialize
			for (int i = 0; i < this->nbNeighbors; i++){
				// initialize 2-hop neighbor table
				this->neighbors[i].myNeighbors = new noeud[CoordNode->nbrNode];
				// Some node may not have 2-hop neighbors, so need to set it to 0 first
				this->neighbors[i].numNeighbors = 0;
				this->neighbors[i].notEmptyNeighborSet = false;
			}
			scheduleAt(simTime()+1.0, this->start2HopDataDiffusion);
#endif
}

#ifdef CRITICALITY_ADAPTATIVE_MAC

void VideoSensorNode::handleMasterSelection() {

	double masterCapturingRate = 0;
	int nodeNoOfCoverSets = 0;

	trace() << " Node " << this->id << " has Capture rate " << myCaptureRate << " :\n ";

	int masterNode = masterSelection();

	trace() << " Node " << this->id << " has chosen Node " << masterNode << " as master :\n ";

	int nIndex = getNeighborIndexById(masterNode);

	if (neighbors[nIndex].capturingRate == -1.0){
		neighbors[nIndex].capturingRate = minCaptureRate*3;}

	masterCapturingRate = neighbors[nIndex].capturingRate / maxCaptureRate;

	if (this->forceNumberOfCoversets > 0)
		nodeNoOfCoverSets = this->forceNumberOfCoversets ;
	else
		nodeNoOfCoverSets = this->nodeCoverset.size();

	trace() << " node " << this->id << " will use " << nodeNoOfCoverSets << " coversets and capturing rate " << masterCapturingRate << " :\n " ;

	TestMacControlCommand *cmd1 = new TestMacControlCommand ("TestMAC control command", MAC_CONTROL_COMMAND);

	cmd1->setTestMacCommandKind(SET_DUTY_CYCLE);
	cmd1->setCapturedRate(masterCapturingRate);
	cmd1->setCoverSets(nodeNoOfCoverSets);
	toNetworkLayer(cmd1);
}
#endif

#ifdef SEND_REAL_IMAGE
void VideoSensorNode::handleDisplayImage(int fromNodeID, int theImageSN) {

	// just in case, we check that we still have information on this node
	if (this->imageInfoRecordFromNodes.count(fromNodeID)!=0) {

		char *outfilename = const_cast<char *>(this->imageInfoRecordFromNodes[fromNodeID].imageFilename.c_str());

		char captionString[50];
		char imageType='I';

		if (this->imageInfoRecordFromNodes[fromNodeID].imageType==INTRUSION_IMG_TYPE)
			imageType='I';
		else if (this->imageInfoRecordFromNodes[fromNodeID].imageType==FORCED_IMG_TYPE)
			imageType='F';
		else if (this->imageInfoRecordFromNodes[fromNodeID].imageType==COVERSET_ACTIVATION_IMG_TYPE)
			imageType='C';
		else if (this->imageInfoRecordFromNodes[fromNodeID].imageType==RANDOM_INTRUSION_IMG_TYPE)
			imageType='R';

		sprintf(captionString,"%c%d(%.1f)-%d(%d)-%dof%d-DL=%.2f-L=%.2f", imageType, this->id, SIMTIME_DBL(simTime()), fromNodeID, theImageSN,
				this->imageInfoRecordFromNodes[fromNodeID].packetSNSet.size(),
				//this->imageInfoRecordFromNodes[fromNodeID].nbPacketsReceived,
				this->imageInfoRecordFromNodes[fromNodeID].nbPackets,
				SIMTIME_DBL(simTime())-this->imageInfoRecordFromNodes[fromNodeID].timePacketWasSent,
				SIMTIME_DBL(simTime())-this->imageInfoRecordFromNodes[fromNodeID].firstPacketReceived);

		ostringstream oss;
		string file = "sensor_images/IR";
		string theFilename = this->imageInfoRecordFromNodes[fromNodeID].imageFilename.c_str();

		// try to find where is the original filename part without the path
		string::size_type position = theFilename.rfind ('#');

		if (position != string::npos)
			// get only the filename part
			oss << file << "-" << captionString << "#" << theFilename.erase(0, position+1);
		else
			oss << file << "-" << captionString << "#" << "defaultname.ext";

		file = oss.str();

		char *newfilename=const_cast<char *>(file.c_str());

#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << ": WRITES IMAGE FILE(" << theImageSN << ") from node " << fromNodeID << ". Size="
				<< this->imageInfoRecordFromNodes[fromNodeID].imageSize << "." << captionString << endl;
#endif

		if (this->imageInfoRecordFromNodes[fromNodeID].cranEncoding) {

			// we just close the file that has been filled on the fly
			fclose(this->imageInfoRecordFromNodes[fromNodeID].pFile);

			// add here the code for decoding the image with the encoding technique from CRAN
			if (this->displayReceivedImage) {

				BMPImageStruct OriginalImage;
				int err;

				// Read the original BMP file to get useful informations
				err = ReadBitmapFile(const_cast<char *>(this->imageInfoRecordFromNodes[fromNodeID].imageBMPOriginalFilename.c_str()),
									 &OriginalImage);

				if (err) {
					ev_v << "Node " << this->id << ": CANNOT read original BMP file "
							<< this->imageInfoRecordFromNodes[fromNodeID].imageBMPOriginalFilename << endl;

				}
				else {
					// reset content
					for (int i=0; i<OriginalImage.imageVsize; i++)
						for (int j=0; j<OriginalImage.imageHsize; j++) OriginalImage.data[i][j]=0.0;

					//  Read the received file
					if ((this->imageInfoRecordFromNodes[fromNodeID].pFile=fopen(outfilename, "r")) == NULL) {
						ev_v << "Node " << this->id << ": CANNOT read received encoded file "
								<< this->imageInfoRecordFromNodes[fromNodeID].imageFilename << endl;
					}
					else {
						// Initialize the quantization matrix
						QTinitialization((unsigned int)this->imageInfoRecordFromNodes[fromNodeID].qualityFactor);

						// JPEG decoding
						while  (JPEGdepacketization(&OriginalImage, this->imageInfoRecordFromNodes[fromNodeID].pFile) == true);
						fclose(this->imageInfoRecordFromNodes[fromNodeID].pFile);

						JPEGdecoding(&OriginalImage, &OriginalImage);

						string tmpBMPFile = file.c_str();
						// try to find where is the extension
						string::size_type position = tmpBMPFile.rfind ('.');

						if (position != string::npos)
							tmpBMPFile.replace(position, 20, "-tmp.bmp");

						err = WriteBitmapFile(const_cast<char *>(tmpBMPFile.c_str()), &OriginalImage);

						if (err) {
							ev_v << "Node " << this->id << ": CANNOT write BMP file from received encoded image"
									<< this->imageInfoRecordFromNodes[fromNodeID].imageFilename << endl;

						}
						else {
							displayImage(tmpBMPFile.c_str(), captionString, this->waitForKeyPressWhenDisplayImage);

							if (!this->keepImageFile)
								remove(const_cast<char *>(tmpBMPFile.c_str()));
						}
					}
				}
			}
			else
				ev_v << captionString;
		}
		else {
			writeImage(outfilename, this->imageInfoRecordFromNodes[fromNodeID].raw_image,
					   this->imageInfoRecordFromNodes[fromNodeID].imageSize);

			// free the memory
			delete [] this->imageInfoRecordFromNodes[fromNodeID].raw_image;

			if (this->displayReceivedImage)
				displayImage(this->imageInfoRecordFromNodes[fromNodeID].imageFilename, captionString,
							 this->waitForKeyPressWhenDisplayImage);
			else
				ev_v << captionString;
		}

#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << ": DISPLAY REAL IMAGE(" << theImageSN << ") from node " << fromNodeID << ". "
				<< this->imageInfoRecordFromNodes[fromNodeID].packetSNSet.size() << "/"
				//<< this->imageInfoRecordFromNodes[fromNodeID].nbPacketsReceived << "/"
				<< this->imageInfoRecordFromNodes[fromNodeID].nbPackets << ". First packet received at "
				<< this->imageInfoRecordFromNodes[fromNodeID].firstPacketReceived << endl;

#endif
		if (!this->keepImageFile)
			remove(outfilename);
		else {
			// remane it in order to get final statistics in the file name
			rename(outfilename, newfilename);
		}

		// indicate that the image has been displayed, we expect the next one
		this->imageInfoRecordFromNodes[fromNodeID].displayedImageSN=theImageSN;
		// will allow the sink to detect a new image
		this->imageInfoRecordFromNodes[fromNodeID].currentImageSN=-1;

		if (this->imageInfoRecordFromNodes[fromNodeID].packetSNSet.size()==this->imageInfoRecordFromNodes[fromNodeID].nbPackets) {
			collectOutput("Image displayed", "complete");
			collectOutput("Image displayed", fromNodeID, "complete");
		}
		else {
			collectOutput("Image displayed", "truncated");
			collectOutput("Image displayed", fromNodeID, "truncated");
		}

		collectOutput("Image displayed", "all");
		collectOutput("Image displayed", fromNodeID, "all");
	}
	else {
#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << " CANNOT display image from node " << fromNodeID << ". Missing record" << endl;
#endif
	}
}
#endif
#endif // CASTALIA

void VideoSensorNode::handleImageCapture() {

	bool intrusionSeen=imageCapture();

	// when the sensor is active, there is one event scheduled each time the sensor is supposed to capture
	// It then decreases its energy at each capture
	if (this->isActive) {
		// find new captureRate
		setCaptureRate(this->getCaptureRate());

		// take care of timing because we need to recompute the time of next image capture according to timing constraints
		double timeToWait = this->timeForImageCapture + this->timeForImageProcessing;

		if (intrusionSeen)
			timeToWait += this->timeForImageEncoding;

		timeToWait = (timeToWait > 1.0/this->myCaptureRate)? timeToWait : 1.0/this->myCaptureRate;

		// schedule the next capture, depending on the capture rate
		scheduleAt(simTime()+timeToWait, this->imageCaptureEv);
	}
}

void VideoSensorNode::handleAllowChangeStatus() {

	this->canChangeStatus=true;
#ifdef RETRY_CHANGE_STATUS
	if (retryChangeStatus) {
#ifdef RETRY_CHANGE_STATUS_WPROBABILITY
		double theProb = retryChangeStatusTo?RETRY_CHANGE_STATUS_PROBABILITY_TO_ACTIVE:RETRY_CHANGE_STATUS_PROBABILITY_TO_INACTIVE;
		if (uniform(0,1)>theProb) {
#endif
#ifdef DEBUG_OUTPUT_LEVEL1
			string activeStr=retryChangeStatusTo?"ACTIVE":"INACTIVE";
			ev << "Node " << this->id <<" RETRY CHANGE STATUS TO " << activeStr << endl;
#endif
			if (ev.isGUI())
				bubble("RETRY");
			retryChangeStatus=false;
			setActiveStatus(retryChangeStatusTo);
			// sends to all neighbors
			// sendActivity();
#ifdef RETRY_CHANGE_STATUS_WPROBABILITY
		}
#endif
	}
#endif
}

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER

void VideoSensorNode::handleToNormalCriticalityLevel() {

	this->criticalityLevel=this->minCriticalityLevel;
	// will alert its neighbors at next intrusion detection
	this->hasAlertedNeighbors=false;
	// not in alerted mode anymore
	this->isAlerted=false;

	// clears information about neighbors
	for (unsigned int i=0; i<this->neighbors.size(); i++) {
		this->neighbors[i].hasSentAlertMsg=false;
		this->neighbors[i].lastSeq=0;
	}

#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
	// can not reinforce, it is for next intrusion detection
	this->canReinforce=false;
	this->nbReinforcementMsgRec=0;
#endif
	// keep track of how many time the node gets back to normal
	this->nbHasBeenInAlertedMode++;
	// the new capture rate will be determined at next capture
#ifdef DEBUG_OUTPUT_LEVEL2
	ev << "Node " << this->id << ": back to NORMAL criticality level" << endl;
#endif
	if (ev.isGUI())
		bubble("NORMAL");
}
#endif

#ifdef SCHEDULE_PROB
void VideoSensorNode::handleProbStopActivity() {

	unsigned int theNextNode;

	this->prob_nbActivityRound--;

	// test whether the node should go back to sleep
	if (this->prob_nbActivityRound>0) {
		// stay active
#ifdef DEBUG_OUTPUT_LEVEL1
		ev << "Node " << this->id << ": still has " << this->prob_nbActivityRound << " rounds" << endl;
#endif
		// schedules when the node should stop from being active
		scheduleAt(simTime()+this->getProbActivityTime(), this->probStopActivityEv);
	}
	else {
		bool nextNodeFound=false;
		bool noNextNodeFound=false;

		// record the number of iteration, i.e. how many time a node got elected during its activity period
		// this->nbChaosIterationStatVector.record(this->chaos_nbIteration-CHAOS_NB_INITIAL_ITERATION);
		// this->chaos_totalIteration=this->chaos_totalIteration+this->chaos_nbIteration-CHAOS_NB_INITIAL_ITERATION + 1;

		while (!nextNodeFound) {

#if defined(SCHEDULE_PROB) && !defined(SCHEDULE_PROB_CHAOS)
			theNextNode=rand() % this->CoordNode->nbrNode;
			assert(theNextNode>=0 && theNextNode<this->CoordNode->nbrNode);
#endif

#ifdef SCHEDULE_PROB_CHAOS
			// get next node by chaotic iterations
			for (int i=0; i<this->chaos_nbIteration; i++)
				this->chaos_in=ci_ver2(this->chaos_in,this->chaos_NQ);

			theNextNode=this->chaos_in.out;
#endif

			// if we have retried CHAOS_NB_RETRY_FIND_NEXT_NODE times then goes to sleep
			if (!this->CoordNode->probTestNode(theNextNode, this->id)
				&& (this->prob_nbRetry > PROB_NB_RETRY_FIND_NEXT_NODE)) {
#ifdef DEBUG_OUTPUT_LEVEL1
			ev << "Node " << this->id << ": has retried " <<  this->prob_nbRetry << " times. Goes to sleep." << endl;
#endif
				// we can say that we wake up ourself, but may increase the number of active nodes
				//theNextNode=this->id;
				noNextNodeFound=true;
				this->prob_nbRetry=0;
			}

#ifdef DEBUG_OUTPUT_LEVEL1
			ev << "Node " << this->id << ": will wake up node " <<  theNextNode << " for prob scheduling" << endl;
#endif
			// we wake up our-selves
			if (theNextNode==this->id) {
				this->prob_nbActivityRound++;
				this->prob_nbSelfSchedule++;
#ifdef SCHEDULE_PROB_CHAOS
				this->chaos_totalIteration++;
				this->chaos_nbIteration=CHAOS_NB_INITIAL_ITERATION;
#endif

#ifdef DEBUG_OUTPUT_LEVEL1
				ev << "Node " << this->id << ": re-schedules itself " << endl;
#endif
				// schedules when the node should stop from being active
				scheduleAt(simTime()+PROB_ACTIVITY_PERIOD, this->probStopActivityEv);
				// we found ourself as next node
				nextNodeFound=true;
				// get out of the loop
				continue;
			}
			else {
				// indicates to the coordNode the next node to be awaken
				nextNodeFound=this->CoordNode->probWakeNode(theNextNode, this->id);

				// if we have already tried the max number of time, set nextNodeFound inconditionaly
				if (noNextNodeFound)
					nextNodeFound=true;
			}

			if (nextNodeFound) {
#ifdef DEBUG_OUTPUT_LEVEL1
				ev << "Node " << this->id << ": goes to sleep from prob scheduling" << endl;
#endif
				// go to sleep mode;
				this->canChangeStatus=true;
				setActiveStatus(false);

				// security schedule of a wake up if no wake up for a long time
				// schedules when the node should stop from being active
				scheduleAt(simTime()+PROB_ACTIVITY_PERIOD*10, this->probStartActivityEv);
			}
			else {
#ifdef DEBUG_OUTPUT_LEVEL1
				ev << "Node " << this->id << ": node " << theNextNode << " already active or dead. Retry("
					<< this->prob_nbRetry << ")" << endl;
#endif
				this->prob_nbRetry++;
			}
		}
	}
}
#endif

#if defined(CASTALIA) && defined(CASTALIA_CONNECTIVITYMAP)

void VideoSensorNode::handleChangeTxPower() {
#ifdef DEBUG_OUTPUT_LEVEL1
	ev_v << "Node " << this->id << ": found " << (int)connectivityTool.neighborTable.size() << " neighbors at " <<
		connectivityTool.currentTxOutputPower << "dBm" << endl;

	ev << "[ ";

	for (int i=0; i < (int)connectivityTool.neighborTable.size(); i++) {
		ev << connectivityTool.neighborTable[i].id << " ";
	}

	ev << "]" << endl;

	ev_v << "TX Power Level " << connectivityTool.currentTxOutputPower << "dBm, Node [" <<
		this->id << "] received from (" << (int)connectivityTool.neighborTable.size() << ") at " << simTime() << ":\n";

#ifdef DEBUG_OUTPUT_LEVEL2
	for (int i=0; i < (int)connectivityTool.neighborTable.size(); i++) {
		ev_v << "[" << this->id << "<--" << connectivityTool.neighborTable[i].id << "]\t -->\t " <<
			connectivityTool.neighborTable[i].receivedPackets << "_out_of_" << SENT_PACKETS_PER_NODE_PER_LEVEL << "_at_" <<
			connectivityTool.currentTxOutputPower << "dBm " <<
			(int)connectivityTool.neighborTable.size() << "\t\t>> " << (int)connectivityTool.neighborTable[i].theDistance
			<< " m<<\t> " << connectivityTool.neighborTable[i].theDistance << " <\n";
	}
#endif
#endif
	// if we found the neighbors corresponding to the default output power, copy neighbors information
	if (connectivityTool.currentTxOutputPower==connectivityTool.initialTxOutputPower) {
#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "Node " << this->id << ": copies neighbors:" << endl;
		ev << "[ ";
#endif
		// copy the collected information into this->neighbors

		for (int i=0; i<(int)connectivityTool.neighborTable.size(); i++) {

			// here we can adjust which neighbors should be moved into the final neighbors list
			// depending on the number of received HELLO message knowing that SENT_PACKETS_PER_NODE_PER_LEVEL have been sent
			// here we take all discovered neighbors, i.e. receivedPackets > 0;
			if (connectivityTool.neighborTable[i].receivedPackets > 0) {

				noeud newNeighbor;

				newNeighbor.id=connectivityTool.neighborTable[i].id;
				newNeighbor.nbCamera=1;
				newNeighbor.p.x=connectivityTool.neighborTable[i].px;
				newNeighbor.p.y=connectivityTool.neighborTable[i].py;
				newNeighbor.v.x=connectivityTool.neighborTable[i].vx;
				newNeighbor.v.y=connectivityTool.neighborTable[i].vy;
				newNeighbor.dov[0]=connectivityTool.neighborTable[i].dov;
				newNeighbor.alpha[0]=connectivityTool.neighborTable[i].alpha;
				newNeighbor.isActive=true;

				// find point b, c for each sensor node
				this->findTriangleCoord(newNeighbor.p,
					newNeighbor.v,
					newNeighbor.b,
					newNeighbor.c,
					newNeighbor.dov[0],
					newNeighbor.alpha[0]);

				newNeighbor.fov[0]=Triangle(newNeighbor.p, newNeighbor.b, newNeighbor.c);

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
				newNeighbor.hasSentAlertMsg=false;
				newNeighbor.lastSeq=0;
#endif

#ifdef DEBUG_OUTPUT_LEVEL2
				ev << connectivityTool.neighborTable[j].id << " ";
#endif
				this->neighbors.push_back(newNeighbor);
				this->nbNeighbors++;
			}
		}
#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "] nbNeighbors=(" << this->nbNeighbors << "," << this->neighbors.size() << ")" << endl;
#endif
	}
	// if we have to change the output power then connectivityTool.nbTxOutputPowerLevels > 1
	// the second condition means we haven't reach the last possible output power
	if (connectivityTool.nbTxOutputPowerLevels > 1 &&
		connectivityTool.currentTxOutputPowerIndex < (connectivityTool.nbTxOutputPowerLevels -1)) {
		connectivityTool.neighborTable.clear();
		connectivityTool.packetsSent = 0;
		connectivityTool.serialNumber = 0;
		connectivityTool.currentTxOutputPowerIndex++;
#ifdef CASTALIA_MODIFIED_RADIO
		// set new output power
		setRadioTxOutputPower(radioModule->getTxOutputPowerByIndex(connectivityTool.currentTxOutputPowerIndex));
#endif
		// a new round of messages for the next TX power level
		scheduleAt(simTime()+ INTER_PACKET_SPACING, new cMessage("hello", TIMER_SERVICE));
		scheduleAt(simTime()+ connectivityTool.txInterval_perOutputPowerLevel,
				   new cMessage("changeTxPower", TIMER_SERVICE));
	}
	else {
		// set back the default TX output power
		setRadioTxOutputPower(connectivityTool.initialTxOutputPower);
		// set as active so that the NODE_STARTUP message will now schedule a coverset computation
		this->isActive=true;
		CoordNode->updateNodesStatus(this->id, this->isActive);

		// schedule a new NODE_STARTUP
		cModule *parent = getParentModule();
		double random_startup_delay = parent->par("startupOffset");
		random_startup_delay += genk_dblrand(0) * (double)parent->par("startupRandomization");

		scheduleAt(simTime() + random_startup_delay,
			new cMessage("Application --> Application (self)", NODE_STARTUP));
	}
}
#endif

#ifdef CASTALIA
#ifdef CASTALIA_CONNECTIVITYMAP
void VideoSensorNode::handleHelloPacket(cMessage* msg) {

	collectOutput("AppPacket received", "HELLO msg");

	PositionPacket *rcvPacket = check_and_cast<PositionPacket *>(msg);
	int theId = rcvPacket->getId();
	int thePx = rcvPacket->getPx();
	int thePy = rcvPacket->getPy();
	int theVx = rcvPacket->getVx();
	int theVy = rcvPacket->getVy();
	double theAlpha = rcvPacket->getAlpha();
	double theDov = rcvPacket->getDov();
	int theSN = rcvPacket->getSN();
#ifdef DEBUG_OUTPUT_LEVEL2
ev << "Node " << this->id << " <-- HELLO from node " << theId << "\tpx " << thePx << "\tpy " << thePy << endl;
#endif
	connectivityTool.updateNeighborTable(theId, theSN, thePx, thePy, theVx, theVy, theDov, theAlpha);
}
#endif
#endif

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER

void VideoSensorNode::handleAlertPacket(cMessage* msg) {
	// count how many times the node gets alerted
	nbAlertMsgRec++;
#ifdef CASTALIA
	collectOutput("AppPacket received", "Alert msg");
#endif
	ActivityPacket *pkt = check_and_cast<ActivityPacket *>(msg);

	/////
	// it is an original alert msg, records the first sender in the neighbor table
	if (pkt->getOriginal()) {

		int nIndex=getNeighborIndexById(pkt->getIdFirstSender());

		// this may happen if during the neighbor discovery process this node has not been discovered
		if (nIndex==-1) {
			ev_v << "Node " << this->id << " <-- ALERT from node " << pkt->getId()
					<< ". ERROR: no neighbor in table. IGNORING msg" << endl;
			recordScalar("error.no_neighbor_in_table_for_alert", simTime());
			return;
		}

		if (!this->neighbors[nIndex].hasSentAlertMsg) {
			this->neighbors[nIndex].hasSentAlertMsg=true;
			this->neighbors[nIndex].lastSeq=pkt->getSeq();
		}

		// Normally, if the sender node has already sent an alert, then the current sensor should already be in alert mode
		if (this->isAlerted) {
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
			this->nbReinforcementMsgRec++;
			this->nbTotalReinforcementMsgRec++;

			double newCriticalityLevel;

			// use step increase of 0.1 and needs 2 msg to move from 0.6 to 0.7 for instance
			// moving from 0.4 to 0.5 needs 2 more msg
			// moving from 0.5 to 0.6 needs 2 more msg
			// moving from 0.6 to 0.7 needs 2 more msg
			// moving from 0.7 to 0.8 needs 2 more msg
			// moving from 0.8 to 0.9 needs 2 more msg
			if (this->nbReinforcementMsgRec >= (int)((this->maxCriticalityLevel-ALERTED_NODE_CRITICALITY_LEVEL)*20))
				newCriticalityLevel=this->maxCriticalityLevel;
			else
				newCriticalityLevel=this->criticalityLevel+(this->nbReinforcementMsgRec >> 1)*0.1;

			setAlertCriticalityLevel(newCriticalityLevel);

			char s[20];
			sprintf(s, "REINFORCE %.2f", newCriticalityLevel);
			if (ev.isGUI())
				bubble(s);
#endif
			// notice that if a rotatable camera is already alerted, nothing is done
		}
		else { // moves from not alerted to alerted, either by an original or a propagated msg
#ifdef DEBUG_OUTPUT_LEVEL2
			ev_v << "Node " << this->id << " <-- ALERT from node " << pkt->getId() << endl;
#endif
			this->nbAlertMsgProc++;
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
			// assign a criticality level ALERTED_NODE_CRITICALITY_LEVEL, e.g. 0.6
			// at the first time then use reinforcement (see above)
			setAlertCriticalityLevel(ALERTED_NODE_CRITICALITY_LEVEL);
#else
			// same level of criticality than the sensor that sees the intrusion
			setAlertCriticalityLevel(pkt->getCriticalityLevel());
#endif
			char s[20];
			sprintf(s, "ALERT %.2f", this->criticalityLevel);
			if (ev.isGUI())
				bubble(s);

#ifdef ROTATECAM_ON_ALERT
			if (this->isCamRotatable) {
				// when a node with a rotatable camera gets alerted, it should start rotating its camera
				// for a given number of rotations and indicates to its neighbors that it is doing so
				this->isCamRotating=true;
				this->startCamRotationTime=SIMTIME_DBL(simTime());
				// reset the rotation count
				this->lastRotationCount = 0.0;
				// schedules when the camera has to stop rotating
				// camRotationTime already takes into account the number of rotation and the time for a complete rotation
				if (!this->stopCamRotationEv->isScheduled())
					scheduleAt(simTime()+this->camRotationTime, this->stopCamRotationEv);
#ifdef DEBUG_OUTPUT_LEVEL1
				ev_v << "Node " << this->id << " : rotates its camera " << simTime() << ":" << simTime()+this->camRotationTime <<
					endl;
#endif
				sprintf(s, "ROTATE %.2f", this->camRotationCount);
				if (ev.isGUI())
					bubble(s);
				recordScalar("info.criticality.cam.trigger_cam_rotation", simTime());
				// notify to neighbors
				sendPosition(true);
			}
#endif
		}
#ifdef DYNAMIC_PROPAGATE_CRITICALITY_WITH_BEZIER
		// propagates the alert message and put the original field in the new msg
		// may change the propagated criticality level here
		sendAlert(pkt->getCriticalityLevel(),pkt->getIdFirstSender(), pkt->getSeq());
#endif
	}
	//////
	// not an original but a propagated one
	else {
		// TODO
		// need to avoid ping-pong behavior. Not treated for the moment
	}
	
#ifdef CRITICALITY_ADAPTATIVE_MAC

	int alertSendingNode = pkt->getId();
	int alertNumb = pkt->getAlertNumber();

	trace() << " node " << this->id << " <-- received alert msg, alert/response, from node " << pkt->getId() << " and Alert Number is " << alertNumb << " :\n";

	CoordNode->countAlertResponse(alertSendingNode, alertNumb);
#endif

}
#endif

void VideoSensorNode::handlePositionPacket(cMessage* msg) {
	// handle incoming position packets
	// TODO: only one position packet is sent currently for Castalia,
	// if there are packet loss the number of neighbors may be smaller that
	// that found during the connectivityMap process.
	PositionPacket *pk = check_and_cast<PositionPacket *>(msg);
	// store position packet from neighbor sequentially, neighbors'id is stored in the table
	// the small number if neighbors justify the sequential ordering, more neighbors would need a
	// smarter indexing mechanism

	noeud newNeighbor;

	newNeighbor.id = pk->getId();
	newNeighbor.nbCamera=1;
	newNeighbor.p.x = pk->getPx();
	newNeighbor.p.y = pk->getPy();
	newNeighbor.v.x = pk->getVx();
	newNeighbor.v.y = pk->getVy();
	newNeighbor.dov[0] = pk->getDov();
	newNeighbor.alpha[0] = pk->getAlpha();
	newNeighbor.isActive = pk->getActive();

	newNeighbor.isMobile=pk->getIsMobile();
	newNeighbor.isCamRotatable=pk->getIsCamRotatable();

	this->findTriangleCoord(newNeighbor.p,
							newNeighbor.v,
							newNeighbor.b,
							newNeighbor.c,
							newNeighbor.dov[0],
							newNeighbor.alpha[0]);

	newNeighbor.fov[0]=Triangle(newNeighbor.p, newNeighbor.b, newNeighbor.c);

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	newNeighbor.hasSentAlertMsg=false;
	newNeighbor.lastSeq=0;
#endif

#ifdef DEBUG_OUTPUT_LEVEL1
	ev << "Node " << this->id << " <-- POSITION packet from node " << newNeighbor.id << endl;
#endif
	this->nbPositionPacketReceived++;

	this->neighbors.push_back(newNeighbor);

#ifdef CASTALIA
	collectOutput("AppPacket received", "Position msg");

	// first message?
	if (this->nbPositionPacketReceived==1) {
		// safeguard as normally this->startComputeCoverEv should have been scheduled
		if (this->startComputeCoverEv->isScheduled())
			cancelEvent(this->startComputeCoverEv);

		// re-schedule the event
		scheduleAt(simTime()+NEIGHBORS_DISCOVERY_DELAY, this->startComputeCoverEv);
	}
#else
	// have received from all neighbors? Then compute the cover-sets
	if (this->nbPositionPacketReceived>=this->nbNeighbors) {
#ifdef DEBUG_OUTPUT_LEVEL1
	ev << "Node " << this->id << " <-- POSITION packets from " << this->nbNeighbors << " neighbors" << endl;
#endif
		computeCoverset();

#ifdef SCHEDULE_PROB
		// node will be active depending on a given probability
		// for the moment, half of them are active
		if (this->proba > 0.5) {
#ifdef DEBUG_OUTPUT_LEVEL1
			ev << "Node " << this->id << ": is INITIALLY ACTIVE FOR prob scheduling" << endl;
#endif
			this->probWakeNode();
		}
		else {
			this->canChangeStatus=true;
			setActiveStatus(false);
		}
#endif
	}
#endif
}

void VideoSensorNode::handleUpdatePosPacket(cMessage* msg) {

	PositionPacket *pk = check_and_cast<PositionPacket *>(msg);
#ifdef CASTALIA
	collectOutput("AppPacket received", "Update pos msg");
#endif

	int nIndex=getNeighborIndexById(pk->getId());

#ifdef DEBUG_OUTPUT_LEVEL2
			ev << "Node " << this->id << " <-- UPDATEPOS packet from node " <<
				this->neighbors[nIndex].id << endl;
#endif

	// this may happen if during the neighbor discovery process this node has not been discovered
	if (nIndex==-1) {
		ev_v << "Node " << this->id << " <-- UPDATEPOS packet from node " <<
			this->neighbors[nIndex].id << ". ERROR: no neighbor in table. IGNORING msg" << endl;
		recordScalar("error.no_neighbor_in_table_for_update_pos", simTime());
		return;
	}

	// just to be sure that the node can really move or rotate its camera
	if (this->neighbors[nIndex].isMobile || this->neighbors[nIndex].isCamRotatable) {
		this->neighbors[nIndex].p.x = pk->getPx();
		this->neighbors[nIndex].p.y = pk->getPy();
		this->neighbors[nIndex].v.x = pk->getVx();
		this->neighbors[nIndex].v.y = pk->getVy();

		this->neighbors[nIndex].isMoving=pk->getIsMoving();
		this->neighbors[nIndex].startMobilityTime=pk->getStartMobilityTime();
		this->neighbors[nIndex].mobilitySpeed=pk->getMobilitySpeed();
		this->neighbors[nIndex].mobilityDuration=pk->getMobilityDuration();
		this->neighbors[nIndex].isCamRotating=pk->getIsCamRotating();
		this->neighbors[nIndex].startCamRotationTime=pk->getStartCamRotationTime();
		this->neighbors[nIndex].camRotationTime=pk->getCamRotationTime();
		this->neighbors[nIndex].camRotationCount=pk->getCamRotationCount();
	}
	// handles a camera rotation from a neighbor node
	if (this->neighbors[nIndex].isCamRotating) {
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
		// if the node has a camera that is currently rotating, do not stop.
		if (this->isAlerted && !this->isCamRotating) {
#ifdef DECREASE_CRITICALITY_IF_ROTATABLECAM
#ifdef DECREASE_CRITICALITY_IF_ROTATABLECAM_IF_G
			// when sensor'g point can be covered by the rotating camera during its rotation
			if (distance(this->g,this->neighbors[nIndex].p)<distance(this->neighbors[nIndex].p, this->neighbors[nIndex].v))
#endif
			// decrease the criticality level, but still greater than the normal criticality level
			// exceptionally uses the staticCaptureRate value as the increment above minCriticalityLevel
			this->criticalityLevel=this->minCriticalityLevel+this->staticCaptureRate;
			// will get back to normal when this->toNormalCriticalityLevelEv will be scheduled
			assert(this->toNormalCriticalityLevelEv->isScheduled());
#endif

#ifdef GET_BACK_NORMAL_CRITICALITY_IF_ROTATABLECAM
			// get back to normal when there is a rotatable camera nearby
#ifdef GET_BACK_NORMAL_CRITICALITY_IF_ROTATABLECAM_IF_G
			// when sensor'g point can be covered by the rotating camera during its rotation
			if (distance(this->g,this->neighbors[nIndex].p)<distance(this->neighbors[nIndex].p, this->neighbors[nIndex].v))
#endif
			// regardless of how long the camera will rotate and whether the rotating camera covers
			// efficiently the sensor's fov
			if (this->toNormalCriticalityLevelEv->isScheduled()) {
				cancelEvent(this->toNormalCriticalityLevelEv);
				scheduleAt(simTime(), this->toNormalCriticalityLevelEv);
			}
			// as the toNormalCriticalityLevelEv event will not be scheduled immediately
			// set isAlerted to false now to avoid processing multiple getback event
			this->isAlerted=false;
			recordScalar("info.criticality.cam.get_back_normal_criticality", simTime());
#endif
		}
		// next step, verify that the rotating camera will efficiently cover the sensor's fov
#endif
	}
}

void VideoSensorNode::handleActivityPacket(cMessage* msg) {
	bool activeStatus;
	int fromNodeId;

	if (strcmp("activity",msg->getName())== 0) {

		ActivityPacket *pkt = check_and_cast<ActivityPacket *>(msg);
#ifdef CASTALIA
		collectOutput("AppPacket received", "Activity msg");
#endif
		this->nbStatusPacketReceived++;
		fromNodeId=pkt->getId();
		activeStatus=pkt->getActive();
#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "Node " << this->id << " <-- ACTIVITY (" << nbStatusPacketReceived << ") from node " << fromNodeId << endl;
#endif
	}

	if (strcmp("dead",msg->getName())== 0) {

		Dead *pkt = check_and_cast<Dead *>(msg);
#ifdef CASTALIA
		collectOutput("AppPacket received", "Dead msg");
#endif
		fromNodeId=pkt->getId();
		// for a dead message, the node status is false
		activeStatus=false;
#ifdef DEBUG_OUTPUT_LEVEL1
		ev_v << "Node " << this->id << " <-- DEAD packet from node " << fromNodeId << endl;
#endif
		// mark all the cover sets where the dead node is involved in as dead
		for (unsigned int i=0; i<nodeCoverset.size(); i++ ) {
			if ((int)std::count(coversetTable[i].aCoverset.begin(), coversetTable[i].aCoverset.end(),
				fromNodeId)) {
				this->coversetTable[i].isActive=false;
				this->coversetTable[i].isDead=true;
#ifdef DEBUG_OUTPUT_LEVEL1
				ev_v << "Node " << this->id << ": marks cover-set " << i << " as dead" << endl;
#endif
			}
		}
	}

	if (activeStatus) {
		// insert this neighbor in the set of active neighbors
		// if duplicate message, don't care as the set removes duplicate keys
		activeNeighbors.insert(fromNodeId);
	}
	else
		// remove the neighbors from the set of active nodes
		activeNeighbors.erase(fromNodeId);

	// don't send again activity status before receiving status from all neighbors at least one time
	// otherwise the number of transient status messages will be too large

	// TODO may be a problem is cover-set has not been computed yet
	// TODO or when we can't get enough status message
	// TODO should probably do differently (timer) with Castalia
	if (this->nbStatusPacketReceived >= this->nbNeighbors) {
		// once we get the status of the neighbors, we can determine which cover-set is possible
		// assume the coversetTable is sorted accordingly, so take the first found cover set
		// if it is a dead notification, do the same procedure
		int tmpCoversetIndex=-1;
		int numActiveCoverset=this->computeActiveCoverset(tmpCoversetIndex);

		// the real number of cover set, those that are not marked as dead
		int currentCoversetSize=getNumberPossibleCoverset();

		// a node with high number of cover sets must remain active to act as sentry node
		if ( currentCoversetSize > this->maxDefinedCoverSetNumber) {
#ifdef DEBUG_OUTPUT_LEVEL2
			ev_v << "Node " << this->id <<": FORCED to be active w/proba 1.0 (" << nodeCoverset.size() << ">" <<
				this->maxDefinedCoverSetNumber << ")" << endl;
#endif
			//indicate that the sensor must be active
			tmpCoversetIndex=-1;
			if (ev.isGUI())
				bubble("FORCED 1.0");
		}
		else
		if ( currentCoversetSize > this->maxDefinedCoverSetNumber/2 && uniform(0,1) > 0.2) {
#ifdef DEBUG_OUTPUT_LEVEL2
			ev_v << "Node " << this->id << ": FORCED to be active w/proba 0.8 (" << nodeCoverset.size() << ">" <<
				this->maxDefinedCoverSetNumber << "/2)" << endl;
#endif
			//indicate that the sensor must be active
			tmpCoversetIndex=-1;
			if (ev.isGUI())
				bubble("FORCED 0.8");
		}
		else
		if ( currentCoversetSize > this->maxDefinedCoverSetNumber/4 && uniform(0,1) > 0.5) {
#ifdef DEBUG_OUTPUT_LEVEL2
			ev_v << "Node " << this->id << ": FORCED to be active w/proba 0.5 (" << nodeCoverset.size() << ">" <<
				this->maxDefinedCoverSetNumber << "/4)" << endl;
#endif
			//indicate that the sensor must be active
			tmpCoversetIndex=-1;
			if (ev.isGUI())
				bubble("FORCED 0.5");
		}

		// found a cover-set
		if (tmpCoversetIndex!=-1) {
			this->activeCoversetIndex=tmpCoversetIndex;

			if (this->isActive) {
				// if was active, can now be inactive
				if (setActiveStatus(false)) {
#ifdef DEBUG_OUTPUT_LEVEL2
					ev_v << "Node " << this->id << ": BECOMES inactive, energy level of " << this->energyLevel <<
						" use coverset index " << this->activeCoversetIndex << endl;
#ifdef DEBUG_OUTPUT_LEVEL2
					if (this->activeCoversetIndex!=-1)
						printSet(coversetTable[this->activeCoversetIndex].aCoverset);
#endif
#endif
				}
			}
			else {
#ifdef DEBUG_OUTPUT_LEVEL2
				ev << "Node " << this->id << ": REMAINS inactive, energy level of " << this->energyLevel
					<< endl;
#endif
			}
		}
		// did not find a cover-set
		else {
			if (this->isActive) {
				// if was active, must remains active
#ifdef DEBUG_OUTPUT_LEVEL2
				ev << "Node " << this->id << ": REMAINS active, energy level of " << this->energyLevel
					<< endl;
#endif
			}
			else {
				// was inactive, so must become active
				if (setActiveStatus(true)) {
#ifdef DEBUG_OUTPUT_LEVEL2
					ev << "Node " << this->id << ": BECOMES active, energy level of " << this->energyLevel
						<< endl;
#endif
				}
			}
		}
	}	// end of if (this->nbStatusPacketReceived >= nbNeighbors)
}

#ifdef CRITICALITY_ADAPTATIVE_MAC

//Reception of coversets and capture rate
void VideoSensorNode::handleCoverSetsAndCaptureRatePacket(cMessage* msg) {

	int fromNodeId;

	if (strcmp("coversets and capturerate", msg->getName())== 0) {

		CoversAndCaptureRatePacket *pkt = check_and_cast<CoversAndCaptureRatePacket *>(msg);

		this->coverAndCapturePacketReceived++;
		
		fromNodeId=pkt->getId();

		int nIndex=getNeighborIndexById(fromNodeId);

		if (nIndex==-1) {
			trace() << " node " << fromNodeId << " is not in the neighbor table of --> node " << this->id << ":\n";
		}
		else {
			neighbors[nIndex].capturingRate = pkt->getCapturingRate();
			neighbors[nIndex].noOfCoverSets = pkt->getNoOfCoverSets();
			//trace() << " number of cover sets in handleCoverSetsAndCaptureRatePacket of node " << fromNodeId << " are " << neighbors[nIndex].noOfCoverSets ;
		}
	}

	// will start selecting the master node after 3s.
	if (coverAndCapturePacketReceived==1) {
		scheduleAt(simTime() + 3.0 , this->startMasterSelectionEv);
	}
}
#endif

#ifdef CASTALIA
void VideoSensorNode::handleCoversetActivationPacket(cMessage* msg) {

	collectOutput("AppPacket received", "Cover activation msg");
	ActivityPacket *pkt = check_and_cast<ActivityPacket *>(msg);

#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << "Node " << this->id << " <-- COVERSET ACTIVATION from node " << pkt->getId() << endl;
#endif
	unsigned int c_size=pkt->getNeighborIdListArraySize();

	for (unsigned int i=0; i<c_size; i++) {
#ifdef DEBUG_OUTPUT_LEVEL1
		ev_v << "Neighbor node " << pkt->getNeighborIdList(i) << " is listed" << endl;
#endif
		if (pkt->getNeighborIdList(i)==this->id) {
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": SELECTED FOR COVERSET ACTIVATION" << endl;
#endif
			this->receivedCoversetActivation=true;
			this->imageCountOnCoversetActivationCounter = this->imageCountOnCoversetActivation;
			imageCapture();
		}
	}
}

void VideoSensorNode::handleIdReportPacket(cMessage* msg) {

	collectOutput("AppPacket received", "ID report msg");
	IdReportDataPacket *rcvPacket = check_and_cast<IdReportDataPacket*>(msg);
	IdReportData theData = rcvPacket->getExtraData();

#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << "Node " << this->id << " <-- IDREPORT from node " << theData.nodeID << " seq=" << rcvPacket->getSequenceNumber() << endl;

#endif
	if (this->isSink) {
		ev_v << "Node " << this->id << ": is the default SINK" << endl;
	}
}

void VideoSensorNode::handleImagePacket(cMessage* msg) {

	ImageDataPacket *rcvPacket = check_and_cast<ImageDataPacket*>(msg);
	IdReportData theData = rcvPacket->getExtraData();

	int fromNodeID = theData.nodeID;
	int theImageSN = rcvPacket->getImageSN();
	int theSN = rcvPacket->getSequenceNumber();
	bool createRecord=false;
	bool updateRecord=false;

#ifdef SEND_REAL_IMAGE

	// packet is lost, but we don't want to loose the first ones which may contain very important data for image size
	if ((uniform(0,1) < this->imagePacketLossProb || this->imageSuccPacketLossCount) && theSN>6) {
#ifdef DEBUG_OUTPUT_LEVEL2
		ev_v << "Node " << this->id << " <-- REAL IMAGE(" << theImageSN << ") from node " << fromNodeID << " LOST seq="
			<< theSN << endl;
#endif
		collectOutput("Image received", "Packet lost");

		if (this->imageSuccPacketLossCount==0)
			this->imageSuccPacketLossCount=par("imageSuccPacketLossCount");
		else
			this->imageSuccPacketLossCount--;

		// don't do anything as the initial content is 0
	}
	else {

		// we already have a record
		if (this->imageInfoRecordFromNodes.count(fromNodeID)!=0) {

			// check whether the image is older than a previous displayed image from that node
			// that may happen in multipath routing and long buffer queues
			if (theImageSN <= this->imageInfoRecordFromNodes[fromNodeID].displayedImageSN) {
#ifdef DEBUG_OUTPUT_LEVEL2
				ev_v << "Node " << this->id << " DISCARD older image " << theImageSN << ") from node " << fromNodeID << " seq="
					<< theSN << "displayedSeq=" << this->imageInfoRecordFromNodes[fromNodeID].displayedImageSN << endl;
#endif
				collectOutput("Image discarded past", "Packets");
				collectOutput("Image discarded past", fromNodeID, "Packets");
				// do nothing because we want to discard these image packets
				return;
			}

			// it is the first packet of a new image, from an already known node
			if (this->imageInfoRecordFromNodes[fromNodeID].currentImageSN < 0) {

				createRecord=true;
				updateRecord=true;
			}
			else {
				// it is not the first packet, the node is already known and the packet belongs to the currently received image
				if (theImageSN==this->imageInfoRecordFromNodes[fromNodeID].currentImageSN) {
					if (this->cranEncoding)
						// add a space to start a new packet
						fputc((int)' ', this->imageInfoRecordFromNodes[fromNodeID].pFile);
				}
				else {
					// the packet belongs to another image, normally with higher seq number than the previous displayed one
					// and the currently being received, so either we discard it, either we discard the one currently being displayed

					// for the moment, we discard it
					collectOutput("Image discarded future", "Packets");
#ifdef DEBUG_OUTPUT_LEVEL2
					ev_v << "Node " << this->id << " DISCARD image in future " << theImageSN << ") from node " << fromNodeID << " seq="
						<< theSN << "displayedSeq=" << this->imageInfoRecordFromNodes[fromNodeID].displayedImageSN << endl;
#endif
					return;
				}

			}
		}
		else // first packet (received, not necessarely the first of the image) of a new image from a new node
			createRecord=true;

		if (createRecord) {

			struct imageInfoRecord newRecord;
			// when the first packet of the image has been sent: getSentTime() returns the same value for all packets
			newRecord.timePacketWasSent=rcvPacket->getSentTime();
			newRecord.nodeID = fromNodeID;
			newRecord.currentImageSN = theImageSN;
			if (updateRecord)
				// get the previous value
				newRecord.displayedImageSN=this->imageInfoRecordFromNodes[fromNodeID].displayedImageSN;
			else
				newRecord.displayedImageSN = -1;
			// will compute the real image size
			newRecord.imageSize = 0;
			newRecord.packetSNSet.clear();
			newRecord.nbPackets = rcvPacket->getNbPackets();

			// CRAN_ENCODING_IMAGE_TYPE is -1, if the size is negative it means that we do not know the size at encoding time
			if (rcvPacket->getImageSize()==CRAN_ENCODING_IMAGE_TYPE) {
				newRecord.cranEncoding = true;
				newRecord.imageSizeX = rcvPacket->getImageSizeX();
				newRecord.imageSizeY = rcvPacket->getImageSizeY();
				newRecord.qualityFactor = rcvPacket->getQualityFactor();
			}
			else {
				newRecord.cranEncoding = false;
				newRecord.imageSize = rcvPacket->getImageSize();
				// create storage for the new image
				newRecord.raw_image = new char[newRecord.imageSize];
				// fill the memory content with 0
				for (int i=0; i<newRecord.imageSize; i++)
					newRecord.raw_image[i]=0;
			}

			newRecord.nbPacketsReceived = 0;
			newRecord.firstPacketReceived = SIMTIME_DBL(simTime());
			newRecord.imageType = rcvPacket->getImageType();

			string imageFilename = rcvPacket->getImageFilename();

			ostringstream oss;
			string file = "sensor_images/IR";

			// try to find if there is a directory part
			string::size_type position = imageFilename.rfind ('/');

			if (position != string::npos)
				// get only the filename part
				oss << file << "-" << fromNodeID << "-" << this->id << "SN-" << newRecord.currentImageSN << "#" <<
						imageFilename.erase(0, position+1);
			else
				oss << file << "-" << fromNodeID << "-" << this->id << "SN-" << newRecord.currentImageSN << "#" << imageFilename;

			file = oss.str();

			// name of the received image
			newRecord.imageFilename = file;

			if (newRecord.cranEncoding) {
				// create the file
				newRecord.pFile = fopen(newRecord.imageFilename.c_str(), "w");

				// keep the original BMP filename, will need it to get some usefull information for display
				newRecord.imageBMPOriginalFilename = rcvPacket->getImageBMPOriginalFilename();
			}

			// create the event that will trigger the display of this image
			newRecord.displayImageEv = new cMessage("displayImage", TIMER_SERVICE);
			newRecord.displayImageEv->addPar("nodeID");
			newRecord.displayImageEv->addPar("imageSN");
			newRecord.displayImageEv->par("nodeID").setLongValue(newRecord.nodeID);
			newRecord.displayImageEv->par("imageSN").setLongValue(newRecord.currentImageSN);

			if (updateRecord) {
				// we first have to erase the existing one, since map.insert does not allow for duplicate nor update
				this->imageInfoRecordFromNodes.erase(fromNodeID);
			}

			// insert the new info record
			this->imageInfoRecordFromNodes.insert(pair<int,imageInfoRecord>(fromNodeID, newRecord));

			// schedule the display of the image
			scheduleAt(simTime()+this->displayReceivedImageTimer, newRecord.displayImageEv);

			collectOutput("Image received", "Images");
			collectOutput("Image received", fromNodeID, "Images");
		}

		// now, should have a record, an existing one or a newly created

		int dataSize=rcvPacket->getDataSize();

#ifdef DEBUG_OUTPUT_LEVEL2
		ev_v << "Node " << this->id << " <-- REAL IMAGE(" << theImageSN << ") from node " << fromNodeID << " seq="
			<< theSN << endl;
#endif

		if (this->imageInfoRecordFromNodes[fromNodeID].cranEncoding) {
			char theValue;
			// here we write in the file on the fly
			// write the packet size
			fprintf(this->imageInfoRecordFromNodes[fromNodeID].pFile, "%.4X", dataSize);

			for (int i=0; i<dataSize; i++) {
				fputc((int)' ', this->imageInfoRecordFromNodes[fromNodeID].pFile);
				theValue=rcvPacket->getImageData(i);
				fprintf(this->imageInfoRecordFromNodes[fromNodeID].pFile, "%.2X", (unsigned char)theValue);
			}

			fflush(this->imageInfoRecordFromNodes[fromNodeID].pFile);

			// now update the real image size
			this->imageInfoRecordFromNodes[fromNodeID].imageSize = this->imageInfoRecordFromNodes[fromNodeID].imageSize+dataSize;
		}
		else {
			// get the offset from the image msg in order to know where to assign new image data
			// note that with this mechanism we can handle late received image packets
			int offset=rcvPacket->getImageOffset();

			for (int i=0; i<dataSize; i++) {
				if (uniform(0,1) < this->imageByteCorruptedProb)
					// corrupted byte
					this->imageInfoRecordFromNodes[fromNodeID].raw_image[offset+i]=0;
				else {
					// get data from the payload
					this->imageInfoRecordFromNodes[fromNodeID].raw_image[offset+i]=rcvPacket->getImageData(i);
				}
			}
		}

		collectOutput("Image received", "Packets");
		collectOutput("Image received", fromNodeID, "Packets");

		// keep the SN of received packet, as the set does not allow for duplicate, packetSNSet.size() will give the number of
		// no-duplicate packet
		if (this->imageInfoRecordFromNodes[fromNodeID].packetSNSet.count(theSN)) {
			collectOutput("Image received", "Duplicate Packets");
			collectOutput("Image received", fromNodeID, "Duplicate Packets");
		} else
			this->imageInfoRecordFromNodes[fromNodeID].packetSNSet.insert(theSN);

		this->imageInfoRecordFromNodes[fromNodeID].nbPacketsReceived++;

		// we have received all the packets, so can display it before the timer expires
		//if ( (this->imageInfoRecordFromNodes[fromNodeID].nbPacketsReceived == this->imageInfoRecordFromNodes[fromNodeID].nbPackets) ||
		if ( (this->imageInfoRecordFromNodes[fromNodeID].packetSNSet.size() == this->imageInfoRecordFromNodes[fromNodeID].nbPackets) ||
				(theSN == (unsigned int)this->imageInfoRecordFromNodes[fromNodeID].nbPackets-1) ) {
			// cancel the initial timer
			if (imageInfoRecordFromNodes[fromNodeID].displayImageEv->isScheduled()) {
				cancelEvent(imageInfoRecordFromNodes[fromNodeID].displayImageEv);
				// display it asap
				scheduleAt(simTime(), imageInfoRecordFromNodes[fromNodeID].displayImageEv);
			}
		}
	}
#else
	// packet loss
	if (uniform(0,1) < this->imagePacketLossProb) {
#ifdef DEBUG_OUTPUT_LEVEL2
	ev_v << "Node " << this->id << " <-- FAKE IMAGE(" << theImageSN << ") from node " << theData.nodeID << " LOST seq="
			<< theSN << endl;
#endif
	collectOutput("Image received", "Packet lost");
	}
	else {
#ifdef DEBUG_OUTPUT_LEVEL2
	ev_v << "Node " << this->id << " <-- FAKE IMAGE(" << theImageSN << ") from node " << theData.nodeID << " seq="
			<< theSN << endl;
#endif
	collectOutput("Image received", "Packets");
	}
#endif
}

#endif

#ifdef SCHEDULE_PROB

double VideoSensorNode::getProbActivityTime() {

	double AP;

#ifdef SCHEDULE_PROB_ACTIVITY_CTE
	// exactly PROB_ACTIVITY_PERIOD
	return (double)PROB_ACTIVITY_PERIOD;
#endif

#ifdef SCHEDULE_PROB_ACTIVITY_RAND
	// between 1 and PROB_ACTIVITY_PERIOD
	AP = rand() % PROB_ACTIVITY_PERIOD + 1;
	return AP;
#endif

#ifdef SCHEDULE_PROB_ACTIVITY_RAND_UNI5
	// between PROB_ACTIVITY_PERIOD-5 and PROB_ACTIVITY_PERIOD+5
	// the mean is then CHAOS_ACTIVITY_PERIOD
	AP = rand() % 10 + (PROB_ACTIVITY_PERIOD - 5);
	return AP;
#endif

#ifdef SCHEDULE_PROB_ACTIVITY_RAND_UNI10
	// between PROB_ACTIVITY_PERIOD-10 and PROB_ACTIVITY_PERIOD+10
	// the mean is then CHAOS_ACTIVITY_PERIOD
	AP = rand() % 20 + (PROB_ACTIVITY_PERIOD - 10);
	return AP;
#endif

}

void VideoSensorNode::probWakeNode() {

	Enter_Method_Silent();
	// cancel the security wake up is the node has been normaly elected to wake up
	if (this->probStartActivityEv->isScheduled())
		cancelEvent(this->probStartActivityEv);

	if (this->isDead) {
#ifdef DEBUG_OUTPUT_LEVEL1
		ev << "Node " << this->id << ": is dead, skip prob wake up." << endl;
#endif
	}
	else {
		this->prob_totalWakeup++;
		this->prob_nbActivityRound=1;
		this->prob_nbRetry=0;
#ifdef SCHEDULE_PROB_CHAOS
		this->chaos_nbIteration=CHAOS_NB_INITIAL_ITERATION;
#endif
		this->canChangeStatus=true;
		this->setActiveStatus(true);

#ifdef DEBUG_OUTPUT_LEVEL1
		ev << "Node " << this->id << ": is waken up for prob scheduling" << endl;
#endif
		// this macro is needed since probWakeNode is called from the coorNode module
		// any interaction with omnet++'s object (such as scheduling an event) needs that macro
		Enter_Method_Silent();
		// schedules when the node should stop from being active
		if (this->probStopActivityEv->isScheduled())
			cancelEvent(this->probStopActivityEv);
		scheduleAt(simTime()+this->getProbActivityTime(), this->probStopActivityEv);
	}
}
#endif

double VideoSensorNode::getCaptureRate()
{
	int X;

	// if we emulate more coversets, set X accordingly
	if (this->forceNumberOfCoversets >=0)
		X=this->forceNumberOfCoversets;
	else
		X=this->getNumberPossibleCoverset();

	return ::getCaptureRate(this->criticalityLevel, this->maxDefinedCoverSetNumber,
				this->minCaptureRate, this->maxCaptureRate, X);
}

void VideoSensorNode::setCaptureRate(double theRate)
{
	this->myCaptureRate = theRate;
	recordScalar("info.cam.captureRate",myCaptureRate);
	CoordNode->meanCaptureRate.collect(this->myCaptureRate);
	// possibly perform other tasks
}

void VideoSensorNode::setIcon(const char* s, int i)
{
#ifdef CASTALIA
	// get position from the mobility manager
	NodeLocation_type nodeLocation;
	Point PP;

	nodeLocation=mobilityModule->getLocation();
	PP.x = nodeLocation.x;
	PP.y = nodeLocation.y;

	// with a scaling factor for the omnet++ display window and for internal coverage computation
	PP = PP * MAXFIELDUNIT/MAXFIELDMETER;

	cDisplayString &nodes = getParentModule()->getDisplayString();

	char si[10];
	sprintf(si, "%d", (int)PP.x);
	nodes.setTagArg("p",0,si);
#ifdef DEBUG_OUTPUT_LEVEL3
	ev << "Node " << this->id <<": updates its position (" << si << ",";
#endif
	sprintf(si, "%d", (int)PP.y);
	nodes.setTagArg("p",1,si);
#ifdef DEBUG_OUTPUT_LEVEL3
	ev << si << ")" << endl;
#endif
#else
	cDisplayString &nodes = getDisplayString();
#endif

	nodes.setTagArg("i",i,s);
}

void VideoSensorNode::updateIconText()
{
#ifdef CASTALIA
	cDisplayString &nodes = getParentModule()->getDisplayString();
#else
	cDisplayString &nodes = getDisplayString();
#endif
	char s[25];
	int nCover=getNumberPossibleCoverset();

	sprintf(s, "%.2f(%d:%.1f)", this->myCaptureRate, nCover,energyLevel);

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	if (this->isAlerted)
		strcat(s,"**");
#endif

	//bubble(s);
	nodes.setTagArg("t",0,s);

	if (!this->isDead) {
		if (nCover>maxDefinedCoverSetNumber)
			setIcon("red",1);
		else
		if (nCover>maxDefinedCoverSetNumber/2)
			setIcon("yellow",1);
		else
		if (nCover>maxDefinedCoverSetNumber/4)
			setIcon("blue",1);
		else
		if (nCover>=1)
			setIcon("cyan",1);
		else
		if (nCover==0)
			setIcon("green",1);
	}

#ifdef DEBUG_OUTPUT_LEVEL3
	ev << "Node " << this->id <<": updates its color status with nCover=" << nCover << endl;
#endif
}

bool VideoSensorNode::setActiveStatus(bool status, bool dead)
{
	if (dead) {
		this->isDead=true;
		this->isActive=false;
		this->energyLevel=-1.0;
		if (ev.isGUI()) {
			if (this->isMobile)
				setIcon("robot-dead",0);
			else
				if (this->isCamRotatable)
					setIcon("sensor-video-rot-dead",0);
				else
					setIcon("sensor-video-dead",0);

			updateIconText();
			setIcon("red",1);
		}
#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id <<": DEAD due to energy shortage" << endl;
#endif
		// broadcast to all neighbors
#ifdef CASTALIA
		Dead *pkt = new Dead("dead", APPLICATION_PACKET);
		pkt->setId(this->id); // BYTE_T*2

		int payloadSize=BYTE_T*2;

		pkt->setByteLength(payloadSize);

		// now send it to the network layer
		toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);

		collectOutput("AppPacket sent", "Dead msg");
#else
		for (int i=0; i<this->gateSize("out");i++) {
			Dead *pkt = new Dead("dead", APPLICATION_PACKET);
			pkt->setId(this->id);
			send(pkt,"out",i);
		}
#endif
		// indicates to the coordNode that the node is dead
		bool isAlive=false;
		CoordNode->updateNodesStatus(this->id, this->isActive, isAlive);

#ifdef SCHEDULE_PROB
		// need to cancel the probStopActivityEv if the node is dead
		if (this->probStopActivityEv->isScheduled())
			cancelEvent(this->probStopActivityEv);
#endif
		return true;
	}

#ifdef CASTALIA
	if (this->alwaysActive) {
		this->isActive=true;
		// if status is false, we indicate that we could not change to false
		return status;
	}
	else {
#endif
		// the required status is the same than previous
		if (this->isActive==status && dead==false)
			return true;

#ifdef NO_CHANGE_STATUS_CTRL
		// always enable status change
		this->canChangeStatus=true;
#endif

		if (this->canChangeStatus) {

			this->isActive = status;

			if (ev.isGUI()) {
				if (status) {
					if (this->isMobile)
						setIcon("robot",0);
					else
						if (this->isCamRotatable)
							setIcon("sensor-video-rot",0);
						else
							setIcon("sensor-video",0);
				}
				else
					if (this->isMobile)
						setIcon("robot-sleep",0);
					else
						if (this->isCamRotatable)
							setIcon("sensor-video-rot-sleep",0);
						else
							setIcon("sensor-video-sleep",0);
			}

			if (this->isActive) {
				if (this->isDead) {
					// TODO should not happen
					cout << "node " << this->id << ": active but should be dead!" << endl;
					if (ev.isGUI())
						setIcon("white",1);
					this->isActive=false;
					return false;
				}

				setCaptureRate(this->getCaptureRate());

                                if (!this->imageCaptureEv->isScheduled()) {
                                    // take care of timing because we need to recompute the time of next image capture according to timing constraints
                                    double timeToWait = this->timeForImageCapture + this->timeForImageProcessing;

                                    timeToWait = (timeToWait > 1.0/this->myCaptureRate)? timeToWait : 1.0/this->myCaptureRate;

                                    // schedule the next capture, depending on the capture rate
                                    scheduleAt(simTime()+timeToWait, this->imageCaptureEv);
                                }
			}

			CoordNode->updateNodesStatus(this->id, this->isActive);
			// broadcast to neighbors
			sendActivity();

#ifndef NO_CHANGE_STATUS_CTRL
			this->canChangeStatus=false;

			double statusChangeDelay;

			// if we change to active, then going next to inactive can be done earlier
			if (status)
				statusChangeDelay=TO_INACTIVE_STATUS_LATENCY_TIME;
			else
				statusChangeDelay=TO_ACTIVE_STATUS_LATENCY_TIME;

			if (!this->allowChangeStatusEv->isScheduled())
				scheduleAt(simTime()+statusChangeDelay, this->allowChangeStatusEv);
#endif
			return true;
		}
		else {
#ifdef DEBUG_OUTPUT_LEVEL2
			const char *s=status?"ACTIVE":"INACTIVE";
			ev << "Node " << this->id <<" CHANGE STATUS TO " <<  s << " IS REFUSED" << endl;
#endif
			if (ev.isGUI())
				bubble("REFUSED");
#ifdef RETRY_CHANGE_STATUS
			this->retryChangeStatus=true;
			this->retryChangeStatusTo=status;
#endif
			return false;
		}
#ifdef CASTALIA
	}
#endif
}

// does not take into account multiple camera systems
void VideoSensorNode::computeCoverset() {

#ifdef DEBUG_OUTPUT_LEVEL2
	// When received from all neighbors display the list of neighbors
	// Normally, this is executed only once, at the simulation beginning     
	for (int i = 0; i<this->neighbors.size(); i++) {
		ev << "Node " << this->id << ": has neighbor " << this->neighbors[i].id << "\t" << 
			this->neighbors[i].p.x << "\t"<< this->neighbors[i].p.y << endl;
	}
#endif     
    
#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << "Node " << this->id << ": has " << this->neighbors.size() << " neighbors, building cover-sets" << endl;
#ifdef WITH_OBSTACLES
	ev_v << "Node " << this->id << ": will take into account obstacles" << endl;
#endif
#endif
	Point PP(p.x,p.y);
	Point PB(b.x,b.y);
	Point PC(c.x,c.y);
	Point PG(g.x,g.y);
	Point PGP(gp.x,gp.y);
	Point PGV(gv.x,gv.y);
	Point PGB(gb.x,gb.y);
	Point PGC(gc.x,gc.y);

	// Build set A, B, C, G, AG, BG, CG to compute cover-sets
	// CoordNode->obstacleList.intersects() will always return false when there is no obstacles or when WITH_OBSTACLES is not set
	for (unsigned int i = 0; i<this->neighbors.size(); i++) {

		if (PP.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PP, this->neighbors[i].p))){
			A.insert(this->neighbors[i].id);
		}
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PP.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PP, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point P because of obstacles" << endl;
#endif

		if (PB.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PB, this->neighbors[i].p))){
            B.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PB.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PB, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point B because of obstacles" << endl;
#endif

		if (PC.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PC, this->neighbors[i].p))){
            C.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PC.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PC, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point C because of obstacles" << endl;
#endif

		if (PG.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PG, this->neighbors[i].p))){
            G.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PG.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PG, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point G because of obstacles" << endl;
#endif

		if (PGP.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PGP, this->neighbors[i].p))){
            GP.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PGP.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PGP, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point GP because of obstacles" << endl;
#endif

		if (PGV.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PGV, this->neighbors[i].p))){
            GV.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PGV.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PGV, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point GV because of obstacles" << endl;
#endif

		if (PGB.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PGB, this->neighbors[i].p))){
            GB.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PGB.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PGB, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point GB because of obstacles" << endl;
#endif

		if (PGC.is_inside(this->neighbors[i].fov[0]) && !CoordNode->obstacleList.intersects(Segment(PGC, this->neighbors[i].p))){
            GC.insert(this->neighbors[i].id);
        }
#ifdef DEBUG_OUTPUT_LEVEL1
		if (PGC.is_inside(this->neighbors[i].fov[0]) && CoordNode->obstacleList.intersects(Segment(PGC, this->neighbors[i].p)))
			ev << "Node " << this->id << ": neighbor " << this->neighbors[i].id << " cannot see point GC because of obstacles" << endl;
#endif
	}

#ifdef COVERAGE_WITH_ONLY_G

	nodeCoverset.clear();

	for (SetInt::iterator iy=G.begin(); iy!=G.end(); iy++) {
		SetInt s;
		s.insert(*iy);
		nodeCoverset.insert(s);
	}

#ifdef DEBUG_OUTPUT_LEVEL1
	ev << "Taking all nodes that cover G" << endl;
	printSetSet(nodeCoverset);
#endif

#else

#ifdef COVERAGE_WITH_G        		
    //AG 
    std::vector<int> vectorA(A.begin(), A.end());
	std::vector<int> setIntersectionAG;

#ifdef COVERAGE_WITH_ALT_G
#ifndef COVERAGE_WITH_XS_AOV
	// if the AoV is very small take only point GP, see below
    std::vector<int> vectorGP(GP.begin(), GP.end());
	set_intersection(vectorA.begin(), vectorA.end(),vectorGP.begin(),
        vectorGP.end(),back_inserter(setIntersectionAG));
#endif
#else        
	std::vector<int> vectorG(G.begin(), G.end());
    set_intersection(vectorA.begin(), vectorA.end(),vectorG.begin(),
        vectorG.end(),back_inserter(setIntersectionAG));
#endif            
    std::vector<int>::iterator k;

#ifdef COVERAGE_WITH_XS_AOV 
	// if the AoV is very small take only point GP  
	AG=GP;
#else         
    for (k = setIntersectionAG.begin(); k != setIntersectionAG.end(); ++k)
        AG.insert(*k);
#endif        
    //BG 
    std::vector<int> vectorB(B.begin(), B.end());
    std::vector<int> setIntersectionBG;

#ifdef COVERAGE_WITH_ALT_G

#ifdef COVERAGE_WITH_ALT_GBC
	std::vector<int> vectorGB(GB.begin(), GB.end());

	if (this->onlyAltBC)
		setIntersectionBG=vectorGB;
	else
		set_intersection(vectorB.begin(), vectorB.end(),vectorGB.begin(),
			vectorGB.end(),back_inserter(setIntersectionBG));
#else
	// take only point GV whatever the AoV is				
	std::vector<int> vectorGV(GV.begin(), GV.end());
    set_intersection(vectorB.begin(), vectorB.end(),vectorGV.begin(),
        vectorGV.end(),back_inserter(setIntersectionBG));
#endif

#else
    set_intersection(vectorB.begin(), vectorB.end(),vectorG.begin(),
        vectorG.end(),back_inserter(setIntersectionBG));
#endif     
    std::vector<int>::iterator l;

    for (l = setIntersectionBG.begin(); l != setIntersectionBG.end(); ++l)
        BG.insert(*l);

    //CG 
    std::vector<int> vectorC(C.begin(), C.end());
    std::vector<int> setIntersectionCG;

#ifdef COVERAGE_WITH_ALT_G

#ifdef COVERAGE_WITH_ALT_GBC
	std::vector<int> vectorGC(GC.begin(), GC.end());

	if (this->onlyAltBC)
		setIntersectionCG=vectorGC;
	else
		set_intersection(vectorC.begin(), vectorC.end(),vectorGC.begin(),
			vectorGC.end(),back_inserter(setIntersectionCG));

#else
    set_intersection(vectorC.begin(), vectorC.end(),vectorGV.begin(),
        vectorGV.end(),back_inserter(setIntersectionCG));
#endif

#else
    set_intersection(vectorC.begin(), vectorC.end(),vectorG.begin(),
        vectorG.end(),back_inserter(setIntersectionCG));
#endif  
    std::vector<int>::iterator m;

    for (m = setIntersectionCG.begin(); m != setIntersectionCG.end(); ++m)
        CG.insert(*m);
#else
	//without point G, to see if coverage of cover-set is not as good
	AG=A;
	BG=B;
	CG=C;
#endif				
    
#ifdef DEBUG_OUTPUT_LEVEL2
	ev << "Node " << this->id << " AG : ";
	printSet(AG);

	ev << "Node " << this->id << " BG : ";
	printSet(BG);

	ev << "Node " << this->id << " CG : ";
	printSet(CG);
#endif 

	// Construct cover-sets by applying AG x BG x CG (x is the cross-product performed on sets)
	SetSetInt agxbg;
	SetInt::iterator ag;
	SetInt::iterator bg;
	SetInt::iterator cg;
	SetInt inter;

	nodeCoverset.clear();

	// first compute AGxBG
	for (ag=AG.begin(); ag!=AG.end(); ag++) {
		for (bg=BG.begin(); bg!=BG.end(); bg++) {
			inter.insert(*ag);
			inter.insert(*bg);
			agxbg.insert(inter);
			inter.clear();
		}
	}
#ifdef DEBUG_OUTPUT_LEVEL2
	// intermediate results
	ev << "AGxBG =======" << endl;
	printSetSet(agxbg);
#endif
	// then compute (AGxBG)xCG
	for (SetSetInt::iterator x=agxbg.begin(); x!=agxbg.end(); x++) {
		for (cg=CG.begin(); cg!=CG.end(); cg++) {
			inter=*x;	
			inter.insert(*cg);
			nodeCoverset.insert(inter);
			inter.clear();			
		}
	}
#ifdef DEBUG_OUTPUT_LEVEL2
	ev << "AGxBGxCG ==== before removing included elements" << endl;
	printSetSet(nodeCoverset);
#endif
	// now need to remove in (AGxBG)xCG the sets that contain the same elements than sets of smaller size
	ListListInt tempvec;

	// put the cover-set in a list of list<int>
	for (SetSetInt::iterator x=nodeCoverset.begin(); x!=nodeCoverset.end(); x++) {
		ListInt t;
		for (SetInt::iterator iy=(*x).begin(); iy!=(*x).end(); iy++)
			t.push_back(*iy);
		tempvec.push_back(t);
	}

	for(ListListInt::iterator ix=tempvec.begin();ix!=tempvec.end();ix++){
		for(ListListInt::iterator iy=tempvec.begin();iy!=tempvec.end();iy++){
			if((*ix)!=(*iy)){
				for(ListInt::iterator iz=(*ix).begin();iz!=(*ix).end();iz++);
				    for(ListInt::iterator iz=(*iy).begin();iz!=(*iy).end();iz++);
				    if(isIncluded(*ix,*iy)){
				        iy=tempvec.erase(iy);
				        iy--;
				    }
			}
		}
	}
	
	nodeCoverset.clear();
	// put back in the coverset
	for (ListListInt::iterator ix=tempvec.begin();ix!=tempvec.end();ix++) {
		SetInt s;
		for (ListInt::iterator iy=(*ix).begin(); iy!=(*ix).end(); iy++)
			s.insert(*iy);
		nodeCoverset.insert(s);
	}

#ifdef DEBUG_OUTPUT_LEVEL1
	ev << "AGxBGxCG ==== when removing included elements" << endl;
	printSetSet(nodeCoverset);
#endif

#endif // COVERAGE_WITH_ONLY_G

#ifdef DEBUG_OUTPUT_LEVEL0

	if (this->forceNumberOfCoversets >= 0) {
		ev_v << "Node " << this->id << ": was forced with cover-sets of size " << this->forceNumberOfCoversets << endl;
	}
	else {
		ev_v << "Node " << this->id << ": built it's cover-sets of size " << nodeCoverset.size() << endl;
	}
		
	ev_v << "Node " << this->id << ": built it's cover-sets of size " << nodeCoverset.size() << endl;
#endif      
	// to get the mean of all the percentage of coverage for a sensor node  
    double meanPercentageCoverage=0.0;

	setCaptureRate(this->getCaptureRate()); 

#ifdef CRITICALITY_WITH_BEZIER	
	if (this->isActive) {
		if (this->imageCaptureEv->isScheduled())
			cancelEvent(this->imageCaptureEv);

                // take care of timing because we need to recompute the time of next image capture according to timing constraints
                double timeToWait = this->timeForImageCapture + this->timeForImageProcessing;

                timeToWait = (timeToWait > 1.0/this->myCaptureRate)? timeToWait : 1.0/this->myCaptureRate;

                // schedule the next capture, depending on the capture rate
                scheduleAt(simTime()+uniform(0,1)+timeToWait, this->imageCaptureEv);
	}
#endif

	// if no other nodes cover the current sensor's FoV then the sensor node must be active
    if (nodeCoverset.size()==0) {
        
#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << ": MUST remain active" << endl;
#endif
		// force a status change
		this->canChangeStatus=true;
		setActiveStatus(true);                 
    }
	else {
		// compute the percentage of coverage of each cover-set
		// try validate the method of cover-set construction with abcg points
#ifdef DEBUG_OUTPUT_LEVEL1
    	ev << "Node " << this->id << ": computes perc. coverage/cover-set" << endl; 
#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "=================================" << endl;
#endif        
#endif        
		int coversetIndex=0;

		for (SetSetInt::iterator x=nodeCoverset.begin(); x!=nodeCoverset.end(); x++, coversetIndex++) {
			SetInt a=*x;

#ifdef COVERAGE_COMPUTE_PERCOVERSET
			int nbPointCovered=0;

			// sample lots of random point and check if the sensor of the cover-set cover this point
			for (int k=1; k<=50000; k++) {						
				// get a random point ('Point' structure) that belong to the sensor FoV
				Point randomP=sampleMyTriangle();
#ifdef DEBUG_OUTPUT_LEVEL3
				// only the first point, just to check
				if (k==1)
				cout << "*** Node " << this->id << " Triangle (" << 
					p.x << "," << p.y << ")" << "(" <<
					b.x << "," << b.y << ")" << "(" <<
					c.x << "," << c.y << ")" << "*** (" <<
					myFoV.v1.get_x() << "," << myFoV.v1.get_y() << ")" << "(" <<
					myFoV.v2.get_x() << "," << myFoV.v2.get_y() << ")" << "(" <<
					myFoV.v3.get_x() << "," << myFoV.v3.get_y() << ")" <<
					"random point (" << P0.x << "," << P0.y << ")" << endl;
#endif
				// if at least one sensor in the cover-set covers that point, it's done
				for (SetInt::iterator i=a.begin(); i!=a.end(); i++) {
					if (isCoveredBy(randomP, *i)) {
						nbPointCovered++;
						break;
					}
				}						
			}
			//divide by 500 (i.e. 50000/100) to get in percentage
			double percentageCoverageCoverSet=(double)nbPointCovered/500.0;
#else
			// just a fake percentage if we are not interested in computing the percentage of coverage of each cover set
			double percentageCoverageCoverSet=-1;
#endif
#if defined(DEBUG_OUTPUT_LEVEL2) && defined(COVERAGE_COMPUTE_PERCOVERSET)
			printSet(a);
    		ev << "This cover-set covers " << nbPointCovered << " points => " <<  
				percentageCoverageCoverSet << "%" << endl; 
#endif        	
			// put the cover-set in a map to associate the percentage of coverage.
			// maybe for future usage
			coveragePerCoversetMap.insert(pair <SetInt,double>(a,percentageCoverageCoverSet));

			// also put the cover-set in an array of coversetElem
			coversetTable[coversetIndex].aCoverset=a;
			coversetTable[coversetIndex].coversetSize=a.size();
			coversetTable[coversetIndex].isActive=false;
			coversetTable[coversetIndex].isDead=false;
			coversetTable[coversetIndex].percentageCoverage=percentageCoverageCoverSet;
			
			// compute the mean percentage coverage over all the coverset
			meanPercentageCoverage += percentageCoverageCoverSet;
		}
#ifdef DEBUG_OUTPUT_LEVEL3
		// just to test the map structure, not needed for the moment as the coverset are stored in
		// a set of set<int>
		ev << "=================================" << endl;
		ev << "contents of the map of cover-sets" << endl;
		for (map<SetInt,double>::iterator it=coveragePerCoversetMap.begin() ; 
				it != coveragePerCoversetMap.end(); it++ ) {
			printSet((*it).first);
#ifdef COVERAGE_COMPUTE_PERCOVERSET
			ev << "has percentage of coverage of " << (*it).second << "%" << endl;
#endif
		}
#endif
#ifdef DEBUG_OUTPUT_LEVEL2
		ev << "=================================" << endl;
		ev << "contents of the cover-sets table" << endl;
		for (unsigned int i=0; i<nodeCoverset.size(); i++ ) {
			printSet(coversetTable[i].aCoverset);
#ifdef COVERAGE_COMPUTE_PERCOVERSET
			ev << "has percentage of coverage of " << coversetTable[i].percentageCoverage << "%" << endl;
#endif
		}
		ev << "=================================" << endl;
#endif 		
	}

	recordScalar("info.position.x", this->p.x);
	recordScalar("info.position.y", this->p.y);
	recordScalar("info.topology.nb_neighbors", this->neighbors.size());

	if (nodeCoverset.size()==0) {
		recordScalar("info.coverset.mean_numberof_coverset_0", 0);
		recordScalar("info.coverset.mean_coverage_coverset_0", 0);

		meanPercentageCoverage = 0;
	}
	else {
		recordScalar("info.coverset.mean_numberof_coverset_X", nodeCoverset.size());
		CoordNode->meanNumberCoverset.collect(nodeCoverset.size());	
		meanPercentageCoverage = meanPercentageCoverage/nodeCoverset.size();
		// record for each sensor node the meanPercentageCoverage
		recordScalar("info.coverset.mean_coverage_coverset_X", meanPercentageCoverage);
		// use a cDoubleHistogram class to collect the meanPercentageCoversetCoverage of each sensor
		// in order to get the mean (for those sensor with cover-set) for the entire network
		CoordNode->meanPercentageCoversetCoverage.collect(meanPercentageCoverage);	
	}

#ifdef CASTALIA
	collectOutput("Coverset stats","Neighbors", this->neighbors.size());
	collectOutput("Coverset stats","Size", nodeCoverset.size());
	collectOutput("Coverset stats","%cov", meanPercentageCoverage);
#endif

#ifdef COVERAGE_STATS_ONLY
	// should end the simulation right after the coverset and their percentage of coverage have been 
	// computed. Actually, let the coordinator node ends the simulation by indicating that the node is dead
	setActiveStatus(false,true);
#else
    // sends status to all neighbors after having determined the cover sets
	sendActivity();
	
#ifdef CRITICALITY_ADAPTATIVE_MAC
	sendCoverSetsAndCaptureRate();
#endif
	
#endif
}

int VideoSensorNode::getNeighborIndexById(int theId) {

	if (this->neighbors.size() != this->nbNeighbors) {
		ev_v << "Node " << this->id << " <-- ERROR: mismatch neighbor size" << endl;
		recordScalar("error.mismatch_neighbor_size", simTime());
	}

	for (unsigned int i=0; i<this->neighbors.size(); i++)
        if (theId == this->neighbors[i].id)
            return i;
	return -1;
}

#ifdef CASTALIA
// for creating command packet to the routing layer
void VideoSensorNode::setSinkInfoInRoutingLayer(int id, int x, int y) {
#ifdef GPSR_ROUTING
	GpsrRoutingControlCommand *cmd = new GpsrRoutingControlCommand("GPSR set sink pos", NETWORK_CONTROL_COMMAND);
	cmd->setGpsrRoutingCommandKind(SET_GPSR_SINK_POS);
	cmd->setDouble1(x);
	cmd->setDouble2(y);
	cmd->setInt1(id);
	toNetworkLayer(cmd);
#endif
}

const char* VideoSensorNode::getSinkAddressString() {

	// if recipientAddress=="NO" then we are in a pro-active routing protocol such as MPRingsRouting (Castalia)
	// or OLSR for instance therefore we send the IDREPORT to the default sink by using the SINK_NETWORK_ADDRESS mechanism
	if (this->recipientAddress.compare("NO") == 0)
		return NULL;
	else {
		// we are specifically indicating a nextRecipient so we assume that we have an on-demand routing protocol such
		// as AODV or GPSR for instance

		// we give the nextRecipient's position in case we have a geographic routing layer
		noeud* nextRecipientInfo = CoordNode->getNodeInfo(atoi(this->recipientAddress.c_str()));

		if (nextRecipientInfo) {
			// send command to the routing layer
			setSinkInfoInRoutingLayer(nextRecipientInfo->id, nextRecipientInfo->p.x, nextRecipientInfo->p.y);
		}
		else
			ev_v << "Node " << this->id << ": can find sink's geo information. Sink is unchanged." << endl;

		return this->recipientAddress.c_str();
	}
}

#endif

int VideoSensorNode::computeActiveCoverset(int& tmpCoversetIndex) {

	int numActiveCoverset=0;

	// try to find a cover set given the set of active nodes
	for (unsigned int i=0; i<nodeCoverset.size(); i++ ) {
		if ( includes(activeNeighbors.begin(),activeNeighbors.end(),
			coversetTable[i].aCoverset.begin(),coversetTable[i].aCoverset.end()) && coversetTable[i].isDead==false) {
			// take the first that is not dead that matches
			// could be that it is the same active set than previously
			if (tmpCoversetIndex==-1)
				tmpCoversetIndex=i;
			this->coversetTable[i].isActive=true;
			numActiveCoverset++;
		}
		else
			this->coversetTable[i].isActive=false;
	}

	return numActiveCoverset;
}

int VideoSensorNode::getNumberPossibleCoverset() {

	int n=0;
	// return the number of possible cover sets, cover set can be inactive but must not be dead.
	for (unsigned int i=0; i<nodeCoverset.size(); i++ )
		if ( !this->coversetTable[i].isDead )
			n++;
	return n;
}

float VideoSensorNode::distance(Point M, Point N) {
    return sqrt((M.x - N.x)*(M.x - N.x) + (M.y - N.y)*(M.y - N.y));
}

// test whether point P is covered by node id
bool VideoSensorNode::isCoveredBy(Point& P, int id) {

	unsigned int index;
	for (index = 0; index<this->neighbors.size(); index++)
		if (this->neighbors[index].id==id) break;

	if (P.is_inside(this->neighbors[index].fov[0]))
		return true;

	return false;
}

// Send position packets to all neighbors: this is a packet that defines the sensor's own position 
// and camera direction
void VideoSensorNode::sendPosition(bool updating) {
#ifdef DEBUG_OUTPUT_LEVEL2
	if (updating)
		ev_v << "Node " << this->id << " --> UPDATE_POS(" << this->p.x << "," << this->p.y << ") to neighbors" << endl;
	else
		ev_v << "Node " << this->id << " --> POSITION(" << this->p.x << "," << this->p.y << ") to neighbors" << endl;
#endif
	PositionPacket *pkt;

#ifndef CASTALIA
	for (int i=0; i<this->gateSize("out"); i++) {
#endif
		if (updating)
			pkt = new PositionPacket("update_pos", APPLICATION_PACKET);
		else
			pkt = new PositionPacket("position", APPLICATION_PACKET);

		pkt->setId(this->id); // BYTE_T*2
		pkt->setPx(this->p.x); // FLOAT_T
		pkt->setPy(this->p.y); // FLOAT_T
		pkt->setVx(this->v.x); // FLOAT_T
		pkt->setVy(this->v.y); // FLOAT_T
                pkt->setDov(this->dov[0]);
		pkt->setAlpha(this->alpha[0]); // FLOAT_T
		pkt->setActive(false); // BYTE_BIT_FIELD_T 1

		pkt->setIsMobile(this->isMobile);
		pkt->setIsCamRotatable(this->isCamRotatable);

		if (updating) {
			pkt->setIsMoving(this->isMoving); // BYTE_BIT_FIELD_T 2
			pkt->setStartMobilityTime(this->startMobilityTime); // FLOAT_T
			pkt->setMobilitySpeed(this->mobilitySpeed); // FLOAT_T
			pkt->setMobilityDuration(this->mobilityDuration); // FLOAT_T
			pkt->setIsCamRotating(this->isCamRotating); // BYTE_BIT_FIELD_T 3
			pkt->setStartCamRotationTime(this->startCamRotationTime); // FLOAT_T
			pkt->setCamRotationTime(this->camRotationTime); // FLOAT_T
			pkt->setCamRotationCount(this->camRotationCount); // BYTE_T
		}
#ifdef CASTALIA
		int payloadSize=BYTE_T*2+FLOAT_T*5+BYTE_BIT_FIELD_T;

		if (updating)
			payloadSize+=FLOAT_T+FLOAT_T+FLOAT_T+FLOAT_T+FLOAT_T+BYTE_T;
		pkt->setByteLength(payloadSize);

		// now send it to the network layer
		toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);

		if (updating)
			collectOutput("AppPacket sent", "Update pos msg");
		else
			collectOutput("AppPacket sent", "Position msg");
#else
        send(pkt,"out",i);
    }
#endif
}

// Notify to all neighbors the sensor's activity status
void VideoSensorNode::sendActivity() {
// chaotic scheduling, we do not sent activity status
#ifndef SCHEDULE_PROB
#ifdef DEBUG_OUTPUT_LEVEL2
	ev_v << "Node " << this->id << " --> ACTIVITY(" << this->isActive << ") to neighbors" << endl;
#endif 
	
#ifndef CASTALIA
	for (int i=0; i<this->gateSize("out");i++) {
#endif
		ActivityPacket *pkt = new ActivityPacket("activity", APPLICATION_PACKET);

		pkt->setId(this->id); // BYTE_T*2
		pkt->setActive(this->isActive); // BYTE_BIT_FIELD_T

#ifdef CASTALIA
		int payloadSize=BYTE_T*2+BYTE_BIT_FIELD_T;
		pkt->setByteLength(payloadSize);

		// now send it to the network layer
		toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);

		collectOutput("AppPacket sent", "Activity msg");
#else
    	send(pkt,"out",i);
	}
#endif
	this->nbActivityMsgSent++;
#endif
}

#ifdef CASTALIA

#ifdef CRITICALITY_ADAPTATIVE_MAC

//Broadcast coversets and capture rate
void VideoSensorNode::sendCoverSetsAndCaptureRate() {

	CoversAndCaptureRatePacket *pkt = new CoversAndCaptureRatePacket("coversets and capturerate", APPLICATION_PACKET);

	pkt->setId(this->id);
	pkt->setCapturingRate(this->myCaptureRate);

	if (this->forceNumberOfCoversets >= 0) {
		pkt->setNoOfCoverSets(this->forceNumberOfCoversets);
	}
	else {
		pkt->setNoOfCoverSets(this->nodeCoverset.size());
	}
	
	int payloadSize=BYTE_T*2+FLOAT_T;
	pkt->setByteLength(payloadSize);

	// now send it to the network layer
	toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);
}
#endif

int VideoSensorNode::getCoversetIndex() {
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
	return bestCoversetSelection();
#else
	// for the moment, get the first active coverset
	for (unsigned int i=0; i<nodeCoverset.size(); i++ ) {

			if (this->coversetTable[i].isActive)
					return i;
			}

	return -1;
#endif
}

// Notify to all neighbors which coverset should be active
void VideoSensorNode::sendCoversetActivation() {

#ifdef DEBUG_OUTPUT_LEVEL0
	ev << "Node " << this->id << ": has active neighbors:" << endl;
	printSet(this->activeNeighbors);

	ev << "Node " << this->id << ": updates its active coverset list" << endl;
#endif

	int tmpCoversetIndex=-1;

	// update the list of active coverset
	computeActiveCoverset(tmpCoversetIndex);

	// get an active coverset, here we could change the way a coverset is selected
	int selectedCoversetIndex=getCoversetIndex();

	if (selectedCoversetIndex==-1)
		ev_v << "Node " << this->id << ": CANNOT ACTIVATE COVERSET, NO ACTIVE COVERSET" << endl;
	else {
		SetInt::iterator nid;
		unsigned int j=0;

#ifdef DEBUG_OUTPUT_LEVEL0
		ev_v << "Node " << this->id << " --> COVERSET ACTIVATION INDEX " << selectedCoversetIndex << endl;
		ev_v << "SetSize = " << coversetTable[selectedCoversetIndex].aCoverset.size() << "\t(";
#endif

		ActivityPacket *pkt = new ActivityPacket("coverset activation", APPLICATION_PACKET);
		pkt->setId(this->id); // BYTE_T*2
		pkt->setActive(this->isActive); // BYTE_BIT_FIELD_T
		pkt->setNeighborIdListArraySize(coversetTable[selectedCoversetIndex].aCoverset.size()); // BYTE_T*2*aCoverset.size()

		for (nid=coversetTable[selectedCoversetIndex].aCoverset.begin(); nid!=coversetTable[selectedCoversetIndex].aCoverset.end(); ) {
			pkt->setNeighborIdList(j,*nid);
#ifdef DEBUG_OUTPUT_LEVEL0
			ev << *nid;
			if (nid!=coversetTable[selectedCoversetIndex].aCoverset.end()) ev << ", ";
#endif
			j++;
			nid++;
		}

#ifdef DEBUG_OUTPUT_LEVEL0
		ev << ")" << endl;
#endif

		int payloadSize=BYTE_T*2+BYTE_BIT_FIELD_T+BYTE_T*2*coversetTable[selectedCoversetIndex].aCoverset.size();

		pkt->setByteLength(payloadSize);

		// now send it to the network layer
		toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);

		collectOutput("AppPacket sent", "Cover activation msg");
		collectOutput("Intrusion breakdown", "Cover activation msg sent");
	}
}
#endif

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
// Notify to all neighbors the alert
void VideoSensorNode::sendAlert(double theCriticalityLevel, int theId=-1, int theSeq=-1) {

#ifdef DEBUG_OUTPUT_LEVEL2
	ev_v << "Node " << this->id << " --> ALERT(" << theCriticalityLevel << ") to neighbors" << endl;
#endif

	this->nbAlertMsgSent++;
	recordScalar("info.criticality.send_alert", simTime());

#ifndef CASTALIA
	for (int i=0; i<this->gateSize("out");i++) {
#endif
		ActivityPacket *pkt = new ActivityPacket("alert", APPLICATION_PACKET);

		pkt->setId(this->id); // BYTE_T*2
		pkt->setActive(this->isActive); // BYTE_BIT_FIELD_T 1
		pkt->setCriticalityLevel(theCriticalityLevel); // FLOAT_T

#ifdef CRITICALITY_ADAPTATIVE_MAC
		pkt->setAlertNumber(this->nbAlertMsgSent);
#endif

		if (theId==-1) {
			pkt->setOriginal(true); // BYTE_BIT_FIELD_T 2
			pkt->setIdFirstSender(this->id); // BYTE_T*2
			pkt->setSeq(this->intrusionViewCount); // BYTE_T
		}
		else {
			// put back the original field in the new alert msg
			pkt->setOriginal(false);
			pkt->setIdFirstSender(theId);
			pkt->setSeq(theSeq);
		}

#ifdef CASTALIA
		int payloadSize=BYTE_T*2+FLOAT_T+BYTE_BIT_FIELD_T+BYTE_T*2+BYTE_T;

		pkt->setByteLength(payloadSize);

		// now send it to the network layer
		toNetworkLayer(pkt, BROADCAST_NETWORK_ADDRESS);

		collectOutput("AppPacket sent", "Alert msg");
#else 	
		send(pkt,"out",i);
	}
#endif
	this->hasAlertedNeighbors=true;
	
#ifdef CRITICALITY_ADAPTATIVE_MAC
	CoordNode->countAlertNumber(this->id);
	trace() << " node " << this->id << " --> sending alert msg, alert/response, and it's neighbor table size is " << this->neighbors.size()  << " :\n" ;
#endif
}

void VideoSensorNode::setAlertCriticalityLevel(double theCriticalityLevel) {
		
	this->criticalityLevel=theCriticalityLevel;
	
	if (!this->isAlerted) {
		this->isAlerted=true;
#ifdef DEBUG_OUTPUT_LEVEL2
    	ev << "Node " << this->id << ": at ALERT criticality level " << this->criticalityLevel << endl;
#endif 
		if (ev.isGUI())
			bubble("ALERT");
		// actually, if toNormalCriticalityLevelEv is scheduled then already in alert mode
		if (!this->toNormalCriticalityLevelEv->isScheduled()) {
			scheduleAt(simTime()+this->maxCriticalityLevelPeriod, this->toNormalCriticalityLevelEv);
		}
	}
}

#endif

// mainly captures an image with the camera, check if an intrusion can be seen and decrease the energy level
// NOTE: the energy level is currently managed inside this component, even with Castalia that provides a
// ResourceGenericManager module. 
bool VideoSensorNode::imageCapture()
{
#ifdef FUNCTION_TRACE_DEBUG
	TRACE_DEBUG_IN("imageCapture()");
#endif

	double extraConsumedEnergy=0.0;
	double theForcedIntrusionProb;
	bool seenIntrusion=false;
#ifdef CASTALIA
	// only if it is an image capture that has not been forced by the imageCountOnIntrusionCounter
	if (this->imageCountOnIntrusionCounter==0)
		theForcedIntrusionProb=uniform(0,1);
	else
		theForcedIntrusionProb=1.0;

        collectOutput("Intrusion breakdown", "image capture");
#endif

        this->imageCaptureCount++;

#ifdef WITH_INTRUSION
#ifdef CASTALIA
    cModule *intrusionMod = getParentModule()->getParentModule()->getSubmodule("intrusion");
#else
    cModule *intrusionMod = getParentModule()->getSubmodule("intrusion");
#endif
    intrusionModule* intrusionM = check_and_cast<intrusionModule *>(intrusionMod);

	Triangle theFoV[MAX_NB_CAMERA];

	// is the camera currently rotating?
	// no support for multiple camera if we have rotating camera, only take camera index 0
	if (this->isCamRotating) {
		// we emulate a camera rotation by computing the new fov position
		// based on the startCamRotationTime and the maxCamRotationTime
		Point v1,b1,c1;
		double newLineOfSight, currentRotationCount, diffRotationCount;

		// determine the number of rotations based on the maxCamRotationTime time and the startCamRotationTime
		// since the beginning of the rotation, i.e. normally between 0.0 and this->maxRotationCount
		currentRotationCount=(SIMTIME_DBL(simTime())-this->startCamRotationTime)/this->maxCamRotationTime;

		// determine the rotation count/diff since the last capture
		diffRotationCount=currentRotationCount-this->lastRotationCount;

		// determine the amount of energy spent for this rotation
		// assuming that 1 complete rotation consumes 5 times an image capture energy consumption
		extraConsumedEnergy+=this->energyPerCapture*5*diffRotationCount;

		this->lastRotationCount = currentRotationCount;

		// find the new line of sight
		newLineOfSight=this->lineOfSight+currentRotationCount * 2 * PI;

		v1.x=this->dov[0]*cos(newLineOfSight)+this->p.x;
		v1.y=this->dov[0]*sin(newLineOfSight)+this->p.y;
	
		// find the new coordinates of the sensor's fov
		this->findTriangleCoord(this->p, v1, b1, c1, this->dov[0], this->alpha[0]);
		// use the virtual FoV
		theFoV[0]=Triangle(p,b1,c1);
	}
	else {
		for (int i=0; i<this->nbCamera; i++)
			theFoV[0]=this->myFoV[0];
	}

	if (intrusionM->intrusionOn) {
#ifdef INTRUSION_POLYGON
		// clear the table that records the points that are seen
		intrusionM->clearPolygonPointTable();
		int myCount=intrusionM->isPolygonInside(theFoV);
		// try to find for each cover set the number of point of the polygon that are seen
		if (myCount) {
			int count;
			// sets it high so that we can determine the min
			int minPolyPoint=1000;
			int maxPolyPoint=0;
			// loops on all the cover sets
	  		for (int i=0; i<nodeCoverset.size(); i++ ) {
				count=0;
				int id;
				// clear the table again
				intrusionM->clearPolygonPointTable();
				//loops on all the sensors of the cover set
				for (SetInt::iterator ix=coversetTable[i].aCoverset.begin(); 
						ix!=coversetTable[i].aCoverset.end(); 
						ix++) {
					int index;
					id=*ix;
					// *it carries the id of the sensor
					// find in the neighbors table the sensor with id=*it
					for (index = 0; index<this->neighbors.size(); index++)
						if (this->neighbors[index].id==id) break;
					// how many different points of the intrusion polygon the sensor is seeing
					count = intrusionM->isPolygonInside(this->neighbors[index].fov);	
				}
				coversetTable[i].nbPointPolygonSeen=count;
				if (count > maxPolyPoint) 
					maxPolyPoint = count;

				if (count < minPolyPoint) 
					minPolyPoint = count;
			}
			// only if there are cover set
			if (nodeCoverset.size()) {
				intrusionM->nbPolyPointOriginal.record(myCount);
				intrusionM->nbMinPolyPoint.record(minPolyPoint);
				intrusionM->nbMaxPolyPoint.record(maxPolyPoint);
			}
			if (ev.isGUI())			
				bubble("SEE");
			intrusionM->intrusionHit(this->id);
			//increase the local intrusion counter for per sensor statistics
			intrusionViewCount++;
		}
#else

		if (this->cameraSimultaneous) {
			for (int i=0; i<this->nbCamera; i++) {
				if (intrusionM->p.is_inside(theFoV[i])) {
					seenIntrusion=true;
					break;
				}
			}
		}

		if (this->cameraCycling) {
			if (intrusionM->p.is_inside(theFoV[this->nextActiveCamera]))
				seenIntrusion=true;
			this->nextActiveCamera = (this->nextActiveCamera + 1) % this->nbCamera;
		}
		else {
			int randCamera=rand() % this->nbCamera;

			if (intrusionM->p.is_inside(theFoV[randCamera]))
				seenIntrusion=true;
		}

#ifdef CRITICALITY_ADAPTATIVE_MAC
		if (seenIntrusion || theForcedIntrusionProb < this->forcedIntrusionProb) {
#else
		if (seenIntrusion) {
#endif

#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": INTRUSION SEEN" << endl;
#endif
			bool occlusion = CoordNode->obstacleList.intersects(Segment(this->p, intrusionM->p));

			if (occlusion) {
				if (ev.isGUI())
					bubble("OCC");
				intrusionOccludedCount++;
				recordScalar("info.cam.intrusion_occluded", simTime());
			}
			else {

				if (ev.isGUI())
					bubble("SEE");

				if (this->isCamRotating)
					recordScalar("info.cam.intrusion_seen_by_rotating_cam", simTime());
				else
					recordScalar("info.cam.intrusion_seen", simTime());

				// the sensor sees the intrusion
				intrusionM->intrusionHit(this->id);

				//increase the local intrusion counter for per sensor statistics
				intrusionViewCount++;
#ifdef CASTALIA
				collectOutput("Intrusion breakdown", "seen");
#endif
				// to further trigger the sending of an image
				// only if it is a new intrusion when we have sent all the requested images
				if (this->imageCountOnIntrusionCounter==0)
					this->imageCountOnIntrusionCounter = this->imageCountOnIntrusion;

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
				// since the node sees the intrusion, it increases its criticality level to better detect other intrusions...
				setAlertCriticalityLevel(this->maxCriticalityLevel);
				// ... and sends an alert to neighbors
				// only send alert to neighbors if not already in alert mode, reinforcement will be manage later on
				if (!this->hasAlertedNeighbors) {
					sendAlert(this->maxCriticalityLevel);
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
					// schedule when reinforcement will be enabled
					if (!this->canReinforceEv->isScheduled()) {
						scheduleAt(simTime()+DYNAMIC_REINFORCEMENT_DEAF_PERIOD, this->canReinforceEv);
					}
#endif
				}
				// TODO: when the node that senses the intrusion can rotate its camera, it may do so
				// and indicates this to its neighbors, not implemented now. Only nodes that gets alerted do so.
				if (this->isCamRotatable) {
#ifdef ROTATECAM_ON_INTRUSION_DETECTION
					// probably starts the camera rotation, program when it stops rotating, in which case
					// get back to the initial position
					// can certainly compute the time needed to perform 2 complete rotations
					// each time the sensor takes a picture, compute the new fov according to the rotationSpeed and the
					// startRotationTime to see if it can see intrusions
					// have to set the working parameters first
					// sendPosition(true);
#endif
				}

#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
				// if I can reinforce, then notify neighbors even if it has already been done
				// actually, can reinforce after an initial DEAF_PERIOD until the maxCriticalityLevelPeriod is over
				if (this->canReinforce) {
					this->nbReinforcementMsgSent++;
					sendAlert(this->maxCriticalityLevel);
				}
#endif
#endif
			}
		}
#endif
	}

	if (theForcedIntrusionProb < this->forcedIntrusionProb) {
		// we artificially set the intrusion flag
		// only if it is a new intrusion when we have sent all the requested images
		if (this->imageCountOnIntrusionCounter==0)
			this->imageCountOnIntrusionCounter = this->imageCountOnIntrusion;
	}

#ifdef CASTALIA
	if ( (this->imageCountOnIntrusionCounter && this->activateCoversetOnIntrusion)
		|| (this->receivedCoversetActivation && this->propagateCoversetActivation)
		|| (this->forcedCoversetActivationAt>0 && SIMTIME_DBL(simTime())>this->forcedCoversetActivationAt) ) {

		sendCoversetActivation();

		// only once
		if (SIMTIME_DBL(simTime())>this->forcedCoversetActivationAt)
			this->forcedCoversetActivationAt=-1.0;
	}
#endif
#else // WITH_INTRUSION is not defined
#ifdef CASTALIA
	// only check whether we are forced to send a coverset activation
	if ( (this->forcedCoversetActivationAt>0 && SIMTIME_DBL(simTime())>this->forcedCoversetActivationAt) ) {

		sendCoversetActivation();

		// only one time
		if (SIMTIME_DBL(simTime())>this->forcedCoversetActivationAt)
			this->forcedCoversetActivationAt=-1.0;
	}
#endif
#endif	

#ifdef CASTALIA

	// energy consumption will be computed for each image
	double tmpEnergy;
	tmpEnergy=this->measuredEnergyPerImageCapture-this->imageNodeBaselineConsumption*this->timeForImageCapture;

	// remaining consumed energy for taking/capturing/doing intrusion detection for one image (sync cam & config cam)
	resMgrModule->consumeEnergy(tmpEnergy);
	// we chose to keep the stats separated, at the end of simulation, we will computed the real part of the baseline consumption
	collectOutput("Consumed AppEnergy", "5-Capture", this->measuredEnergyPerImageCapture);

	this->appConsumedEnergy += this->measuredEnergyPerImageCapture;

	tmpEnergy=this->measuredEnergyPerImageProcessing-this->imageNodeBaselineConsumption*this->timeForImageProcessing;

	// consumed energy for processing one picture (polling data from cam)
	resMgrModule->consumeEnergy(tmpEnergy);
	// we chose to keep the stats separated, at the end of simulation, we will computed the real part of the baseline consumption
        collectOutput("Consumed AppEnergy", "6-Processing", this->measuredEnergyPerImageProcessing);

	this->appConsumedEnergy += this->measuredEnergyPerImageProcessing;

#ifdef DEBUG_OUTPUT_LEVEL3
	ev_v << "Node " << this->id << ": has consumed " << resMgrModule->getSpentEnergy() << endl;
#endif

	if ((this->imageCountOnIntrusionCounter && this->sendImageOnIntrusion)
		|| this->forceSendImageCounter
		|| this->imageCountOnCoversetActivationCounter) {

		// get either NULL in which case it is the default sink string which is "SINK" for pro-active routing layers
		string theSink=(getSinkAddressString())?this->recipientAddress:SINK_NETWORK_ADDRESS;

		int imageType=INTRUSION_IMG_TYPE;

		if (this->forceSendImageCounter) {

			imageType=FORCED_IMG_TYPE;
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": SEND FORCED IMAGE" << endl;
#endif
			collectOutput("Image sent", "forced");
			this->forceSendImageCounter--;
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": " << this->forceSendImageCounter
					<< " FORCED IMAGE(S) REMAINING" << endl;
#endif
		}

		if (this->imageCountOnCoversetActivationCounter) {

			imageType=COVERSET_ACTIVATION_IMG_TYPE;
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": SEND IMAGE BY COVERSET ACTIVATION" << endl;
#endif
			collectOutput("Image sent", "by coverset activation");
			this->imageCountOnCoversetActivationCounter--;
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": " << this->imageCountOnCoversetActivationCounter
					<< " IMAGE(S) BY COVERSET ACTIVATION REMAINING" << endl;
#endif
		}

		if (this->imageCountOnIntrusionCounter) {
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": SEND INTRUSION IMAGE" << endl;
#endif
			this->imageCountOnIntrusionCounter--;
#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": " << this->imageCountOnIntrusionCounter << " INTRUSION IMAGE(S) REMAINING" << endl;
#endif
			if (theForcedIntrusionProb < this->forcedIntrusionProb) {
				imageType=RANDOM_INTRUSION_IMG_TYPE;
				collectOutput("Image sent", "on random intrusion");
				collectOutput("Intrusion breakdown", "random");
			}
			else
				collectOutput("Image sent", "on intrusion");
		}
		// the energy consumption for encoding will be taken into account by sendImage
		sendImage(theSink.c_str(), imageType);

		// for the moment, consider that sending 1 image cost 5 times the energy for capturing an image
		extraConsumedEnergy+=this->energyPerCapture*5;
	}
#endif // CASTALIA

	// the node is capturing an image, so decrease the energy level
	this->energyLevel = this->energyLevel - this->energyPerCapture - extraConsumedEnergy;
							
	if (this->energyLevel <= 0 && !this->isDead) // just to avoid broadcasting several time a dead status
		setActiveStatus(false,true);

	return(seenIntrusion);

#ifdef FUNCTION_TRACE_DEBUG
	TRACE_DEBUG_OUT("imageCapture()");
#endif
}

#ifdef CASTALIA

void VideoSensorNode::sendImage(const char *sink, int imageType=INTRUSION_IMG_TYPE) {

	IdReportData tmpData;
	tmpData.nodeID = this->id;
	tmpData.locX = p.x;
	tmpData.locY = p.y;

	int nb_packets=0;

	double tmpEnergy=this->measuredEnergyPerImageEncoding-this->imageNodeBaselineConsumption*this->timeForImageEncoding;

	// remaining consumed energy for encoding
	resMgrModule->consumeEnergy(tmpEnergy);
	// we chose to keep the stats separated, at the end of simulation, we will computed the real part of the baseline consumption
	// consumption for encoding the image, in preparation for its transmission
	collectOutput("Consumed AppEnergy", "7-Encoding", this->measuredEnergyPerImageEncoding);

	this->appConsumedEnergy += this->measuredEnergyPerImageEncoding;

	// the sent time of the current image
	this->imageSentTime = SIMTIME_DBL(simTime());

	string image = this->imageFilename ;

#ifdef SEND_REAL_IMAGE

	if (this->cranEncoding) {

		FILE* pFile;

		pFile = fopen(image.c_str(), "r");

		if (pFile) {
			int chunkSize;
			unsigned char theValue;
			int res;
			unsigned int storedNbPackets, imageSizeX, imageSizeY, qualityFactor;

			fscanf(pFile, "%4X", &storedNbPackets);
			fscanf(pFile, "%4X", &imageSizeX);
			fscanf(pFile, "%4X", &imageSizeY);
			fscanf(pFile, "%4X", &qualityFactor);

			do {
				// get the size of the packet
				res=fscanf(pFile, "%4X", &chunkSize);

				if (res==EOF)
					continue;

				ImageDataPacket *packet2Net = new ImageDataPacket("image", APPLICATION_PACKET);
				packet2Net->setExtraData(tmpData); // (BYTE_T*2+FLOAT_T*2)
				packet2Net->setSentTime(this->imageSentTime);
				packet2Net->setImageSN(this->imageSN); // BYTE_T
				packet2Net->setSequenceNumber(nb_packets); // BYTE_T
				packet2Net->setDataSize(chunkSize);
				// don't know it, put CRAN_ENCODING_IMAGE_TYPE so that the receiver knows it is CRAN_ENCODING
				packet2Net->setImageSize(CRAN_ENCODING_IMAGE_TYPE);
				// don't use it
				packet2Net->setImageOffset(-1);
				// don't know it
				packet2Net->setNbPackets(storedNbPackets);
				packet2Net->setImageSizeX(imageSizeX);
				packet2Net->setImageSizeY(imageSizeY);
				packet2Net->setQualityFactor(qualityFactor);
				packet2Net->setImageFilename(const_cast<char *>(image.c_str()));
				packet2Net->setImageBMPOriginalFilename(const_cast<char *>(this->imageBMPOriginalFilename.c_str()));
				packet2Net->setImageDataArraySize(chunkSize);  // BYTE_T*chunkSize
				packet2Net->setImageType(imageType); // BYTE_T

				for (int j=0; j<chunkSize; j++) {
					res = fscanf(pFile, "%2X", &theValue);
					packet2Net->setImageData(j, theValue);
				}

	#ifdef DEBUG_OUTPUT_LEVEL2
				ev_v << "Node " << this->id << " --> REAL IMAGE(" << this->imageSN <<") to node " << sink << " seq=" << nb_packets << endl;
	#endif
				int payloadSize=(BYTE_T*2+FLOAT_T*2)+BYTE_T+BYTE_T+BYTE_T*chunkSize+BYTE_T;

				packet2Net->setByteLength(payloadSize);

				// TODO: there is no extra time for the send() to return, behaves like a multi-threaded system
				toNetworkLayer(packet2Net, sink);

				// count the number of packet for final statistics
				nb_packets++;

			} while (res!=EOF);

#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << " --> REAL IMAGE(" << this->imageSN << ") to node " << sink << ". #packet: "
					<< nb_packets << ", stored #packets: " << storedNbPackets << endl;
#endif
		}
		else {
			ev_v << "Node " << this->id << ": CANNOT LOAD IMAGE (CRAN_ENCODING) " << image << endl;
		}
	}
	else {
		// use default image if it is not specified in the ini file
		if (image.compare("NO")==0)
			image="sensor_images/default.bmp";

		char *raw_image = loadImage(image, &this->imageByteSize);

		// check if file exists
		if (raw_image) {

#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << ": LOAD IMAGE " << image << " of size " << this->imageByteSize << " to node " << sink << endl;
#endif

			nb_packets = ((this->imageByteSize%this->imageChunkSize)==0)? this->imageByteSize/this->imageChunkSize:
						 this->imageByteSize/this->imageChunkSize+1;
			int size_paquet = 0;
			int chunkSize = this->imageChunkSize;

			for (int i=0; i<nb_packets; i++) {

				ImageDataPacket *packet2Net = new ImageDataPacket("image", APPLICATION_PACKET);
				packet2Net->setExtraData(tmpData); // (BYTE_T*2+FLOAT_T*2)
				packet2Net->setSentTime(this->imageSentTime);
				packet2Net->setImageSN(this->imageSN); // BYTE_T
				packet2Net->setSequenceNumber(i); // BYTE_T
				packet2Net->setImageSize(this->imageByteSize);
				packet2Net->setNbPackets(nb_packets);
				packet2Net->setDataSize(this->imageChunkSize); // BYTE_T*chunkSize
				packet2Net->setImageOffset(i*this->imageChunkSize);
				packet2Net->setImageFilename(const_cast<char *>(image.c_str()));
				packet2Net->setImageType(imageType); // BYTE_T

				if ((i == (nb_packets - 1)) && ((this->imageByteSize%this->imageChunkSize) != 0)) {
					chunkSize = this->imageByteSize%this->imageChunkSize;
					packet2Net->setDataSize(chunkSize);
				}

				packet2Net->setImageDataArraySize(chunkSize);

				for(int j=0; j<chunkSize; j++) {
					packet2Net->setImageData(j, raw_image[size_paquet++]);
				}

	#ifdef DEBUG_OUTPUT_LEVEL2
				ev_v << "Node " << this->id << " --> REAL IMAGE(" << this->imageSN <<") to node " << sink << " seq=" << i << endl;
	#endif
				int payloadSize=(BYTE_T*2+FLOAT_T*2)+BYTE_T+BYTE_T+BYTE_T*chunkSize+BYTE_T;

				packet2Net->setByteLength(payloadSize);

				toNetworkLayer(packet2Net, sink);
			}

	#ifdef DEBUG_OUTPUT_LEVEL0
			ev_v << "Node " << this->id << " --> REAL IMAGE(" << this->imageSN <<") to node " << sink << endl;
	#endif
			// release buffer
			if (raw_image)
				delete[] raw_image;
		}
		else {
			ev_v << "Node " << this->id << ": CANNOT LOAD IMAGE " << image << endl;
		}
	}
#else

	nb_packets = this->imageByteSize/this->imageChunkSize;
	int chunkSize = this->imageChunkSize;

	for (int i=0; i<nb_packets; i++) {

		ImageDataPacket *packet2Net = new ImageDataPacket("image", APPLICATION_PACKET);
		packet2Net->setExtraData(tmpData); // (BYTE_T*2+FLOAT_T*2)
		packet2Net->setImageSN(this->imageSN); // BYTE_T
		packet2Net->setSequenceNumber(i); // BYTE_T
		packet2Net->setImageSize(this->imageByteSize);
		packet2Net->setNbPackets(nb_packets);
		packet2Net->setDataSize(this->imageChunkSize); // BYTE_T*chunkSize
		packet2Net->setImageOffset(i*this->imageChunkSize);
		packet2Net->setImageFilename(const_cast<char *>(image.c_str()));
		packet2Net->setImageType(imageType); // BYTE_T

		if ((i == (nb_packets - 1)) && ((this->imageByteSize%this->imageChunkSize) != 0)) {
			chunkSize = this->imageByteSize%this->imageChunkSize;
			packet2Net->setDataSize(chunkSize);
		}

		packet2Net->setImageDataArraySize(this->imageChunkSize);

#ifdef DEBUG_OUTPUT_LEVEL2
		ev_v << "Node " << this->id << " --> FAKE IMAGE(" << this->imageSN <<") to node " << sink << " seq=" << i << endl;
#endif
		int payloadSize=(BYTE_T*2+FLOAT_T*2)+BYTE_T+BYTE_T+BYTE_T*chunkSize+BYTE_T;

		packet2Net->setByteLength(payloadSize);

		toNetworkLayer(packet2Net, sink);
	}

#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << "Node " << this->id << " --> FAKE IMAGE(" << this->imageSN <<") to node " << sink << endl;
#endif

#endif

	// only if we sent something
	if (nb_packets) {
		this->imageSN++;

		collectOutput("Image sent", "Images");
		collectOutput("Image sent", "Packets", nb_packets);
	}
}

#endif

void VideoSensorNode::findTriangleCoord(Point p, Point v, Point& b, Point& c, double dov, double alpha) {
    double h = dov*sin(alpha);
    double d = dov*cos(alpha);

    b.x = v.x + h*(v.y - p.y)/d;
    b.y = v.y - h*(v.x - p.x)/d;

    c.x = v.x - h*(v.y - p.y)/d;
    c.y = v.y + h*(v.x - p.x)/d;
}

// Based on the triangle sampling method found on http://mathworld.wolfram.com/TrianglePointPicking.html
// Basically, use x = a1V1+a2V2; where a1 & a2 are between 0 and 1, and V1 & V2 two vector of the triangle
Point VideoSensorNode::sampleMyTriangle() {

	double point_a1;
	double point_a2;
	bool found=false;
	Point randomP;

	Point AB = myFoV[0].v2-myFoV[0].v1;
#ifdef DEBUG_OUTPUT_LEVEL3
	ev << "** Point AB(" << AB.get_x() << "," << AB.get_y() << ")" << endl;
#endif
	Point AC = myFoV[0].v3-myFoV[0].v1;
#ifdef DEBUG_OUTPUT_LEVEL3
	ev << "** Point AC(" << AC.get_x() << "," << AC.get_y() << ")" << endl;
#endif
	while (!found) {
		point_a1=uniform(0,1); //(double)rand() / (double)RAND_MAX;
		point_a2=uniform(0,1); //(double)rand() / (double)RAND_MAX;
#ifdef DEBUG_OUTPUT_LEVEL3
		ev << a1 << ", " << a2 << ", "; 
#endif
		randomP = point_a1*AB+point_a2*AC;
		randomP = randomP + myFoV[0].v1;
#ifdef DEBUG_OUTPUT_LEVEL3
		ev << "(" << randomP.get_x() << "," << randomP.get_y() << ")" << endl;
#endif
		if (randomP.is_inside(myFoV[0])) {
			found=true;
#ifdef DEBUG_OUTPUT_LEVEL3
			ev << "found" << endl;
#endif
		}
#ifdef DEBUG_OUTPUT_LEVEL3
		else ev << "not found - retry" << endl;
#endif
	}
	return (randomP);
}

#ifdef CASTALIA


double VideoSensorNode::getCastaliaEnergyLevel()
{
	return resMgrModule->getSpentEnergy();
}

void VideoSensorNode::setRadioTxOutputPower(double o)
{
	toNetworkLayer(createRadioCommand(SET_TX_OUTPUT, o));
	connectivityTool.currentTxOutputPower=o;
}

#ifdef CASTALIA_VA
void VideoSensorNode::fromNetworkLayer(ApplicationPacket * genericPacket,
		 const char *source, double rssi, double lqi)
{
	if (isSink)
		ev_v << "Sink received packet" << endl;
}
#else
// Taken from Castalia's VirtualApplication.cc

// A function used to send control messages
void VideoSensorNode::toNetworkLayer(cMessage * msg)
{
	if (msg->getKind() == APPLICATION_PACKET)
		opp_error("toNetworkLayer() function used incorrectly to send APPLICATION_PACKET without destination Network address");
	send(msg, "toCommunicationModule");
}

// A function used to send data packets
void VideoSensorNode::toNetworkLayer(cPacket * pkt, const char *dst)
{
	ApplicationPacket *appPkt = check_and_cast <ApplicationPacket*>(pkt);
	appPkt->getAppNetInfoExchange().destination = string(dst);
	appPkt->getAppNetInfoExchange().source = selfAddress;
	appPkt->getAppNetInfoExchange().timestamp = simTime();
	appPkt->setApplicationID(applicationID.c_str());
	int size = appPkt->getByteLength();
	if (size == 0)
		size = constantDataPayload;
	if (packetHeaderOverhead > 0) size += packetHeaderOverhead;
	appPkt->setByteLength(size);

	// get the delay for being able to send the next packet and set the time when the packet should arrived at the destination
	// because there is communication protocol nor OS overhead in the communication stack of Castalia
	simtime_t sendDelay = this->getTimeBetweenSend(size);
	sendDelayed(appPkt, sendDelay, "toCommunicationModule");

#ifdef DEBUG_OUTPUT_LEVEL0
	ev_v << "Sending [" << appPkt->getName() << "] of size " <<
		size << " bytes to communication layer after " << sendDelay.dbl() << endl;
#endif
}
#endif

simtime_t VideoSensorNode::getTimeBetweenSend(int size)
{
	simtime_t t_delay, t_now;
	double sendTime;
	
	sendTime = (size>90)?this->globalSendTime:this->globalSendTime/2;
	
	//t_next = this->a_time_send*size+this->b_time_send;

	t_now=simTime();

	// we cannot send another packet yet
	if (t_now < this->nextSendTime) {
		t_delay = sendTime+this->nextSendTime-t_now;
		this->nextSendTime += sendTime;
	}
	else {
		// don't have to wait
		t_delay=sendTime;
		this->nextSendTime=t_now+sendTime;
	}

	return t_delay;
}

#endif

#ifdef CASTALIA
void VideoSensorNode::finishSpecific()
#else
void VideoSensorNode::finish()
#endif
{
#ifdef FUNCTION_TRACE_DEBUG
	TRACE_DEBUG_IN("finishSpecific");
#endif

	recordScalar("------------------------------------- simTime=", simTime());
#ifdef WITH_INTRUSION
	recordScalar("stat.scheduling.nbActivityMsgSent", this->nbActivityMsgSent);
	recordScalar("stat.cam.intrusionViewCount", this->intrusionViewCount);
        recordScalar("stat.cam.imageCaptureCount", this->imageCaptureCount);
	recordScalar("stat.cam.intrusionOccludedCount", this->intrusionOccludedCount);
#ifdef CASTALIA
	recordScalar("stat.cam.nbImageSent", this->imageSN);
#endif

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER	
	recordScalar("stat.criticality.nbAlertMsgRec", this->nbAlertMsgRec);
	recordScalar("stat.criticality.nbAlertMsgSent", this->nbAlertMsgSent);
	recordScalar("stat.criticality.nbAlertMsgProc", this->nbAlertMsgProc);
	recordScalar("stat.criticality.nbHasBeenInAlertedMode", this->nbHasBeenInAlertedMode);
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER	
	recordScalar("stat.criticality.nbTotalReinforcementMsgRec", this->nbTotalReinforcementMsgRec);
	recordScalar("stat.criticality.nbReinforcementMsgSent", this->nbReinforcementMsgSent);
#endif	
#endif	
#endif	

#ifdef CASTALIA
	double nodeConsumedEnergy = resMgrModule->getSpentEnergy();

	recordScalar("stat.energy.spentEnergy", nodeConsumedEnergy);
	cancelAndDelete(this->forceSendImageEv);

	collectOutput("Consumed AppEnergy", "1-NodeAll", nodeConsumedEnergy);

	// we compute here the real baseline consumption by removing the time spent in image capture and processing
	double baseLineConsumedEnergy=(SIMTIME_DBL(simTime())
								   - this->intrusionViewCount*(this->timeForImageCapture+this->timeForImageProcessing+this->timeForImageEncoding)
								   - this->imageCaptureCount*(this->timeForImageCapture+this->timeForImageProcessing))
									*this->imageNodeBaselineConsumption;

	collectOutput("Consumed AppEnergy", "2-Baseline", baseLineConsumedEnergy);
	collectOutput("Consumed AppEnergy", "3-Radio", nodeConsumedEnergy-this->appConsumedEnergy-baseLineConsumedEnergy);
	collectOutput("Consumed AppEnergy", "4-AppAll", this->appConsumedEnergy);
#endif	

	cancelAndDelete(this->imageCaptureEv);
	cancelAndDelete(this->allowChangeStatusEv);
	cancelAndDelete(this->stopCamRotationEv);
						
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	cancelAndDelete(this->toNormalCriticalityLevelEv);
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
	cancelAndDelete(this->canReinforceEv);
#endif
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
	cancelAndDelete(this->startMasterSelectionEv);
#endif

#ifdef SCHEDULE_PROB
#ifdef SCHEDULE_PROB_CHAOS
	recordScalar("stat.scheduling.chaos.totalChaosIteration", this->chaos_totalIteration);
#endif
	recordScalar("stat.scheduling.chaos.totalProbSelfSchedule", this->prob_nbSelfSchedule);
	recordScalar("stat.scheduling.chaos.totalProbWakeup", this->prob_totalWakeup);
	cancelAndDelete(this->probStopActivityEv);
	cancelAndDelete(this->probStartActivityEv);
#endif
	this->neighbors.clear();

#ifdef FUNCTION_TRACE_DEBUG
	TRACE_DEBUG_OUT("finishSpecific");
#endif
}

#ifdef CASTALIA
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
#include "videoSensorNode-BestCover-methods.c"
#endif
#endif
