// C. Pham, Congduc.Pham@univ-pau.fr, www.univ-pau.fr/~cpham 

#ifndef __VIDEOSENSORNODE_H
#define __VIDEOSENSORNODE_H

using namespace std;

#include <set>
#include <list>
#include <iostream>
#include <sstream>
#include <string>
#include <omnetpp.h>
#include "geometry/Point.h"
#include "geometry/Segment.h"
#include "geometry/Triangle.h"
#include "geometry/Polygon.h"
#include "videoModelDefine.h"

#ifdef CASTALIA

// to define payload size
#define BYTE_T				1
#define FLOAT_T				6
#define BYTE_BIT_FIELD_T	1

#ifdef BEST_COVERSET_SELECTION_STRATEGIES
#include "twoHopData_m.h"
#include "coverMember_m.h"
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
#include "TestMacControl_m.h"
#include "coversAndCaptureRate_m.h"
#endif

#include "teestream.h"
// use ev_v to write both in the OMNET++'s ev output stream and the Castalia's trace() stream
 #define ev_v videoOutputStream << setw(16) << simTime() << setw(40) << getFullPath() << " "
// #define ev_v trace()
// #define ev_v ev

#ifdef CASTALIA_VA
#include "VirtualApplication.h"
#else
#include "ApplicationPacket_m.h"
#include "SensorManagerMessage_m.h"
#include "ResourceManager.h"
#include "Radio.h"
#include "VirtualMobilityManager.h"
#include "CastaliaModule.h"
#endif

#include "connectivityMap.h"

#define SELF_NETWORK_ADDRESS selfAddress.c_str()

#ifdef GPSR_ROUTING
// needed to pass geographic information to the GPSR routing layer
#include "GpsrRoutingControl_m.h"
#endif

#ifdef CASTALIA_CONNECTIVITYMAP
#define NEIGHBORS_DISCOVERY_DELAY 2 // maybe we need less time as a prior connectivity map has been performed ?
#else 
#define NEIGHBORS_DISCOVERY_DELAY 5 // seconds
#endif

#define NB_IDREPORT_PACKET 5

#else
// TAKEN FROM CASTALIA v3.1
#define BROADCAST_MAC_ADDRESS -1
#define BROADCAST_NETWORK_ADDRESS "-1"
#define SINK_NETWORK_ADDRESS "SINK"
#define PARENT_NETWORK_ADDRESS "PARENT"

enum CastaliaMessageTypes {
	NODE_STARTUP = 1,
	TIMER_SERVICE = 2,

	OUT_OF_ENERGY = 3,
	DESTROY_NODE = 4,

	SENSOR_READING_MESSAGE = 5,
	RESOURCE_MANAGER_DRAW_POWER = 6,
	PHYSICAL_PROCESS_SAMPLING = 7,

	WC_SIGNAL_START = 9,
	WC_SIGNAL_END = 10,
	WC_NODE_MOVEMENT = 11,

	RADIO_CONTROL_MESSAGE = 21,
	RADIO_CONTROL_COMMAND = 22,
	RADIO_ENTER_STATE = 23,
	RADIO_CONTINUE_TX = 24,

	MAC_LAYER_PACKET = 30,
	MAC_CONTROL_MESSAGE = 31,
	MAC_CONTROL_COMMAND = 32,

	NETWORK_LAYER_PACKET = 40,
	NETWORK_CONTROL_MESSAGE = 41,
	NETWORK_CONTROL_COMMAND = 42,

	APPLICATION_PACKET = 50
};
// if only OMNET++ is used then ev_v is simply the ev stream
#define ev_v ev
#endif

// set to a relatively high number to avoid segmentation fault. Need to implement a more dynamic scheme.
#define MAX_COVERSET_TABLESIZE 1200
#define MAX_NB_CAMERA 4

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER

#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
// time during which we do not perform reinforcement, avoids reinforcing for the same intrusion
#define DYNAMIC_REINFORCEMENT_DEAF_PERIOD 2
#define ALERTED_NODE_CRITICALITY_LEVEL 0.6
#endif

#endif

#define STARTUP_DELAY 0.05

#ifdef BEST_COVERSET_SELECTION_STRATEGIES
#define THRESHOLD_ENERGY 10
#endif

typedef std::set<int>        SetInt;
typedef std::set< SetInt > SetSetInt;
typedef std::list<int>        ListInt;
typedef std::list< ListInt > ListListInt;

#ifdef SEND_REAL_IMAGE
typedef std::vector<unsigned char> PacketsForOneImage;
typedef std::vector<PacketsForOneImage> AllPacketsImage;

struct imageInfoRecord {

	int nodeID; // the node who sent the image
	int currentImageSN; // the SN of the current received image
	int displayedImageSN; // the SN of the last displayed image
	int nbPackets; // the total number of packets for that image, if known
	int imageSize; // the size of the image file (BMP,JPG,...) in case CRAN_ENCODING is used, the real size in bytes of the encoded
	int nbPacketsReceived; // how many packet from the image has been received
	SetInt packetSNSet;
	double firstPacketReceived; // keeps track of when the first packet was received
	double timePacketWasSent; // keeps track of when the packet was sent
	string imageFilename; // name under which the file will be saved by the sink on the disk
	cMessage* displayImageEv; // msg for displaying the image after a given time period, needed to handle packet losses
	// could be a vector of vector in order to store several images at the same time and choose which one to display
	// in case of multipath routing and packet losses
	std::vector<unsigned char> allPackets; // stores all the received image data
	char* raw_image;

	bool cranEncoding;
	FILE* pFile;
	int imageSizeX;
	int imageSizeY;
	int qualityFactor;
	string imageBMPOriginalFilename;
	int imageType;
};

#endif

class VideoSensorNode;

// the structure that define a sensor node
// it is used by the coordinator node to get info on all sensor nodes (see coordNode.h)
// and by the sensor itself to keep information on its neighbors
struct noeud {
	int id;
	VideoSensorNode* theNode;
	Point p;
	Point v;
	Point b;
	Point c;
	int nbCamera;
	double dov[MAX_NB_CAMERA];
	double alpha[MAX_NB_CAMERA];
	Triangle fov[MAX_NB_CAMERA];
#ifdef DISPLAY_SENSOR
	Triangle fov_display[MAX_NB_CAMERA];
#endif
	int Energy;
	bool isActive;
	bool isAlive;
#ifdef SCHEDULE_PROB
	int nbWakeup;
#endif
	bool isMobile;
	bool isMoving;
	double startMobilityTime;
	double mobilitySpeed;
	double mobilityDuration;
	// for camera rotation capability
	bool isCamRotatable;
	bool isCamRotating;
	double startCamRotationTime;
	double camRotationTime;
	double camRotationCount;

#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	bool hasSentAlertMsg;
	int lastSeq;
#endif

#ifdef BEST_COVERSET_SELECTION_STRATEGIES
	int numNeighbors;
	bool notEmptyNeighborSet;
	noeud *myNeighbors;
	double speed; // current capture rate
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
   	double capturingRate;
   	int noOfCoverSets;

	noeud(){
		capturingRate = -1.0;
		noOfCoverSets = 0;
	}
#endif
};

#ifdef BEST_COVERSET_SELECTION_STRATEGIES

struct elem {
	   int id;
	   Point p;
	   double speed; // current capture rate
	   noeud *listNeighbors;
	   int listNeighborsSize;
};

#endif

struct coversetElem {
	SetInt aCoverset;
	int coversetSize;
	double percentageCoverage;
	// for future use
	double energy;
	// for future use
	double captureRate;
	// means that all nodes are active
	bool isActive;
	// means that one or more nodes of the cover set is dead, not inactive
	bool isDead;
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
	elem *aCompleteCoverset;
	int degreeOfJoint; // number of shared nodes with other covers of the same node.
	// weight of metrics
	double weightSize;
	double weightDependency;
	double weightDistance;
	double weightAv1HopNeighbors;
	double weightClosestAv1HopNeighbors;
	double weightAv2HopNeighbors;
	double weightClosestAv2HopNeighbors;
	double score;
	// Members
	bool belong(int); // to check if a node belongs to the cover set
#endif
#ifdef WITH_INTRUSION
#ifdef INTRUSION_POLYGON
	int nbPointPolygonSeen;
#endif
#endif
};

#ifdef SCHEDULE_PROB
// time in seconds of the activity period
#define PROB_ACTIVITY_PERIOD 20
// when the node got wake up several rounds, each round is a bit shorter
#define PROB_ACTIVITY_PERIOD_DECREMENT 5
// how many times do we retry when the next node is either already active or dead
#define PROB_NB_RETRY_FIND_NEXT_NODE 5

#ifdef SCHEDULE_PROB_CHAOS
struct chaos_x {
	unsigned int xq, yq, y, out;
};

// how many minimum chaotic iterations
#define CHAOS_NB_INITIAL_ITERATION 8
#endif
#endif

// the maximum depth of view in meter and in internal representation
#define MAX_DOV 80
#define MAX_DOV_INTERNAL (MAX_DOV*MAXFIELDUNIT/MAXFIELDMETER)

// time during which no change of status can occur, to avoid oscillations
#define TO_INACTIVE_STATUS_LATENCY_TIME 2.0
#define TO_ACTIVE_STATUS_LATENCY_TIME 4.0

#ifdef RETRY_CHANGE_STATUS_WPROBABILITY
// threshold probability to change the status after change as been refused: uniform(0,1) > RETRY_CHANGE_STATUS_PROBABILITY
#define RETRY_CHANGE_STATUS_PROBABILITY_TO_ACTIVE 	0.6
#define RETRY_CHANGE_STATUS_PROBABILITY_TO_INACTIVE	0.2
#endif

class coordNode;

// a sensor node
#ifdef CASTALIA
#ifdef CASTALIA_VA
class VideoSensorNode : public VirtualApplication
#else
class VideoSensorNode : public CastaliaModule
#endif
#else
class VideoSensorNode : public cSimpleModule
#endif
{ 
#ifdef CASTALIA
#ifndef CASTALIA_VA
	/* TAKEN FROM THE CASTALIA PACKAGE  */ 
	/*--- The .ned file's parameters ---*/
	string applicationID;
	int priority;
	int packetHeaderOverhead;
	int constantDataPayload;
	bool isSink;

	/*--- Custom class parameters ---*/
	int self;				// the node's ID
	string selfAddress;
	ResourceManager *resMgrModule;			//a pointer to the object of the Resource Manager Module
	VirtualMobilityManager *mobilityModule;	//a pointer to the mobilityModule object
	Radio *radioModule;				//a pointer to the radio module object
	int disabled;
	double cpuClockDrift;
#endif
	// with Castalia, we use both ev and trace(), so we define here a tee stream variable for being able to write
	// to both streams at the same time
	teestream videoOutputStream;
	// Our own neighbor discovery mechanism when running with Castalia
	connectivityMap connectivityTool;
#endif

	int id; // the node's ID
	coordNode *CoordNode;

	// sending overhead properties
	///////////////////////////////
	double a_time_send; // in .ini file, the a parameter to define the minimum time between 2 packet generation
	double b_time_send; // in .ini file, the b parameter to define the minimum time between 2 packet generation
	double globalSendTime; // in .ini file, in ms, if non-zero, will overwrite the aTimeSend and bTimeSend definations
	double globalRelayTime; // in .ini file, in ms

	simtime_t nextSendTime; // indicates the time for the next possible packet send

	// physical position properties
	///////////////////////////////
	Point p,v,b,c,g,gp,gv,gb,gc; // points that define the sensor's FoV
	double lineOfSight; // line of sight

	// coverage & camera properties
	///////////////////////////////
	double myCaptureRate; // in .ini file
	double minCaptureRate; // in .ini file
	double maxCaptureRate; // in .ini file
	double staticCaptureRate; // in .ini file
	int nbCamera; // in .ini file
	double aov[MAX_NB_CAMERA]; // in .ini file, not used the moment
	double dov[MAX_NB_CAMERA];	// in .ini file, not used the moment
	double alpha[MAX_NB_CAMERA]; // 2.alpha=angle of view
	Triangle myFoV[MAX_NB_CAMERA];
	bool cameraCycling; // in .ini file
	bool cameraSimultaneous; // in .ini file
	int nextActiveCamera;

	std::set<int> A;
	std::set<int> B;
	std::set<int> C;
	std::set<int> G;
	std::set<int> GP;
	std::set<int> GV;
	std::set<int> GB;
	std::set<int> GC;
	std::set<int> AG;
	std::set<int> BG;
	std::set<int> CG;
	SetSetInt nodeCoverset; // only stores the set the node's ID
	coversetElem coversetTable[MAX_COVERSET_TABLESIZE];
	std::map<SetInt, double> coveragePerCoversetMap;
	int activeCoversetIndex;

	bool selectiveFOV;
	bool virtualFOV;
	bool onlyAltBC;

	// scheduling properties
	///////////////////////////////
#ifdef CASTALIA
	bool alwaysActive; // in .ini file
	double forcedCoversetActivationAt; // in .ini file
#endif
	bool isActive;
	bool isDead;
	double proba;
	double load; // in .ini file, not used for the moment
	double maxCriticalityLevel;
	double criticalityLevel; // in .ini file
	double minCriticalityLevel; // in .ini file
	double maxCriticalityLevelPeriod; // in .ini file
	int maxDefinedCoverSetNumber; // in .ini file
	int forceNumberOfCoversets; // in .ini file
	int minimumCoverSetNumber;
	bool canChangeStatus;
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	bool hasAlertedNeighbors;
	bool isAlerted;
	unsigned int nbAlertMsgRec;
	unsigned int nbAlertMsgSent;
	unsigned int nbAlertMsgProc;
	unsigned int nbHasBeenInAlertedMode;
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
	bool canReinforce;
	unsigned int nbReinforcementMsgRec;
	unsigned int nbTotalReinforcementMsgRec;
	unsigned int nbReinforcementMsgSent;
#endif
#endif
	unsigned int nbActivityMsgSent;
	unsigned int nbPositionPacketReceived;
	unsigned int nbStatusPacketReceived;
	
#ifdef CRITICALITY_ADAPTATIVE_MAC
	unsigned int coverAndCapturePacketReceived;
	unsigned int alertResponseCount;
#endif	
	
#ifdef RETRY_CHANGE_STATUS
	bool retryChangeStatus;
	bool retryChangeStatusTo;
#endif
#ifdef SCHEDULE_PROB
#ifdef SCHEDULE_PROB_CHAOS
	struct chaos_x chaos_in;
	unsigned int chaos_NQ;
	cOutVector nbChaosIterationStatVector;
#endif
#endif

	// mobility properties
	///////////////////////////////
	bool isMobile; // in .ini file
	double isMobileProb; // in .ini file, if isMobile probability to define whether the node is mobile or not
	bool isCamRotatable; // in .ini file
	double isCamRotatableProb; // in .ini file, if isCamRotatable probability to define whether the camera is rotatable or not
	double maxMobilitySpeed; // in .ini file, in m/s
	double maxCamRotationTime; // in .ini file, in time (s) for a complete rotation, 5s -> 1,256637 rad /s for instance
	double maxCamRotationCount; // in .ini file, in number of rotations

	bool isMoving;
	bool isCamRotating;
	double mobilitySpeed;
	double mobilityDuration;
	double startMobilityTime;
	double startCamRotationTime;
	double camRotationCount;
	double camRotationTime;
	double lastRotationCount;

	// energy properties
	///////////////////////////////
	double maxBatteryLevel; // in .ini file
	double energyPerCapture; // in .ini file
	double imageNodeBaselineConsumption;
	double measuredEnergyPerImageCapture; // in .ini file, in mJ. Here in J
	double measuredEnergyPerImageEncoding; // in .ini file, in mJ. Here in J
	double measuredEnergyPerImageProcessing; // in .ini file, in mJ. Here in J
	double timeForImageCapture; // in .ini file, in ms. Here in s
	double timeForImageEncoding; // in .ini file, in ms. Here in s
	double timeForImageProcessing; // in .ini file, in ms. Here in s
	double energyLevel;
	double appConsumedEnergy;

	// neighbors properties
	///////////////////////////////
	unsigned int nbNeighbors;
	// store the neighbors' info
	// noeud *neighbors;
	std::vector<noeud> neighbors;
	std::set<int> activeNeighbors;

	// network properties
	///////////////////////////////
#ifdef CASTALIA
	bool idReportToSink; // in .ini file
	string recipientAddress; // in .ini file
#endif

	// intrusion properties
	///////////////////////////////
#ifdef WITH_INTRUSION
	unsigned int intrusionViewCount;
	unsigned int intrusionOccludedCount;
	double forcedIntrusionProb; // in .ini file
#ifdef CASTALIA
	bool activateCoversetOnIntrusion; // in .ini file
	bool propagateCoversetActivation; // in .ini file
	bool receivedCoversetActivation;
	int imageCountOnIntrusion; // in .ini file
	int imageCountOnIntrusionCounter;
	int imageCountOnCoversetActivation; // in .ini file
	int imageCountOnCoversetActivationCounter;
#endif
#endif

	// image properties
	///////////////////////////////
	unsigned int imageCaptureCount;
#ifdef CASTALIA
	double imageSentTime; // keep the sent time of the current image
	int imageSN; // keep the SN of the images sent
	bool sendImageOnIntrusion; // in .ini file
	bool forceSendImage; // in .ini file
	int forceSendImageCounter;
	double forceSendImageInterval; // in .ini file
	int imageByteSize; // in .ini file
	int imageChunkSize; // in .ini file
	string imageFilename; // in .ini file
	bool cranEncoding; // in .ini file
	string imageBMPOriginalFilename; // in .ini file

	bool displayReceivedImage; // in .ini file
	bool waitForKeyPressWhenDisplayImage; // in .ini file
	double displayReceivedImageTimer; // in .ini file
	bool keepImageFile; // in .ini file
	double imagePacketLossProb; // in .ini file
	double imageByteCorruptedProb; // in .ini file
	int imageSuccPacketLossCount;
#ifdef SEND_REAL_IMAGE
	std::map <int,imageInfoRecord> imageInfoRecordFromNodes; // keeps information on all the images received by other nodes
#endif
#endif

	// simulation model's messages
	///////////////////////////////
	cMessage *diffusionEv; 	// trigger a broadcast of sensor's info (position,...)
#ifdef CASTALIA
	// indicates when cover-sets computation can begin as Castalia need a neighbor discovery process prior
	// to launch the cover-set computation
	cMessage *startComputeCoverEv;
	cMessage *forceSendImageEv;
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
	// indicates when 2-hop data diffusion can begin as we need a neighbor discovery process and cover-set computation
	// prior to launch the 2-hop data diffusion
	cMessage *start2HopDataDiffusion;
	cMessage *startCoverMemberDiffusion;
	cMessage *startComputeScore;
#endif
#endif
	cMessage *imageCaptureEv; // periodic event, according to the capture rate
	cMessage *allowChangeStatusEv; // timer for enabling status (active/sleep) changes
	cMessage *stopCamRotationEv; // notify that the camera stops rotating
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	// notifies that should go back to normal after a period of maximum criticality level
	cMessage *toNormalCriticalityLevelEv;
#ifdef DYNAMIC_REINFORCEMENT_CRITICALITY_WITH_BEZIER
	cMessage *canReinforceEv; // indicates that node can do reinforcement (timer)
#endif	
#endif		
#ifdef SCHEDULE_PROB
	cMessage *probStopActivityEv; // stop prob-based scheduling activity
	cMessage *probStartActivityEv; // stop prob-based scheduling activity
#endif

#ifdef CRITICALITY_ADAPTATIVE_MAC
	cMessage *startMasterSelectionEv;
#endif

protected:
	// The following redefined virtual function holds the algorithm.
	virtual void initialize();
	virtual void handleMessage(cMessage *msg);
#ifdef CASTALIA
	virtual void finishSpecific();
#else
	virtual void finish();
#endif

public:

	// scheduling properties
	///////////////////////////////
#ifdef SCHEDULE_PROB
#ifdef SCHEDULE_PROB_CHAOS
	unsigned int chaos_nbIteration;
	int chaos_totalIteration;
#endif
	int prob_totalWakeup;
	unsigned int prob_nbActivityRound;
	unsigned int prob_nbRetry;
	unsigned int prob_nbSelfSchedule;
#endif

	// video sensor node's methods
	///////////////////////////////
	void setIcon(const char*, int);
	double getCaptureRate();
	void setCaptureRate(double);
	bool setActiveStatus(bool status, bool dead=false);
	void updateIconText();
	void sendPosition(bool updating=false);
	void sendActivity();
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	void sendAlert(double, int, int);
	void setAlertCriticalityLevel(double);
#endif

	void handlePositionBroadcast();
#ifdef CASTALIA
	void handleNodeStartup();
#ifdef CASTALIA_CONNECTIVITYMAP
	void handleHelloInitiate();
#endif
	void handleIdReport();
	void handleComputeCover();
#ifdef SEND_REAL_IMAGE
	void handleDisplayImage(int, int);
#endif
#endif
	void handleImageCapture();
	void handleAllowChangeStatus();
	void handleToNormalCriticalityLevel();
#ifdef SCHEDULE_PROB
	void handleProbStopActivity();
#endif
	void handleChangeTxPower();
#ifdef CASTALIA
#ifdef CASTALIA_CONNECTIVITYMAP
	void handleHelloPacket(cMessage*);
#endif
#endif
#ifdef DYNAMIC_CRITICALITY_WITH_BEZIER
	void handleAlertPacket(cMessage*);
#endif
#ifdef CRITICALITY_ADAPTATIVE_MAC
	void handleMasterSelection();
	int masterSelection();
	void handleAlertResponse(cMessage*);
	void sendCoverSetsAndCaptureRate();
	void handleCoverSetsAndCaptureRatePacket(cMessage*);
#endif
	void handlePositionPacket(cMessage*);
	void handleUpdatePosPacket(cMessage*);
	void handleActivityPacket(cMessage*);
#ifdef CASTALIA
	void handleCoversetActivationPacket(cMessage*);
	void handleIdReportPacket(cMessage*);
	void handleImagePacket(cMessage*);
#ifdef BEST_COVERSET_SELECTION_STRATEGIES
	void handle2hopDataDiffusion();
	void handleCoverMemberDiffusionInitiate();
	void handleComptuteScore();

	void handle2hopDataDiffusionPacket(cMessage*);
	void handleStartCoverMemberDiffusionPacket(cMessage*);
	void handleCoverMemberDiffusionPacket(cMessage*);
#endif
#endif

#ifdef SCHEDULE_PROB
	void probWakeNode();
	double getProbActivityTime();
#endif
	void computeCoverset();
	int getNeighborIndexById(int theId);
	int computeActiveCoverset(int&);
	int getNumberPossibleCoverset();
	double getEnergyLevel() { return energyLevel; }
	double getCastaliaEnergyLevel();
	float distance(Point, Point);
	bool isCoveredBy(Point&, int);
	Point sampleMyTriangle();	
	bool imageCapture();
	void findTriangleCoord(Point, Point, Point&, Point&, double, double);
#ifdef CASTALIA
	int getCoversetIndex();
	void sendCoversetActivation();
	void setSinkInfoInRoutingLayer(int id, int x, int y);
	const char* getSinkAddressString();
	void sendImage(const char *, int);
	void setRadioTxOutputPower(double);

	// sending overhead properties
	///////////////////////////////
	simtime_t getTimeBetweenSend(int size);

#ifdef BEST_COVERSET_SELECTION_STRATEGIES
#include "videoSensorNode-BestCover-methods.h"
#endif
#ifdef CASTALIA_VA
	void fromNetworkLayer(ApplicationPacket *, const char *, double, double);
#else
	void toNetworkLayer(cMessage * msg);
	void toNetworkLayer(cPacket *, const char *);
#endif
#endif
};
#endif
